// CustomStateTreeView.cs: Contributed by Chris Sells [csells@sellsbrothers.com]
// A treeview that allows a custom image list for the check state.
#region Copyright  2002-2003 The Genghis Group
/*
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from the
 * use of this software.
 * 
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, subject to the following restrictions:
 * 
 * 1. The origin of this software must not be misrepresented; you must not claim
 * that you wrote the original software. If you use this software in a product,
 * an acknowledgment in the product documentation is required, as shown here:
 * 
 * Portions Copyright  2002-2003 The Genghis Group (http://www.genghisgroup.com/).
 * 
 * 2. No substantial portion of the source code of this library may be redistributed
 * without the express written permission of the copyright holders, where
 * "substantial" is defined as enough code to be recognizably from this library. 
*/
#endregion
#region Limitiations
// -Can't use 256-color images for custom states
#endregion

using System;
using System.Reflection;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using System.Runtime.InteropServices;

namespace Genghis.Windows.Forms
{
    /// <summary>
    /// A treeview that allows a custom image list for the check state.
    /// </summary>
    public class CustomStateTreeView : TreeView
    {
        /// <summary>
        /// Set the check state of a node in the tree.
        /// </summary>
        /// <param name="node">The node</param>
        /// <param name="state">0-based index into image list (0 means no check state)</param>
        public void SetCheckState(TreeNode node, int state)
        {
            TVITEMA item = new TVITEMA();
            item.mask = TVIF_HANDLE | TVIF_STATE;
            item.hItem = node.Handle;
            item.stateMask = TVIS_STATEIMAGEMASK;
            item.state = (uint)IndexToStateImageMask(state);

            uint    ret = GetSetCheckState(this.Handle, TVM_SETITEMA, 0, item);
            if( ret == 0 ) throw new ApplicationException("Unable to set check state: " + state);
        }

        /// <summary>
        /// Get the check state of a node in the tree.
        /// </summary>
        /// <param name="node">The node</param>
        /// <returns>The check state</returns>
        public int GetCheckState(TreeNode node)
        {
            TVITEMA item = new TVITEMA();
            item.mask = TVIF_HANDLE | TVIF_STATE;
            item.hItem = node.Handle;
            item.stateMask = TVIS_STATEIMAGEMASK;

            uint    ret = GetSetCheckState(this.Handle, TVM_GETITEMA, 0, item);
            if( ret == 0 ) throw new ApplicationException("Unable to get check state");

            return (int)StateImageMaskToIndex((int)item.state);
        }

        /// <summary>
        /// Image list for the check state (image 0 is ignored)
        /// </summary>
        public ImageList CheckStateImageList
        {
            get { return checkStateImages; }
            set
            {
                SetImageList(this.Handle, TVM_SETIMAGELIST, TVSIL_STATE, value != null ? value.Handle : IntPtr.Zero);
                checkStateImages = value;
            }
        }
    
        // HACK: Unfortunately there's no notification that the image list has changed
        // in the control, so there's no way to catch this behavior if the base class
        // treeview type is used instead of this one. Luckily the designer-generated
        // code, where this will most likely be changed, is type-safe
        public new System.Windows.Forms.ImageList ImageList
        {
            get
            {
                return base.ImageList;
            }

            set
            {
                base.ImageList = value;

                // HACK: Fix bug in unmanaged control that requires state image list
                // to be set after normal image list
                if( checkStateImages != null ) CheckStateImageList = checkStateImages;
            }
        }

        protected ImageList checkStateImages;

    #region Interop
        private const uint    TVSIL_STATE = 2;
        private const uint    TVIF_STATE = 0x0008;
        private const uint    TVIF_HANDLE = 0x0010;
        private const uint    TV_FIRST = 0x1100;
        private const uint    TVM_GETITEMA = TV_FIRST + 12;
        private const uint    TVM_SETITEMA = TV_FIRST + 13;
        private const uint    TVM_SETIMAGELIST = TV_FIRST + 9;
        private const uint    TVIS_STATEIMAGEMASK = 0xF000;

        [ StructLayout(LayoutKind.Sequential) ]
        private class TVITEMA
        {
            public uint     mask = 0;
            public IntPtr   hItem = IntPtr.Zero;
            public uint     state = 0;
            public uint     stateMask = 0;
            public uint     pszText = 0;
            public int      cchTextMax = 0;
            public int      iImage = 0;
            public int      iSelectedImage = 0;
            public int      cChildren = 0;
            public uint     lParam = 0;
        }

        protected int IndexToStateImageMask(int i)
        {
            return (int)((uint)i << 12);
        }

        protected int StateImageMaskToIndex(int i)
        {
            return (int)((uint)i >> 12);
        }

        [DllImport("user32.dll", EntryPoint = "SendMessage")]
        private extern static uint GetSetCheckState(IntPtr hwnd, uint msg, uint zero, TVITEMA item);

        [DllImport("user32.dll", EntryPoint = "SendMessage")]
        private extern static uint SetImageList(IntPtr hwnd, uint msg, uint imageListType, IntPtr imageList);
    #endregion
    }
}
