// Command.cs: Contributed by Paul Bartrum [paulbartrum@mail.com]
// Command abstraction classes.
#region Copyright 2002 The Genghis Group
//    Copyright 2002 The Genghis Group
//    This software is provided 'as-is', without any express or implied warranty. 
//    In no event will the authors be held liable for any damages arising from the use of this software.
//    Permission is granted to anyone to use this software for any purpose, including commercial applications, 
//    subject to the following restrictions:
//    
//    The origin of this software must not be misrepresented; you must not claim that you wrote 
//    the original software. If you use this software in a product, an acknowledgment in the product 
//    documentation is required, as shown here:
//
//        Portions copyright 2002 The Genghis Group (http://www.genghisgroup.com/).
//  
//    No substantial portion of the source code of this library may be redistributed without the 
//    express written permission of the copyright holders, where "substantial" is defined as enough 
//    code to be recognizably from this library.
#endregion Copyright

#region History
//	6 April
//		- Initial version.
#endregion

using System;
using System.IO;
using System.Collections;
using System.Xml;
using System.Diagnostics;
using System.Reflection;
using System.Drawing;
using System.Windows.Forms;
using T = Genghis.Command;

// TODO: Add Visible property(?)
// TODO: Define routing on tree.
// TODO: Thread safety.

namespace Genghis {

	public delegate void CommandEventHandler(Command command, CommandEventArgs e);

	public class CommandEventArgs : System.EventArgs {
		public enum EventType {
			Execute,
			Update,
		}
	
		internal Command command;
		internal object source;
		internal EventType type;
		internal CommandCollection commandPath;
		internal bool executeHandled;
		internal bool updateHandled;
	
		/// <summary>
		/// Initializes a new instance of the CommandEventArgs class with the Source
		/// property set to the given value.</summary>
		public CommandEventArgs(object source) {
			this.source = source;
			commandPath = new CommandCollection();
			executeHandled = false;
			updateHandled = false;
		}
	
		/// <summary>
		/// The command that was invoked.</summary>
		public Command Command {
			get { return command; }
		}
	
		/// <summary>
		/// The command that the handler is attached to (can be an ancestor of Command).</summary>	
		public Command HandlerCommand {
			get { return commandPath.Count > 0 ? commandPath[0] : null; }
		}
	
		/// <summary>
		/// The UI element that triggered this event.  Can be null.</summary>
		public object Source {
			get { return source; }
		}
		
		public EventType Type {
			get { return type; }
		}
		
		public void Route() {
			HandlerCommand.DoRouting(this);
		}
	}

	public class Command {

		/// <summary>
		/// Represents the amount to the which the command has been applied to the selection.</summary>
		public enum SelectedState {
			/// <summary>
			/// The standard state.</summary>
			Unselected,

			/// <summary>
			/// The command will be represented by a check mark in a menu, and a pushed toolbar button in a toolbar.</summary>
			Selected,

			/// <summary>
			/// The command has been partially applied.</summary>
			Indeterminate,
		}

		/// <summary>
		/// Represents the various different behaviours that commands can have.</summary>
		public enum CommandType {
			/// <summary>
			/// Command is standard or command is toggle or radio and has not been applied.</summary>
			Standard,

			/// <summary>
			/// Applying the command twice in succession will have no effect
			/// (ie. the second command will reverse the effects of the first).</summary>
			Toggle,

			/// <summary>
			/// The child commands of this command are mutually exclusive to each other.</summary>
			Radio
		}
		
		[AttributeUsage(AttributeTargets.Method, AllowMultiple = false)]
		public class ExecuteHandlerAttribute : Attribute {
			string id;

			public ExecuteHandlerAttribute(string id) {
				this.id = id;
			}

			public string ID {
				get { return id; }
			}
		}

		[AttributeUsage(AttributeTargets.Method, AllowMultiple = false)]
		public class UpdateHandlerAttribute : Attribute {
			string id;

			public UpdateHandlerAttribute(string id) {
				this.id = id;
			}

			public string ID {
				get { return id; }
			}
		}




		// Fields.
		string id;

		string defaultText;
		string defaultDescription;

		CommandType type;

		Image icon;
		Shortcut shortcut;

		Command parent;
		CommandCollection commands;	// Child commands.

		// UI state.
		//  - modified by the update handler.

		string text;				// Text of the command.  Can contain & for use as a keyboard mnemonic.
		string description;			// Long description, for use in displaying help.
		bool enabled;				// If true, the command can be executed,
									// otherwise the command cannot be executed at this time.
		SelectedState selected;



		// Constructors.


		/// <summary>
		/// Initializes a Command with null values (command is of the standard type).</summary>
		public Command() : this(null, CommandType.Standard) {
		}

		/// <summary>
		/// Initializes a Command with a label (command is of the standard type).</summary>
		public Command(string text) : this(text, CommandType.Standard) {
		}

		/// <summary>
		/// Initializes a Command with a label and a type.</summary>
		public Command(string text, CommandType type) {
			this.defaultText = text;
			this.type = type;

			// Setup the child commands.
			commands = new CommandCollection();
			commands.ItemsAdded += new CommandCollectionEventHandler(OnCommandsChanged);
			commands.ItemsRemoved += new CommandCollectionEventHandler(OnCommandsChanged);

			// Ensure that the update handler is called the first time a property is queried.
			cacheTime = Environment.TickCount - maxCachePeriod;
		}

		/// <summary>
		/// Loads the command from an XML file.  Not to be called directly.</summary>
		protected Command(XmlReader reader) : this() {	
			Debug.Assert(reader.Name == "command");
			Debug.Assert(reader.NodeType == XmlNodeType.Element);

			string temp;

			id = reader.GetAttribute("id");
			//string imagePath = reader.GetAttribute("image");
			if ((temp = reader.GetAttribute("type")) != null)
				type = (CommandType) Enum.Parse(type.GetType(), temp, true);
			defaultText = reader.GetAttribute("text");
			defaultDescription = reader.GetAttribute("description");
			if ((temp = reader.GetAttribute("shortcut")) != null)
				shortcut = (Shortcut) System.Enum.Parse(shortcut.GetType(), temp, true);

			if (reader.IsEmptyElement)
				return;		// No child commands.

			int initialDepth = reader.Depth;
			while (reader.Read() && reader.Depth != initialDepth) {
				if (reader.Name == "command")
					commands.Add(new Command(reader));
			}
		}


		// Properties.

		/// <summary>
		/// A unique identifier for this command.</summary>
		public string ID {
			get { return id; }
			set { id = value; }
		}

		/// <summary>
		/// The default text for this command.</summary>
		public string DefaultText {
			get { return defaultText; }
			set { defaultText = value; }
		}

		/// <summary>
		/// The default description of what the command does.</summary>
		public string DefaultDescription {
			get { return defaultDescription; }
			set { defaultDescription = value; }
		}

		/// <summary>
		/// The type of the command.  See the definition for
		/// <see cref="CommandType">CommandType</see>.</summary>
		public CommandType Type {
			get { return type; }
		}

		/// <summary>
		/// <c>true</c> if this command is part of a mutually-exclusive group.</summary>
		public bool IsRadioMember {
			get { return Parent != null && Parent.Type == CommandType.Radio; }
		}

		/// <summary>
		/// The keyboard shortcut that invokes this command.</summary>
		public Shortcut Shortcut {
			get { return shortcut; }
			set { shortcut = value; }
		}

		/// <summary>
		/// The icon representing this command.</summary>
		public Image Image {
			get { return icon; }
			set { icon = value; }
		}

		/// <summary>
		/// A collection representing the commands grouped under this command.</summary>
		public CommandCollection Commands {
			get { return commands; }
		}

		/// <summary>
		/// The command that contains this one.</summary>
		public Command Parent {
			get { return parent; }
		}


		// UI properties.

		/// <summary>
		/// A cache of the command state.  This represents what is currently shown in the UI.</summary>
		/// <remarks>
		/// This value may not be completely up to date.  Call
		/// <see cref="PerformUpdate">PerformUpdate</see> to update this value.</remarks>


		/// <summary>
		/// A concise name for the command.</summary>
		public string Text {
			get { PerformUpdate(); return text; }
			set { text = value; }
		}

		/// <summary>
		/// A full description of what the command does.</summary>
		public string Description {
			get { PerformUpdate(); return description; }
			set { description = value; }
		}

		/// <summary>
		/// <c>true</c> if the command can be executed,
		/// <c>false</c> otherwise.</summary>
		public bool Enabled {
			get { PerformUpdate(); return enabled; }
			set { enabled = value; }
		}

		/// <summary>
		/// <c>true</c> if the command has been applied to the current selection,
		/// <c>false</c> otherwise.</summary>
		public bool Selected {
			get { PerformUpdate(); return selected == SelectedState.Unselected ? false : true; }
			set { selected = value ? SelectedState.Selected : SelectedState.Unselected; }
		}

		/// <summary>
		/// <c>SelectedState.Selected</c> if the command has been applied to the
		/// current selection, <c>SelectedState.Unselected</c> if it hasn't, or
		/// <c>SelectedState.Indeterminate</c> if the command has been partially
		/// applied.</summary>
		public SelectedState Selected3 {
			get { PerformUpdate(); return selected; }
			set { selected = value; }
		}

		// Methods.

		/// <summary>
		/// A command was added or removed from the Commands collection.</summary>
		protected virtual void OnCommandsChanged(object sender, CommandCollectionEventArgs e) {
			// Update parent of each command.
			if (e.Added) {
				foreach (Command command in e.Items)
					command.parent = this;
			}
			else {
				foreach (Command command in e.Items)
					command.parent = null;
			}
		}

		internal class ImageList : IDisposable {
			Image imageList;
			Size imageSize;
			System.Collections.Specialized.StringCollection imageIDs;

			public ImageList(string path) {
				imageList = Image.FromFile(path);
				imageSize = new Size(imageList.Height, imageList.Height);

				imageIDs = new System.Collections.Specialized.StringCollection();
			}

			public ImageList(Assembly assembly, string resourceName) {
				Stream stream = assembly.GetManifestResourceStream(resourceName);
				if (stream == null)
					throw new ArgumentException("The specified resource does not exist: " + resourceName);

				imageList = Image.FromStream(stream);
				imageSize = new Size(imageList.Height, imageList.Height);

				imageIDs = new System.Collections.Specialized.StringCollection();
			}

			public void Dispose() {
				if (imageList != null) {
					imageList.Dispose();
					imageList = null;
				}
			}

			public Image GetImage(int index) {
				if (imageList == null)
					throw new InvalidOperationException("This object has already been disposed.");

				Image image = new Bitmap(ImageSize.Width, ImageSize.Height);

				Graphics graphics = Graphics.FromImage(image);
				graphics.DrawImage(imageList, 0, 0, new Rectangle(new Point(index * ImageSize.Width, 0), ImageSize), GraphicsUnit.Pixel);
				graphics.Dispose();

				return image;
			}

			public Size ImageSize {
				get { return imageSize; }
				set { imageSize = value; }
			}

			public void AddID(string id) {
				if (id == null)
					imageIDs.Add("");
				else
					imageIDs.Add(id);
			}

			public int IndexOf(string id) {
				return imageIDs.IndexOf(id);
			}
		}

		/// <summary>
		/// Load the command (and subcommands) from a XML file in the
		/// specified assembly.</summary>
		/// <param baseName="url">
		/// The URL for the file containing the XML data.</param>
		public static Command Load(Assembly assembly, string name) {
			Stream stream = assembly.GetManifestResourceStream(name);
			XmlTextReader reader = new XmlTextReader(stream);
			Command commands = Load(assembly, reader);
			reader.Close();

			return commands;
		}

		/// <summary>
		/// Load the command (and subcommands) from a XML document.</summary>
		/// <param baseName="reader">
		/// The XmlTextReader representing the XML document.</param>
		public static Command Load(Assembly assembly, XmlTextReader reader) {
			Command root = null;
			ArrayList imageLists = new ArrayList();

			while (reader.Read()) {
				if (reader.NodeType == XmlNodeType.Element) {
					if (reader.Name == "command") {
						root = new Command(reader);
					}
					else if (reader.Name == "imageList") {
						if (reader["path"] == null)
							throw new XmlException("No path attribute exists on the imageList element.", null);
						ImageList imageList = new ImageList(assembly, reader["path"]);

						// Read width and height.
						int width = imageList.ImageSize.Width;
						int height = imageList.ImageSize.Height;
						if (reader["width"] != null)
							width = Int32.Parse(reader["width"]);
						if (reader["height"] != null)
							height = Int32.Parse(reader["height"]);
						imageList.ImageSize = new Size(width, height);

						// Read image IDs.
						int initialDepth = reader.Depth;
						while (reader.Read() && reader.Depth != initialDepth) { 
							if (reader.Name == "image")
								imageList.AddID(reader["id"]);
						}

						imageLists.Add(imageList);
					}
				}
			}

			reader.Close();

			// Associate the commands with the images.
			if (root != null)
				root.ApplyImages(imageLists);

			// Dispose of the image lists.
			foreach (ImageList imageList in imageLists)
				imageList.Dispose();

			return root;
		}

		private void ApplyImages(ArrayList imageLists) {
			// Apply images.
			if (ID != null) {
				// Search all image lists.
				foreach (ImageList imageList in imageLists) {
					int index = imageList.IndexOf(ID);
					if (index != -1) {
						Image = imageList.GetImage(index);
						break;
					}
				}
			}

			foreach (Command child in Commands)
				child.ApplyImages(imageLists);
		}

		/// <summary>
		/// Gets the command with the specified ID.</summary>
		/// <param baseName="id">
		/// The ID of the command.</param>
		public Command GetCommandFromID(string id) {
			if (id == null)
				return null;
			if (this.id == id)
				return this;

			foreach (Command subcommand in commands) {
				Command result = subcommand.GetCommandFromID(id);
				if (result != null)
					return result;
			}
			return null;
		}

		/// <summary>
		/// Registers all the command event handlers on an object.</summary>
		/// <param baseName="target">
		/// The class instance on which the event handlers exist.</param>
		public void RegisterHandlers(object target) {
			DoRegisterHandlers(target, true);
		}

		/// <summary>
		/// Registers all the command event handlers on an object.</summary>
		/// <param baseName="target">
		/// The class instance on which the event handlers exist.</param>
		public void UnregisterHandlers(object target) {
			DoRegisterHandlers(target, false);
		}

		// TODO: Catch security exceptions.
		private void DoRegisterHandlers(object target, bool add) {
			Hashtable handlers = GetExecuteHandlers(target);
			IDictionaryEnumerator enumerator = handlers.GetEnumerator();
			while (enumerator.MoveNext()) {
				// Find the command with the specified ID.
				Command command = GetCommandFromID((string) enumerator.Key);

				if (command != null) {
					// Add or remove the handler from the list.
					if (add)	// Register execute handler.
						command.Execute += (CommandEventHandler) enumerator.Value;
					else		// Unregister execute handler.
						command.Execute -= (CommandEventHandler) enumerator.Value;
				}
				else
					Trace.WriteLine("The execute handler with the command ID " + id + " cannot be registered because there is no corresponding command.");
			}

			handlers = GetUpdateHandlers(target);
			enumerator = handlers.GetEnumerator();
			while (enumerator.MoveNext()) {
				// Find the command with the specified ID.
				Command command = GetCommandFromID((string) enumerator.Key);
				if (command != null) {
					// Add or remove the handler from the list.
					if (add)	// Register update handler.
						command.Update += (CommandEventHandler) enumerator.Value;
					else		// Unregister update handler.
						command.Update -= (CommandEventHandler) enumerator.Value;
				}
				else
					Trace.WriteLine("The update handler with the command ID " + id + " cannot be registered because there is no corresponding command.");
			}
		}

		static public Hashtable GetExecuteHandlers(object target) {
			Hashtable result = new Hashtable();

			// Iterate through the methods of the specified instance.
			BindingFlags flags = BindingFlags.Instance | BindingFlags.Public | BindingFlags.NonPublic;
			System.Reflection.MethodInfo[] methods = target.GetType().GetMethods(flags);
			foreach (System.Reflection.MethodInfo method in methods) {

				// Iterate through the attributes attached to the method.
				object[] attributes = method.GetCustomAttributes(true);
				for (int i = 0; i < attributes.Length; i ++) {

					// If the attribute is a ExecuteHandlerAttribute...
					if (attributes[i] is ExecuteHandlerAttribute) {
						string id = ((ExecuteHandlerAttribute) attributes[i]).ID;

						// Get the handler.
						CommandEventHandler handler = null;
						try {
							handler = (CommandEventHandler) Delegate.CreateDelegate(typeof(CommandEventHandler), target, method.Name);
						}
						catch (ArgumentException exception) {
							if (exception.Message == "Error binding to target method.")
								throw new ArgumentException("The method " + method.Name + " can not be converted to a CommandEventHandler delegate, because one or more arguments are of the wrong type.", exception);
							throw;
						}

						result[id] = handler;
					}
				}
			}

			return result;
		}

		static public Hashtable GetUpdateHandlers(object target) {
			Hashtable result = new Hashtable();

			// Iterate through the methods of the specified instance.
			BindingFlags flags = BindingFlags.Instance | BindingFlags.Public | BindingFlags.NonPublic;
			System.Reflection.MethodInfo[] methods = target.GetType().GetMethods(flags);
			foreach (System.Reflection.MethodInfo method in methods) {

				// Iterate through the attributes attached to the method.
				object[] attributes = method.GetCustomAttributes(true);
				for (int i = 0; i < attributes.Length; i ++) {

					// If the attribute is a UpdateHandlerAttribute...
					if (attributes[i] is UpdateHandlerAttribute) {
						string id = ((UpdateHandlerAttribute) attributes[i]).ID;

						// Get the handler.
						CommandEventHandler handler = null;
						try {
							handler = (CommandEventHandler) Delegate.CreateDelegate(typeof(CommandEventHandler), target, method.Name);
						}
						catch (ArgumentException exception) {
							if (exception.Message == "Error binding to target method.")
								throw new ArgumentException("The method " + method.Name + " can not be converted to a CommandEventHandler delegate, because one or more arguments are of the wrong type.", exception);
							throw;
						}

						result[id] = handler;
					}
				}
			}

			return result;
		}


		// Events/Delegates and related methods.

		/// <summary>
		/// Occurs when a control needs to update the command state so
		/// it can draw itself properly.</summary>
		public CommandEventHandler Update;

		/// <summary>
		/// Updates the state of the command by calling the
		/// Update delegate.</summary>
		/// <remarks>
		/// If this command doesn't have an UpdateState handler, the ancestors of
		/// this command are called, in order, until one of the commands handles
		/// this event or the top of the heirarchy is reached.
		/// In the absence of any UpdateState handler, the following rules apply:
		/// 1. If this command (or ancestors) has one or more Execute event handler(s),
		///    the command is considered enabled and unselected.
		/// 2. If this command (and ancestors) has no Execute event handler(s)
		///    but has one or more enabled child commands,
		///    the command is considered enabled and unselected.
		/// 3. If this command has no Execute event handler(s) and
		///    no enabled child commands, the command is disabled and unselected.
		/// </remarks>

		int cacheTime;
		static int cachePeriod = 100;
		const int maxCachePeriod = 10000;

		public void PerformUpdate() {
			// Don't call recursively.
			if (System.Threading.Monitor.TryEnter(this) == false)
				return;

			int tickCount = Environment.TickCount;

			// Check if the state is cached.
			if (tickCount - cacheTime < cachePeriod)
				return;
			cacheTime = tickCount;

			System.Threading.Monitor.Enter(this);
			try {
				string backupText = text;
				string backupDescription = description;
				bool backupEnabled = enabled;
				SelectedState backupSelected = selected;

				text = defaultText;
				description = defaultDescription;
				enabled = true;
				selected = SelectedState.Unselected;

				CommandEventArgs e = new CommandEventArgs(null);
				e.command = this;
				e.type = CommandEventArgs.EventType.Update;
				e.commandPath.Clear();

				Command ancestor = this;
				while (ancestor != null) {
					e.commandPath.Insert(0, ancestor);
					ancestor = ancestor.Parent;
				}

				// Execute the first filter.
				if (e.HandlerCommand.Filter != null)
					e.HandlerCommand.Filter(e.Command, e);
				else
					DefaultFilter(e.Command, e);

				if (e.updateHandled == false) {
					// UpdateState event was not handled.

					// If there is no execute handlers and every subcommand is
					// disabled, then this command is disabled.
					if (e.executeHandled == false) {
						bool disable = true;
						foreach (Command subcommand in commands) {
							if (subcommand.enabled == true) {
								disable = false;
								break;
							}
						}
						if (disable)
							enabled = false;
					}
				}

				if (backupText == text &&
					backupDescription == description &&
					backupEnabled == enabled &&
					backupSelected == selected)
					return;

				OnChanged(new CommandEventArgs(null));
			}
			finally {
				System.Threading.Monitor.Exit(this);
			}
		}

		protected internal virtual void OnUpdate(CommandEventArgs e) {
			if (Update != null)
				Update(e.Command, e);
			else
				e.Route();

			/*
			// First call the UpdateState handler for this command.  If there is
			// no handler, then call ancestors of this command, in order, until one
			// of the commands handles this event or the top of the heirarchy is
			// reached.
			Command ancestor = this;
			while (ancestor != null) {
				if (ancestor.Update != null)
					break;
				ancestor = ancestor.Parent;
			}

			if (ancestor != null) {
				CommandEventArgs e = new CommandEventArgs(null);
				e.Command = this;
				e.HandlerCommand = ancestor;
				ancestor.Update(this, e);
			}
			else {
				// UpdateState event was not handled.

				// Determine if there are any execute handlers.
				bool executeHandlerExists = false;
				ancestor = this;
				while (ancestor != null) {
					if (ancestor.Execute != null) {
						executeHandlerExists = true;
						break;
					}
					ancestor = ancestor.Parent;
				}

				// If there is no execute handlers and every subcommand is
				// disabled, then this command is disabled.
				if (!executeHandlerExists) {
					bool disable = true;
					foreach (Command subcommand in commands) {
						if (subcommand.enabled == true) {
							disable = false;
							break;
						}
					}
					if (disable)
						enabled = false;
				}
			}

			if (backupText == text &&
				backupDescription == description &&
				backupEnabled == enabled &&
				backupSelected == selected)
				return;

			OnChanged(new CommandEventArgs(null));*/
		}


		/// <summary>
		/// Occurs when the UI state of this command changes.</summary>
		public event CommandEventHandler Changed;

		protected virtual void OnChanged(CommandEventArgs e) {
			// Call the changed handler all of the ancestors of this command, in
			// order, until the top of the heirarchy is reached.
			Command ancestor = this;
			while (ancestor != null) {
				if (ancestor.Changed != null)
					ancestor.Changed(this, e);
				ancestor = ancestor.Parent;
			}
		}


		/// <summary>
		/// Occurs when the user clicks on a menu item or toolbar item
		/// that represents this command.</summary>
		public event CommandEventHandler Execute;

		/// <summary>
		/// Executes the command by calling the Execute event.</summary>
		/// <remarks>
		/// If <c>State.Enabled</c> is <c>true</c> then calling this method will
		/// have no effect.
		/// The execute event handlers of this command's ancestors are all called.</remarks>
		public void PerformExecute(object source) {
			// Check if this command can be executed.
			if (enabled == false)
				return;

			CommandEventArgs e = new CommandEventArgs(source);
			e.command = this;
			e.type = CommandEventArgs.EventType.Execute;
			e.commandPath.Clear();

			Command ancestor = this;
			while (ancestor != null) {
				e.commandPath.Insert(0, ancestor);
				ancestor = ancestor.Parent;
			}

			if (e.HandlerCommand.Filter != null)
				e.HandlerCommand.Filter(e.Command, e);
			else
				DefaultFilter(e.Command, e);
		}

		protected internal virtual void OnExecute(CommandEventArgs e) {

		}

		public CommandEventHandler Filter;

		internal void DoRouting(CommandEventArgs e) {
			if (Execute != null)
				e.executeHandled = true;
			if (Update != null)
				e.updateHandled = true;

			if (e.Type == CommandEventArgs.EventType.Execute) {
				if (Execute != null)
					Execute(e.Command, e);
			}
			else if (e.Type == CommandEventArgs.EventType.Update) {
				if (Update != null)
					Update(e.Command, e);
			}

			// Call the filter for the next in the chain.
			e.commandPath.RemoveAt(0);
			if (e.commandPath.Count == 0)
				return;

			if (e.HandlerCommand.Filter != null)
				e.HandlerCommand.Filter(e.Command, e);
			else
				DefaultFilter(e.Command, e);
		}

		private void DefaultFilter(Command command, CommandEventArgs e) {
			e.Route();
		}
	}


	/*
	Genghis.CommandCollection
	Genghis.CommandCollection.Enumerator

	These C# classes implement a strongly-typed collection of 
	Command objects.

	The internal representation is an array of Command, so the performance 
	characteristics will be more like a vector than a list, to use STL terminology.

	The implementation is optimized for value-types, as it goes to great length to 
	avoid the overhead of boxing and unboxing.  But it should also work well for 
	reference types.

	Mad props to Nick Wienholt <sheyenne@bigpond.com> for assisting me in 
	this research, and the testing, the benchmarking, and of course, the 
	implementation!

	Last but not least, a quick shout out to Kit George, for his generous 
	contribution to the dotnet mailing list -- a code-generator for 
	CollectionBase-derived classes:
		http://discuss.develop.com/archives/wa.exe?A2=ind0107C&L=DOTNET&P=R35911
	This was the original inspiration for the fine code you are now enjoying.

	- Shawn Van Ness, 28 Jan 2002
	*/

	[Serializable]
	public class CommandCollection : ICollection, IList, IEnumerable
	{
		// Construction

		public CommandCollection()
		{ }

		public CommandCollection(CommandCollection x)
		{ AddRange(x); }

		public CommandCollection(T[] x)
		{ AddRange(x); }

		// Operations (type-safe ICollection)

		public int Count
		{
			get
			{ return m_count; }
		}

		public void CopyTo(T[] array)
		{
			this.CopyTo(array, 0);
		}

		public void CopyTo(T[] array, int start)
		{
			if (m_count > array.GetUpperBound(0)+1-start)
				throw new System.ArgumentException("Destination array was not long enough.");
			for (int i=0; i < m_count; ++i)
				array[start+i] = m_array[i];
		}

		// Operations (type-safe IList)

		public T this[int i]
		{
			get
			{
				ValidateIndex(i); // throws
				return m_array[i]; 
			}
			set
			{
				ValidateIndex(i); // throws
				
				OnItemRemoved(m_array[i]);
				
				++m_version;
				m_array[i] = value;
				
				OnItemAdded(value);
			}
		}

		public int Add(T x)
		{
			++m_version;
			
			if (NeedsGrowth())
				Grow();

			m_array[m_count] = x;
			m_count ++;
			
			OnItemAdded(new T[] { x });
			
			return m_count - 1;
		}
		
		public void Clear()
		{
			if (ItemsRemoved != null) {
				T[] members = new T[m_count];
				CopyTo(members);
				OnItemRemoved(members);
			}
		
			++m_version;
			m_array = new T[16];
			m_count = 0;
		}

		public bool Contains(T x)
		{
			for (int i=0; i < m_count; ++i)
				if (m_array[i].Equals(x))
					return true;
			return false;
		}

		public int IndexOf(T x)
		{
			for (int i=0; i < m_count; ++i)
				if (m_array[i].Equals(x))
					return i;
			return -1;
		}

		public void Insert(int pos, T x)
		{
			ValidateIndex(pos,true); // throws
			++m_version;
			
			if (NeedsGrowth())
				Grow();

			for (int i=m_count; i > pos; --i)
				m_array[i] = m_array[i-1];

			m_array[pos] = x;
			m_count++;
			
			OnItemAdded(x);
		}

		public void Remove(T x)
		{			
			int i = IndexOf(x);
			if (i < 0)
				throw new System.ArgumentException("Cannot remove the specified item because it was not found in the specified Collection.");
			
			RemoveAt(i);
		}

		public void RemoveAt(int pos)
		{
			ValidateIndex(pos); // throws
			
			OnItemRemoved(m_array[pos]);
			
			++m_version;
			
			m_count--;
			for (int i=pos; i < m_count; ++i)
				m_array[i] = m_array[i+1];

			if (NeedsTrimming())
				Trim();
		}

		// Operations (type-safe IEnumerable)

		public Enumerator GetEnumerator()
		{
			return new Enumerator(this);
		}

		// Public helpers (just to mimic some nice features of ArrayList)

		public int Capacity
		{
			get
			{ return m_array.Length; }
			set
			{
				++m_version;

				if (value < m_count)
					value = m_count;

				if (value != Capacity)
				{
					T[] temp = new T[value];
					for (int i=0; i < m_count; ++i)
						temp[i] = m_array[i];
					m_array = temp;
				}
			}
		}

		public int AddRange(CommandCollection x)
		{
			++m_version;
			
			int r = m_count;
			for (int i=0; i < x.Count; ++i)
				Add(x[i]);
			
			if (ItemsAdded != null) {
				T[] members = new T[x.Count];
				x.CopyTo(members);
				OnItemAdded(members);
			}
			
			return r;
		}

		public int AddRange(T[] x)
		{
			++m_version;
			
			int r = m_count;
			for (int i=0; i < x.Length; ++i)
				Add(x[i]);
			
			OnItemAdded(x);
			
			return r;
		}

		// Implementation (helpers)

		private void ValidateIndex(int i)
		{
			ValidateIndex(i,false);
		}

		private void ValidateIndex(int i, bool allowEqualEnd)
		{
			int max = (allowEqualEnd)?(m_count):(m_count-1);
			if (i < 0 || i > max)
				throw new System.ArgumentOutOfRangeException("Index was out of range.  Must be non-negative and less than the size of the collection.", (object)i, "Specified argument was out of the range of valid values.");
		}

		private bool NeedsGrowth()
		{
			return (m_count >= Capacity);
		}

		private void Grow()
		{
			++m_version;
			
			if (NeedsGrowth())
			{
				T[] temp = new T[m_count*2];
				for (int i=0; i < m_count; ++i)
					temp[i] = m_array[i];
				m_array = temp;
			}
		}

		private bool NeedsTrimming()
		{
			return (m_count <= Capacity/3);
		}

		private void Trim()
		{
			++m_version;
			
			if (NeedsTrimming())
			{
				T[] temp = new T[m_count];
				for (int i=0; i < m_count; ++i)
					temp[i] = m_array[i];
				m_array = temp;
			}
		}
		
		// Events
		public event CommandCollectionEventHandler ItemsAdded;
		public event CommandCollectionEventHandler ItemsRemoved;
		
		// OnItemAdded is called after the member has been added.
		protected virtual void OnItemAdded(T x) {
			OnItemAdded(new T[] { x });
		}
		protected virtual void OnItemAdded(T[] x) {
			if (ItemsAdded != null)
				ItemsAdded(this, new CommandCollectionEventArgs(true, x));
		}
		
		// OnItemRemoved is called before the member has been removed.
		protected virtual void OnItemRemoved(T x) {
			OnItemRemoved(new T[] { x });
		}
		protected virtual void OnItemRemoved(T[] x) {
			if (ItemsRemoved != null)
				ItemsRemoved(this, new CommandCollectionEventArgs(false, x));
		}
		

		// Implementation (ICollection)

		int ICollection.Count
		{
			get
			{ return m_count; }
		}

		bool ICollection.IsSynchronized
		{
			get
			{ return m_array.IsSynchronized; }
		}

		object ICollection.SyncRoot
		{
			get
			{ return m_array.SyncRoot; }
		}

		void ICollection.CopyTo(Array array, int start)
		{
			this.CopyTo((T[])array, start);
		}

		// Implementation (IList)

		bool IList.IsFixedSize
		{
			get
			{ return false; }
		}

		bool IList.IsReadOnly
		{
			get
			{ return false; }
		}

		object IList.this[int i]
		{
			get
			{ return (object)this[i]; }
			set
			{ this[i] = (T)value; }
		}

		int IList.Add(object x)
		{
			return this.Add((T)x);
		}

		void IList.Clear()
		{
			this.Clear();
		}

		bool IList.Contains(object x)
		{
			return this.Contains((T)x);
		}

		int IList.IndexOf(object x)
		{
			return this.IndexOf((T)x);
		}

		void IList.Insert(int pos, object x)
		{
			this.Insert(pos, (T)x);
		}

		void IList.Remove(object x)
		{
			this.Remove((T)x);
		}

		void IList.RemoveAt(int pos)
		{
			this.RemoveAt(pos);
		}

		// Implementation (IEnumerable)

		IEnumerator IEnumerable.GetEnumerator()
		{
			return (IEnumerator)(this.GetEnumerator());
		}

		// Implementation (data)

		private T[] m_array = new T[16];
		private int m_count = 0;
		private int m_version = 0;

		// Nested enumerator class
		
		public class Enumerator : IEnumerator
		{
			// Construction
	
			public Enumerator(CommandCollection tc)
			{
				m_collection = tc;
				m_index = -1;
				m_version = tc.m_version;
			}
	
			// Operations (type-safe IEnumerator)
	
			public T Current
			{
				get
				{ return m_collection[m_index]; }
			}
	
			public bool MoveNext()
			{
				if (m_version != m_collection.m_version)
					throw new System.InvalidOperationException("Collection was modified; enumeration operation may not execute.");

				++m_index;
				return (m_index < m_collection.Count)?true:false;
			}
	
			public void Reset()
			{
				m_index = -1;
			}
	
			// Implementation (IEnumerator)
	
			object IEnumerator.Current
			{
				get
				{ return (object)(this.Current); }
			}
	
			bool IEnumerator.MoveNext()
			{
				return this.MoveNext();
			}
	
			void IEnumerator.Reset()
			{
				this.Reset();
			}
	
			// Implementation (data)
	
			private CommandCollection m_collection;
			private int m_index;
			private int m_version;
		}
	}
	
	public delegate void CommandCollectionEventHandler(object sender, CommandCollectionEventArgs e);
	
	public class CommandCollectionEventArgs : EventArgs
	{
		bool added;
		T[] items;
		
		public CommandCollectionEventArgs(bool added, T[] items) {
			this.added = added;
			this.items = items;
		}
		
		public T[] Items {
			get { return items; }
		}
		
		public bool Added {
			get { return added; }
		}
	}
}

#region To Do
// Allow the command description to be displayed elsewhere.
// Look into implementing CloneMenu().
// Quiz: What happens if the statusBar is disposed while menu is open?
#endregion

namespace Genghis.Windows.Forms {

#region Command-driven menu implementation

	/// <summary>
	/// Implements a main menu that reads the state of the menu items
	/// from a command set.</summary>
	public class CommandMainMenu : MainMenu {
		bool isMenuOpen;
		StatusBar statusBar;
		bool backupShowPanels;
		string backupText;

		/// <summary>
		/// Initializes an empty CommandMainMenu.</summary>
		public CommandMainMenu() {
			isMenuOpen = false;
		}

		/// <summary>
		/// Gets or sets the status bar that command descriptions will be displayed in.</summary>
		/// <value>
		/// The status bar that command descriptions will be displayed in.</value>
		public StatusBar StatusBar {
			get { return statusBar; }
			set { statusBar = value; }
		}
		
		/// <summary>
		/// Called when any menu item in the menu is clicked by the user.</summary>
		/// <param name="sender">The sender of the event.</param>
		/// <param name="e">The EventArgs that contains the event data.</param>
		internal void OnMenuOpen(object sender, EventArgs e) {
			if (isMenuOpen == true)
				return;
			isMenuOpen = true;
			
			Form form = GetForm();
			if (form == null)
				return;		// Without the form, we don't know when the menu closes.
			form.MenuComplete += new EventHandler(OnMenuClose);

			// Backup the state of the status bar.
			if (StatusBar != null) {
				backupShowPanels = StatusBar.ShowPanels;
				backupText = StatusBar.Text;
			}
		}

		/// <summary>
		/// Called when the user dismisses the menu.</summary>
		/// <param name="sender">The sender of the event.</param>
		/// <param name="e">The EventArgs that contains the event data.</param>
		private void OnMenuClose(object sender, EventArgs e) {
			Form form = GetForm();
			if (form == null)
				return;
		
			// Remove this event handler.
			form.MenuComplete -= new EventHandler(OnMenuClose);

			// Restore the state of the status bar.
			if (StatusBar != null) {
				StatusBar.ShowPanels = backupShowPanels;
				StatusBar.Text = backupText;
			}
			
			// Menu is no longer open.
			isMenuOpen = false;
		}
	}

	/// <summary>
	/// Implements a menu item that reads it's state from a command.</summary>
	public class CommandMenuItem : MenuItem {
		Command command;

		/// <summary>
		/// Initializes a CommandMenuItem with a blank caption.</summary>
		public CommandMenuItem() {
		}

		/// <summary>
		/// Initializes a CommandMenuItem with a command.</summary>
		/// <param name="command">The command that this menu item represents.</param>
		public CommandMenuItem(Command command) {
			Command = command;
		}
		
		/// <summary>
		/// Gets the command that this menu item represents.</summary>
		/// <value>
		/// The command that this menu item represents.</value>
		public Command Command {
			get { return command; }
			set {
				command = value;
			
				// Get the command properties that don't change.
				if (command != null) 
				{
					Text = command.DefaultText;
					Shortcut = command.Shortcut;
					RadioCheck = command.IsRadioMember;
				}
			}
		}
		
		/// <summary>
		/// This member overrides
		/// <paramref name="MenuItem.OnPopup"/>.</summary>
		protected override void OnPopup(EventArgs e) {
			// Indicate to the main menu that the menu is open.
			CommandMainMenu mainMenu = GetMainMenu() as CommandMainMenu;
			if (mainMenu != null)
				mainMenu.OnMenuOpen(this, EventArgs.Empty);
		
			// Delegate popup event to the Popup event handlers.
			base.OnPopup(e);

			foreach (MenuItem menuItem in MenuItems) {
				if (menuItem is CommandMenuItem) {
					// Update the child menu items.
					if (((CommandMenuItem) menuItem).Command != null)
						((CommandMenuItem) menuItem).OnCommandChanged(EventArgs.Empty);
				}
			}
		}

		/// <summary>
		/// This member overrides
		/// <paramref name="MenuItem.OnSelect"/>.</summary>
		protected override void OnSelect(EventArgs e) {
			// Delegate select event to the Select event handlers.
			base.OnSelect(e);

			CommandMainMenu mainMenu = GetMainMenu() as CommandMainMenu;
			if (mainMenu != null && mainMenu.StatusBar != null && command != null) {
				// Show the command description in the status bar.
				mainMenu.StatusBar.Text = command.Description;
			}
		}

		/// <summary>
		/// This member overrides
		/// <paramref name="MenuItem.OnClick"/>.</summary>
		protected override void OnClick(EventArgs e) {
			// Delegate click event to the Click event handlers.
			base.OnClick(e);

			// Execute command.
			if (command != null)
				command.PerformExecute(this);
		}
		
		/// <summary>
		/// Occurs when the state of the command that this menu item represents changes.</summary>
		/// <param name="e">An EventArgs that contains the event data.</param>
		protected virtual void OnCommandChanged(EventArgs e) {
			// Refreshes menu item state from the command.
			Text = command.Text;
			Enabled = command.Enabled;
			Checked = command.Selected;
		}
	}

#endregion

#region Command-driven tool bar implementation

	/// <summary>
	/// Implements a tool bar that reads it's state from a list of commands.</summary>
	public class CommandToolBar : ToolBar {
		Timer updateTimer;

		/// <summary>
		/// Initializes an empty CommandToolBar.</summary>
		public CommandToolBar() {
			// Start a timer that will update the toolbar.
			updateTimer = new Timer();
			updateTimer.Interval = 200;
			updateTimer.Tick += new EventHandler(OnTick);
			updateTimer.Start();
		}
		
		/// <summary>
		/// This member overrides
		/// <paramref name="System.ComponentModel.Component.Dispose"/>.</summary>
		protected override void Dispose(bool disposing) {
			if (disposing) {
				if (updateTimer != null) {
					updateTimer.Dispose();
					updateTimer = null;
				}
			}
		}

		/// <summary>
		/// Called regularly to update the state of the toolbar commands.</summary>
		private void OnTick(object sender, EventArgs e) {
			// For each button in the toolbar...
			foreach (ToolBarButton button in Buttons) {
				// If the button has a command attached, update the command.
				CommandToolBarButton commandButton = button as CommandToolBarButton;
				if (commandButton != null && commandButton.Command != null)
					commandButton.Command.PerformUpdate();
			}
		}
		
		/// <summary>
		/// This member overrides
		/// <paramref name="ToolBar.OnButtonClick"/>.</summary>
		protected override void OnButtonClick(ToolBarButtonClickEventArgs e) {
			// Call event handlers.
			base.OnButtonClick(e);
		
			// If the button has a command attached, execute the command.
			CommandToolBarButton button = e.Button as CommandToolBarButton;
			if (button != null && button.Command != null)
				button.Command.PerformExecute(button);
		}
		
		// HACK: This was the only way I could find of determining when buttons
		// are added to the toolbar.
		
		/// <summary>
		/// This member overrides
		/// <paramref name="ToolBar.OnResize"/>.</summary>
		protected override void OnResize(EventArgs e) {
			if (Buttons == null)
				return;

			foreach (ToolBarButton button in Buttons) {
				CommandToolBarButton commandButton = button as CommandToolBarButton;
				if (commandButton == null || commandButton.Command == null)
					continue;
				if (commandButton.Command.Image == null || commandButton.ImageIndex != -1)
					continue;

				if (ImageList == null) {
					ImageList = new ImageList();
					ImageList.ColorDepth = ColorDepth.Depth32Bit;
				}

				int index = ImageList.Images.Count;
				ImageList.Images.Add(commandButton.Command.Image);

				button.ImageIndex = index;
			}
		}
	}

	/// <summary>
	/// Implements a tool bar button that reads it's state from a command.</summary>
	public class CommandToolBarButton : ToolBarButton {
		Command command;

		/// <summary>
		/// Initializes a CommandToolBarButton with a command.</summary>
		/// <param name="command">The command that this tool bar button represents.</param>
		public CommandToolBarButton(Command command) {
			this.command = command;
			if (command == null)
				return;

			// If the command is a toggle command or is a part of a radio group,
			// then make this button a toggle button.
			if (command.Type == Genghis.Command.CommandType.Toggle || command.IsRadioMember)
				Style = ToolBarButtonStyle.ToggleButton;
			else
				Style = ToolBarButtonStyle.PushButton;

			// Receive changed events from the command.
			command.Changed += new CommandEventHandler(OnChanged);
			OnChanged(command, null);	// Fake a changed event.
		}

		/// <summary>
		/// Gets the command that this menu item represents.</summary>
		/// <value>
		/// The command that this menu item represents.</value>
		public Command Command {
			get { return command; }
			set { command = value; }
		}

		/// <summary>
		/// Called when the state of the command that this tool bar button represents changes.</summary>
		/// <param name="command">The command that changed.</param>
		/// <param name="e">A CommandEventArgs that contains the event data.</param>
		private void OnChanged(Command command, CommandEventArgs e) {
			// The command state has changed; update the button.
			ToolTipText = command.Text.EndsWith("...") ? command.Text.Substring(0, command.Text.Length - 3) : command.Text;
			Enabled = command.Enabled;
			PartialPush = command.Selected3 == Genghis.Command.SelectedState.Indeterminate;
			Pushed = command.Selected3 == Genghis.Command.SelectedState.Selected;
		}
	}
#endregion
}