using System;
using System.Collections;
using System.IO;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using System.Windows.Forms;
using System.Reflection;
using Genghis;

namespace Genghis.DocView {

	public class Document : ICommandTarget {
		Uri savedPath;
		string title;

		ViewCollection views;
		View locker;
		
		Hashtable executeHandlers;
		Hashtable updateHandlers;
		
		bool modified;		// Has the document been modified since the last save?
		
		static ArrayList registeredDocumentTypes;

		/// <summary>
		/// Constructs an empty document.</summary>
		public Document() {
			savedPath = null;
			Title = "Untitled";

			views = new ViewCollection();
			views.ItemsAdded += new ViewCollectionEventHandler(OnViewsChanged);
			views.ItemsRemoved += new ViewCollectionEventHandler(OnViewsChanged);
			
			executeHandlers = Command.GetExecuteHandlers(this);
			updateHandlers = Command.GetUpdateHandlers(this);
		}
		
		// Used to hook up command event handlers.
		Hashtable ICommandTarget.ExecuteHandlers {
			get { return executeHandlers; }
		}
		Hashtable ICommandTarget.UpdateHandlers {
			get { return updateHandlers; }
		}

		/// <summary>
		/// The fully qualifed path of the document, or a null reference
		/// (Nothing in Visual Basic) if the document has not yet been
		/// saved.</summary>
		public Uri SavedPath {
			get { return savedPath; }
			set {
				if (savedPath != value) {
					savedPath = value;
					OnSavedPathChanged(EventArgs.Empty);
				}
			}
		}

		public event EventHandler SavedPathChanged;
		protected void OnSavedPathChanged(EventArgs e) {
			if (SavedPathChanged != null)
				SavedPathChanged(this, e);
		}

		/// <summary>
		/// The document title, as displayed in the title bar of the window
		/// hosting the document.</summary>
		public string Title {
			get { return title; }
			set {
				if (title != value) {
					title = value;
					OnTitleChanged(EventArgs.Empty);
				}
			}
		}

		public event EventHandler TitleChanged;
		protected void OnTitleChanged(EventArgs e) {
			if (TitleChanged != null)
				TitleChanged(this, e);
		}

		/// <summary>
		/// Gets or sets a boolean indicating whether the document has been
		/// modified.</summary>
		/// <value>
		/// A boolean indicating whether the document has been modified.</value>
		/// <remarks>
		/// <c>True</c> if the document has been modified since it was last
		/// saved (or, if it has not yet been saved, since it was created).
		/// <c>False</c> otherwise.</remarks>
		public bool Modified {
			get { return modified; }
			set { modified = value; OnModifiedChanged(EventArgs.Empty); }
		}
		
		public event EventHandler ModifiedChanged;
		protected void OnModifiedChanged(EventArgs e) {
			if (ModifiedChanged != null)
				ModifiedChanged(this, e);
		}

		// TODO: Fix this!
		public virtual string FileDialogFilter {
			get { return "All files (*.*)|*.*"; }
		}

		// Views

		/// <summary>
		/// The views currently attached to this document.</summary>
		public ViewCollection Views {
			get { return views; }
		}

		/// <summary>
		/// A view was added or removed from the Views collection.</summary>
		protected virtual void OnViewsChanged(object sender, ViewCollectionEventArgs e) {
			// Update document of each view.
			if (e.Added) {
				foreach (View view in e.Items)
					view.SetDocument(this);
			}
			else {
				foreach (View view in e.Items)
					view.SetDocument(null);
			}
		}
		
		public void Lock(View view) {
			locker = view;
		}
		
		public void Unlock() {
			foreach (View view in views) {
				if (view != locker)
					view.OnDocumentChanged();
			}
			locker = null;
		}



		// --------------------------------------------------------------------
		//
		//   DOCUMENT LOADING
		//
		// --------------------------------------------------------------------


		/// <summary>
		/// Registers a document type so it becomes a candidate for loading.</summary>
		static public void RegisterDocumentType(Type documentType) {
			if (documentType.IsSubclassOf(typeof(Document)) == false)
				throw new ArgumentException("documentType must be a subclass of Document.", "documentType");

			if (registeredDocumentTypes == null)
				registeredDocumentTypes = new ArrayList();
			registeredDocumentTypes.Add(documentType);
		}

		static public Document Load(Stream stream) {
			return Load(new StreamInfo(stream));
		}

		/// <summary>
		/// Deserializes a document from a stream.</summary>
		/// <remarks>
		/// If there is a registered document type with a constructor of the
		/// form Constructor(StreamInfo), this method will invoke it to deserialize
		/// the document; otherwise, it will use the BinaryFormatter class.</remarks>
		static public Document Load(StreamInfo streamInfo) {
			Document loadedDocument = null;

			// Open the file as a stream.
			using (streamInfo.Stream) {

				bool handled = false;

				// Calls a static method of each document to see if they will load the stream.
				// Searches for a derived class with a constructor of the forms:
				// 1. Constructor(StreamInfo streamInfo).
				// 2. Constructor(Stream stream).
				foreach (Type documentType in registeredDocumentTypes) {
					ConstructorInfo loadConstructor = null;
					
					loadConstructor = documentType.GetConstructor(new Type[] { typeof(StreamInfo) });
					if (loadConstructor != null) {
						handled = true;
						// Invoke the derived document's construtor.
						loadedDocument = (Document) loadConstructor.Invoke(new object[] { streamInfo });
						break;
					}

					loadConstructor = documentType.GetConstructor(new Type[] { typeof(Stream) });
					if (loadConstructor != null) {
						handled = true;
						// Invoke the derived document's construtor.
						loadedDocument = (Document) loadConstructor.Invoke(new object[] { streamInfo.Stream });
						break;
					}
				}

				// If no document handled it, attempt to use serialization to
				// load the document.
				if (handled == false)
					loadedDocument = (Document) new BinaryFormatter().Deserialize(streamInfo.Stream);
			}

			if (loadedDocument == null)
				return null;

			loadedDocument.SavedPath = streamInfo.Uri;
			loadedDocument.Title = streamInfo.NameWithoutExtension;
			loadedDocument.Modified = false;

			return loadedDocument;
		}




		// --------------------------------------------------------------------
		//
		//   DOCUMENT SAVING
		//
		// --------------------------------------------------------------------


		// File menu.

		/// <summary>
		/// Handle the Save command.</summary>
		[Command.ExecuteHandler("File.Save")]
		public void OnFileSave(Command command, CommandEventArgs e) {
			DoSave(false, false);
		}
		
		[Command.UpdateHandler("File.Save")]
		public void UpdateFileSave(Command command, CommandEventArgs e) {
			command.Enabled = Modified;
		}

		/// <summary>
		/// Handle the Save As command.</summary>
		[Command.ExecuteHandler("File.SaveAs")]
		public void OnFileSaveAs(Command command, CommandEventArgs e) {
			DoSave(true, false);
		}


		/// <summary>
		/// Opens a file dialog box, serializes to disk and changes the
		/// document's saved path.</summary>
		/// <param name="saveAs">
		/// <c>true</c> if the document should be saved under another name.
		/// </param>
		/// <param name="saveCopy">
		/// <c>true</c> if a copy of the document should be saved under another
		/// name.</param>
		protected void DoSave(bool saveAs, bool saveCopy) {
			if (saveCopy == true)
				saveAs = true;

			StreamInfo streamInfo = null;

			if (saveAs || SavedPath == null) {
				// We're either doing a Save As or Save Copy As, or a Save and
				// the document hasn't been saved yet.
			
				// Show a save file dialog box.
				SaveFileDialog saveFileDialog = new SaveFileDialog();
				if (SavedPath != null && SavedPath.LocalPath != null) {
					saveFileDialog.InitialDirectory = Path.GetDirectoryName(SavedPath.LocalPath);
					saveFileDialog.FileName = SavedPath.LocalPath;
				}
				saveFileDialog.Filter = FileDialogFilter;
				if (saveCopy)
					saveFileDialog.Title = "Save Copy As";
				if (saveFileDialog.ShowDialog() == DialogResult.Cancel)
					return;
				
				string fileName = null;
				try {
					fileName = saveFileDialog.FileName;
				}
				catch (System.Security.SecurityException) {
				}
				
				// Check if the fileName is valid.
				if (fileName != null)
					streamInfo = new FileStreamInfo(saveFileDialog.OpenFile(), fileName);
				else
					streamInfo = new StreamInfo(saveFileDialog.OpenFile());
			}
			else {
				// We're doing a Save.
				// SavedPath != null.
				
				// Open the document file for writing.
				streamInfo = new FileStreamInfo(
					new FileStream(SavedPath.LocalPath, FileMode.Create, FileAccess.Write),
					SavedPath.LocalPath);
			}

			if (streamInfo == null || streamInfo.Stream == null)
				return;

			try {
				Save(streamInfo);
			}
			catch (System.Exception exception) {
				// Display any errors thrown by the Save() method.
				MessageBox.Show(exception.Message);
			}

			// Close the output stream.
			streamInfo.Stream.Close();

			if (saveCopy == false) {
				SavedPath = streamInfo.Uri;
				if (SavedPath == null)
					Title = "Untitled";
				else
					Title = streamInfo.NameWithoutExtension;
			}
			
			// Not modified.
			Modified = false;
		}

		/// <summary>
		/// Serializes the document.</summary>
		/// <notes>
		/// Serializes the document using the BinaryFormatter class.  The
		/// class should be marked [Serializable] for this to work.</notes>
		virtual public void Save(StreamInfo streamInfo) {
			Save(streamInfo.Stream);
		}
		
		/// <summary>
		/// Serializes the document.</summary>
		/// <notes>
		/// Serializes the document using the BinaryFormatter class.  The
		/// class should be marked [Serializable] for this to work.</notes>
		virtual public void Save(Stream stream) {
			new BinaryFormatter().Serialize(stream, this);
		}
	}

}

