using System;
using Microsoft.Win32;
using System.Runtime.InteropServices;

namespace Genghis {

	public class FileAssociation {
		string progID;
		RegistryKey registryKey;

		public FileAssociation(string progID) {
			ValidateRegistryName(progID, "progID");

			this.progID = progID;
		}

		public string ProgID {
			get { return progID; }
		}

		public string FriendlyName {
			get { return ReadString(null, null); }
			set { WriteString(null, null, value); }
		}

		// Eg. %SystemRoot%\shell32.dll,-1
		public string IconPath {
			get { return ReadString("DefaultIcon", null); }
			set { WriteString("DefaultIcon", null, value); }
		}

		// Eg. @%SystemRoot%\shell32.dll,-154
		public string FriendlyTypeNamePath {
			get { return ReadString(null, "FriendlyTypeName"); }
			set { WriteString(null, "FriendlyTypeName", value); }
		}

		public string LatestProgID {
			get { return ReadString("CurVer", null); }
			set { WriteString("CurVer", null, value); }
		}

		public string EditFlags {
			get { return ReadString(null, "EditFlags"); }
			set { WriteString(null, "EditFlags", value); }
		}

		// Eg. @%SystemRoot%\shell32.dll,-54
		public string InfoTipPath {
			get { return ReadString(null, "InfoTip"); }
			set { WriteString(null, "InfoTip", value); }
		}

		public void SetVerb(string verb, string exePath) {
			WriteString("shell\\" + verb + "\\command", null, exePath);
		}



		public bool IsAssociated(string extension) {
			ValidateExtension(extension);
			string associatedProgID = ReadString(Registry.ClassesRoot, extension, null);
			return associatedProgID == progID;
		}

		public void Associate(string extension) {
			Associate(extension, PerceivedType.Unknown);
		}

		//Shell32.dll
		//SHSTDAPI_(void) SHChangeNotify(LONG wEventId, UINT uFlags, LPCVOID dwItem1, LPCVOID dwItem2);
		[DllImport("shell32.dll")]
		private static extern int SHChangeNotify(uint eventID, uint flags, IntPtr item1, IntPtr item2);

		public void Associate(string extension, PerceivedType perceivedType) {
			ValidateExtension(extension);

			WriteString(Registry.ClassesRoot, extension, null, progID);
			if (perceivedType != PerceivedType.Unknown)
				WriteString(Registry.ClassesRoot, extension, "PerceivedType", perceivedType.ToString().ToLower());

			// SHChangeNotify(LONG wEventId, UINT uFlags, LPCVOID dwItem1, LPCVOID dwItem2);
			// SHChangeNotify(SHCNE_ASSOCCHANGED, SHCNF_IDLIST, NULL, NULL);
			uint SHCNE_ASSOCCHANGED = 0x08000000;
			uint SHCNF_IDLIST = 0x0000;
			SHChangeNotify(SHCNE_ASSOCCHANGED, SHCNF_IDLIST, IntPtr.Zero, IntPtr.Zero);
		}



		// PRIVATE MEMBERS

		string ReadString(string path, string valueName) {
			return ReadString(registryKey, progID + "\\" + path, valueName);
		}

		string ReadString(RegistryKey baseKey, string path, string valueName) {
			RegistryKey key = null;
			if (path == null)
				key = baseKey;
			else
				key = Registry.ClassesRoot.OpenSubKey(path);
			if (key == null)
				return null;

			object data = null;
			try {
				data = key.GetValue(valueName);
			}
			finally {
				if (path == null)
					key.Close();
			}

			if (data is string)
				return (string) data;

			return null;
		}

		void WriteString(string path, string valueName, string valueString) {
			if (registryKey == null)
				registryKey = Registry.ClassesRoot.CreateSubKey(progID);
			WriteString(registryKey, progID + "\\" + path, valueName, valueString);
		}

		void WriteString(RegistryKey baseKey, string path, string valueName, string valueString) {
			RegistryKey key = null;
			if (path == null)
				key = baseKey;
			else
				key = Registry.ClassesRoot.CreateSubKey(path);
			if (key == null)
				return;

			try {
				key.SetValue(valueName, valueString);
			}
			finally {
				if (path == null)
					key.Close();
			}
		}

		void ValidateRegistryName(string argument, string argumentName) {
			int invalidCharacterIndex = argument.IndexOfAny("\\/:*?\"<>|".ToCharArray());
			if (invalidCharacterIndex != -1)
				throw new ArgumentException(argumentName + " contains an invalid character: '" + argument[invalidCharacterIndex] + "'.", argumentName);
		}

		void ValidateExtension(string extension) {
			ValidateRegistryName(extension, "extension");
			if (extension.StartsWith(".") == false)
				throw new ArgumentException("The extension must begin with a period.", "extension");
			if (extension.IndexOf('.', 1) != -1)
				throw new ArgumentException("The extension cannot contain more than one period.", "extension");
		}
	}

	/// <summary>
	/// Specifies broad categories of file format types.</summary>
	/// <remarks>
	/// <para>Perceived types are similar to file types except they refer to
	/// broad categories of file format types, rather than specific file types.
	/// File types should define a perceived type when one of the defined
	/// categories is appropriate.</para>
	/// <para>Perceived types were introduced in Microsoft Windows XP.</para>
	/// </remarks>
	public enum PerceivedType {
		/// <summary>The perceived type is unknown, or none are suited.</summary>
		Unknown,
		/// <summary>A type of image eg. GIF, PNG, JPEG, TIFF, etc.</summary>
		Image,
		/// <summary>A type of text file eg. TXT, DOC, XML, etc.</summary>
		Text,
		/// <summary>A type of audio file eg. MP3, WAV, WMA, etc.</summary>
		Audio,
		/// <summary>A type of video file eg. AVI, MPG, WMV, etc.</summary>
		Video,
		/// <summary>A compressed archive eg. ZIP, RAR, ACE, etc.</summary>
		Compressed,
		/// <summary>A system file eg. DLL, DRV, SYS, etc.</summary>
		System,
	}

	[Flags]
	public enum EditFlags {
		/// <summary>Excludes the file class.</summary>
		Exclude = 0x00000001,

		/// <summary>Shows file classes, such as folders, that aren't associated with a file name extension.</summary>
		Show = 0x00000002,

		/// <summary>Indicates that the file class has a file name extension.</summary>
		HasExtension = 0x00000004,

		/// <summary>Prohibits editing of the registry entries associated with this file class, the addition of new entries, and the deletion or modification of existing entries.</summary>
		NoEdit = 0x00000008,

		/// <summary>Prohibits deletion of the registry entries associated with this file class.</summary>
		NoRemove = 0x00000010,

		/// <summary>Prohibits the addition of new verbs to the file class.</summary>
		NoNewVerb = 0x00000020,

		/// <summary>Prohibits the modification or deletion of canonical verbs such as open and print.</summary>
		NoEditVerb = 0x00000040,

		/// <summary>Prohibits the deletion of canonical verbs such as open and print.</summary>
		NoRemoveVerb = 0x00000080,

		/// <summary>Prohibits the modification or deletion of the description of the file class.</summary>
		NoEditDesc = 0x00000100,

		/// <summary>Prohibits the modification or deletion of the icon assigned to the file class.</summary>
		NoEditIcon = 0x00000200,

		/// <summary>Prohibits the modification of the default verb.</summary>
		NoEditDflt = 0x00000400,

		/// <summary>Prohibits the modification of the commands associated with verbs.</summary>
		NoEditVerbCmd = 0x00000800,

		/// <summary>Prohibits the modification or deletion of verbs.</summary>
		NoEditVerbExe = 0x00001000,

		/// <summary>Prohibits the modification or deletion of the DDE-related entries.</summary>
		NoDDE = 0x00002000,

		/// <summary>Prohibits the modification or deletion of the content type and default extension entries.</summary>
		NoEditMIME = 0x00008000,

		/// <summary>Indicates that the file class's open verb can be safely invoked for downloaded files.</summary>
		OpenIsSafe = 0x00010000,

		/// <summary>Prevents the "Never ask me" check box from being enabled. The user can override this attribute through the File Type dialog box.</summary>
		AlwaysUnsafe = 0x00020000,

		/// <summary>Causes the file class's file name extension is always shown, even if the user has selected the "Hide Extensions" option.</summary>
		AlwaysShowExt = 0x00040000,

		/// <summary>Prohibits the addition of members of this file class to the Recent Documents folder.</summary>
		NoRecentDocs = 0x00100000,
	}

}