using System;
using System.Collections;
using System.Diagnostics;
using System.Windows.Forms;
using System.Drawing;
using System.IO;
using System.Xml;
using System.Reflection;

using Genghis;
using Genghis.DocView;
using Genghis.Windows.Forms;

class SimplePadApp : Form {
	WindowSerializer windowSerializer;
	CommandToolBar toolBar;
	StatusBar statusBar;
	Command commands;
	Splitter splitter = new Splitter();
	
	ArrayList documents;
	Genghis.DocView.View activeView;

	[STAThread]
	public static void Main() {
		TextWriterTraceListener consoleWriter = new TextWriterTraceListener(System.Console.Error);
    	Trace.Listeners.Add(consoleWriter);
		//StreamInfoUnitTests.Main();
		//Preferences.Test();
		
		Application.Run(new SimplePadApp());
	}

	public SimplePadApp() {
		windowSerializer = new WindowSerializer(this);
	
		//string exePath = Path.Combine(Environment.CurrentDirectory, Path.GetFileName(Environment.GetCommandLineArgs()[0]));
		string exePath = Process.GetCurrentProcess().MainModule.FileName;
		string exeDir = Path.Combine(Path.GetPathRoot(exePath), Path.GetDirectoryName(exePath));
		Environment.CurrentDirectory = exeDir;
		
		documents = new ArrayList();
		Document.RegisterDocumentType(typeof(SimplePadDocument));
		
		this.Size = new Size(700, 500);
		this.Text = "SimplePad";
		// Load form icon as a managed resource with the name "SimplePad.ico".
		this.Icon = new Icon(this.GetType().Assembly.GetManifestResourceStream("SimplePad.res.SimplePad.ico"));
		this.IsMdiContainer = true;
		this.MdiChildActivate += new EventHandler(OnMdiChildActivate);

		// Load commands from a managed resource with the name "CommandInfo.xml".
		commands = Command.Load(this.GetType().Assembly, "SimplePad.res.CommandInfo.xml");
		commands.RegisterHandlers(this);
		commands.Filter += new CommandEventHandler(FilterCommandEvents);
		
		statusBar = new StatusBar();
		this.Controls.Add(statusBar);
		
		CoolMainMenu mainMenu = LoadMenu(commands, this.GetType().Assembly, "SimplePad.res.CommandInfo.xml");
		mainMenu.StatusBar = statusBar;
		//mainMenu.MenuItems[3].MenuItems[2].MdiList = true;
		this.Menu = mainMenu;

		toolBar = LoadToolBar(commands, this.GetType().Assembly, "SimplePad.res.CommandInfo.xml");
		this.Controls.Add(toolBar);
		
		FileAssociation fileAssociation = new FileAssociation("SimplePad.Document.1");
		//if (fileAssociation.IsAssociated(".spd") == false) {
			fileAssociation.FriendlyName = "SimplePad document";
			fileAssociation.IconPath = Path.Combine(exeDir, "SimplePadDoc.ico");
			fileAssociation.SetVerb("open", "\"" + exePath + "\" \"%1\"");
			fileAssociation.Associate(".spd");
		//}
		
		if (Environment.GetCommandLineArgs().Length > 1) {
			string path = Environment.GetCommandLineArgs()[1];
			OnFileOpen(path);
		}
	}
	
	
	// MENU LOADING
	
	CoolMainMenu LoadMenu(Command commands, Assembly assembly, string name) {
		Stream stream = assembly.GetManifestResourceStream(name);
	
		XmlTextReader reader = new XmlTextReader(stream);
		CoolMainMenu root = null;

		while (reader.Read()) {
			if (reader.NodeType == XmlNodeType.Element) {
				if (reader.Name == "mainMenu") {
					string commandID = reader.GetAttribute("command");
					root = new CoolMainMenu();
					//root.VisualStyle = CoolMenuVisualStyle.WindowsMovieMaker;
					
					int initialDepth = reader.Depth;
					while (reader.Read() && reader.Depth != initialDepth) {
						if (reader.Name == "menuItem")
							root.MenuItems.Add(LoadMenuItem(commands, reader));
					}
				}
			}
		}

		reader.Close();
		return root;
	}
	
	CoolMenuItem LoadMenuItem(Command commands, XmlTextReader reader) {
		Debug.Assert(reader.Name == "menuItem");
		Debug.Assert(reader.NodeType == XmlNodeType.Element);

		string commandID = reader.GetAttribute("command");
		CoolMenuItem menuItem = new CoolMenuItem(commands.GetCommandFromID(commandID));

		if (reader.IsEmptyElement)
			return menuItem;		// No child commands.

		int initialDepth = reader.Depth;
		while (reader.Read() && reader.Depth != initialDepth) {
			if (reader.Name == "menuItem")
				menuItem.MenuItems.Add(LoadMenuItem(commands, reader));
			else if (reader.Name == "seperator") {
				CoolMenuItem separator = new CoolMenuItem("-");
				menuItem.MenuItems.Add(separator);
			}
		}
		
		return menuItem;
	}
	
	// TOOLBAR LOADING
		
	CommandToolBar LoadToolBar(Command commands, Assembly assembly, string name) {
		Stream stream = assembly.GetManifestResourceStream(name);
	
		XmlTextReader reader = new XmlTextReader(stream);
		CommandToolBar toolBar = null;

		while (reader.Read()) {
			if (reader.NodeType == XmlNodeType.Element) {
				if (reader.Name == "toolBar") {
					toolBar = new CommandToolBar();
					toolBar.Appearance = ToolBarAppearance.Flat;

					int initialDepth = reader.Depth;
					while (reader.Read() && reader.Depth != initialDepth) {
						if (reader.Name == "button") {
							Command command = commands.GetCommandFromID(reader["command"]);
							CommandToolBarButton button = new CommandToolBarButton(command);
							toolBar.Buttons.Add(button);
						}
						else if (reader.Name == "seperator") {
							ToolBarButton separator = new ToolBarButton();
							separator.Style = ToolBarButtonStyle.Separator;
							toolBar.Buttons.Add(separator);
						}
					}
				}
			}
		}

		reader.Close();
		return toolBar;
	}
	
	
	// Document / View architecture.
	
	public Document ActiveDocument {
		get {
			if (activeView != null)
				return activeView.Document;
			return null;
		}
	}
	
	public Genghis.DocView.View ActiveView {
		get { return activeView; }
	}
	
	
	
	// Handle when a view gets the focus.
	void OnMdiChildActivate(object sender, EventArgs e) {
		if (this.ActiveMdiChild == null) {
			// No active view.
			activeView = null;
		}
		else if (this.ActiveMdiChild is ViewContainer) {
			// Set the new active view.
			ViewContainer childForm = (ViewContainer) this.ActiveMdiChild;
			if (childForm.Views.Count > 0)
				activeView = childForm.Views[0];
		}
	}
	
	void FilterCommandEvents(Command command, CommandEventArgs e) {
		// NOTE: the active document/view could change during the course of an event.
		ICommandTarget[] targets = new ICommandTarget[] {
			ActiveDocument, ActiveView, ActiveMdiChild as ViewContainer
		};

		foreach (ICommandTarget target in targets) {
			if (target != null) {
				command.Execute += (CommandEventHandler) target.ExecuteHandlers[e.Command.ID];
				command.Update += (CommandEventHandler) target.UpdateHandlers[e.Command.ID];
			}
		}

		// Route to child commands.
		e.Route();
		
		foreach (ICommandTarget target in targets) {
			if (target != null) {
				command.Execute -= (CommandEventHandler) target.ExecuteHandlers[e.Command.ID];
				command.Update -= (CommandEventHandler) target.UpdateHandlers[e.Command.ID];
			}
		}
	}
	

	// File menu.

	// New.
	[Command.ExecuteHandler("File.New")]
	public void OnFileNew(Command command, CommandEventArgs e) {
		// Create a new document.
		Document document = new SimplePadDocument();
		documents.Add(document);

		// Create a view for that document.
		Genghis.DocView.View view = new SimplePadView();
		document.Views.Add(view);
		
		// Create a window to host the view.
		ViewContainer childForm = new ViewContainer();
		childForm.MdiParent = this;
		childForm.Views.Add(view);
		childForm.Show();
	}

	// Open.
	[Command.ExecuteHandler("File.Open")]
	public void OnFileOpen(Command command, CommandEventArgs e) {
		OpenFileDialog openFileDialog = new OpenFileDialog();
		openFileDialog.Filter = "SimplePad documents (*.spd)|*.spd|All files (*.*)|*.*";
		if (openFileDialog.ShowDialog() == DialogResult.Cancel)
			return;
			
		OnFileOpen(openFileDialog.FileName);
	}
	
	public void OnFileOpen(string path) {
		// Load the document.
		Document document = null;

		try {
			document = Document.Load(new FileStreamInfo(File.OpenRead(path), path));
		}
		catch (System.Exception exception) {
			MessageBox.Show(exception.Message);
			return;
		}
		
		// Generic error message.
		if (document == null) {
			MessageBox.Show("The file could not be opened.");
			return;
		}
		
		// Create a view for that document.
		Genghis.DocView.View view = new SimplePadView();
		document.Views.Add(view);

		// Create a window to host the view.
		ViewContainer childForm = new ViewContainer();
		childForm.MdiParent = this;
		childForm.Views.Add(view);
		childForm.Show();
	}

	// Exit.
	[Command.ExecuteHandler("File.Exit")]
	public void OnFileExit(Command command, CommandEventArgs e) {
		Application.Exit();
	}

	// View menu.

	// View tool bar.
	[Command.UpdateHandler("View.ToolBar")]
	public void UpdateViewToolBar(Command command, CommandEventArgs e) {
		command.Selected = toolBar.Visible;
	}

	[Command.ExecuteHandler("View.ToolBar")]
	public void OnViewToolBar(Command command, CommandEventArgs e) {
		toolBar.Visible = !toolBar.Visible;
	}

	// View status bar.
	[Command.UpdateHandler("View.StatusBar")]
	public void UpdateViewStatusBar(Command command, CommandEventArgs e) {
		command.Selected = statusBar.Visible;
	}

	[Command.ExecuteHandler("View.StatusBar")]
	public void OnViewStatusBar(Command command, CommandEventArgs e) {
		statusBar.Visible = !statusBar.Visible;
	}
	
	// View notes.
	/*[Command.UpdateHandler("View.NotesBar")]
	public void UpdateViewNotesBar(Command command, CommandEventArgs e) {
		command.Selected = notesBar.Visible;
	}

	[Command.ExecuteHandler("View.NotesBar")]
	public void OnViewNotesBar(Command command, CommandEventArgs e) {
		notesBar.Visible = !notesBar.Visible;
	}*/

	// Window menu.

	// New window.
	[Command.UpdateHandler("Window.NewWindow")]
	public void UpdateWindowNewWindow(Command command, CommandEventArgs e) {
		command.Enabled = ActiveDocument != null;
	}
	
	[Command.ExecuteHandler("Window.NewWindow")]
	public void OnWindowNewWindow(Command command, CommandEventArgs e) {
		// Create a view for the active document.
		SimplePadView view = new SimplePadView();
		ActiveDocument.Views.Add(view);

		ViewContainer childForm = new ViewContainer();
		childForm.MdiParent = this;
		childForm.Views.Add(view);
		childForm.Show();
	}

	// Cascade.
	[Command.ExecuteHandler("Window.Cascade")]
	public void OnWindowCascade(Command command, CommandEventArgs e) {
		this.LayoutMdi(MdiLayout.Cascade);
	}

	// Tile.
	[Command.ExecuteHandler("Window.Tile")]
	public void OnWindowTile(Command command, CommandEventArgs e) {
		this.LayoutMdi(MdiLayout.TileHorizontal);
	}

	// Arrange Icons.
	/*[CommandExecuteHandler("Window.ArrangeIcons")]
	public void OnWindowArrangeIcons(Command command, CommandEventArgs e) {
	}*/

	// Help menu.

	[Command.ExecuteHandler("Help.Topics")]
	public void OnHelpTopics(Command command, CommandEventArgs e) {
		Help.ShowHelp(this, "SimplePad.chm");
	}

	[Command.ExecuteHandler("Help.About")]
	public void OnAbout(Command command, CommandEventArgs e) {
		SimplePad.AboutDialog aboutDialog = new SimplePad.AboutDialog();
		aboutDialog.ShowDialog();
	}
}