using System;
using System.Windows.Forms;
using System.Drawing;
using System.Drawing.Printing;
using Genghis.DocView;
using Genghis;

class SimplePadView : Genghis.DocView.View {
	TextBox textBox;

	public SimplePadView() : this(null) {
	}
	
    public SimplePadView(Document document) : base(document) {
		textBox = new TextBox();
		textBox.Multiline = true;
		textBox.Dock = DockStyle.Fill;
		textBox.TextChanged += new EventHandler(OnTextChanged);
		this.Controls.Add(textBox);
	}
    
	public new SimplePadDocument Document {
		get { return (SimplePadDocument) base.Document; }
	}
	
	public TextBox TextBox {
		get { return textBox; }
	}

	protected void OnTextChanged(object sender, EventArgs e) {
		if (Document.Text == textBox.Text)
			return;
	
		Document.Lock(this);
		Document.Text = textBox.Text;
		Document.Unlock();
	}
	
	protected internal override void OnDocumentChanged() {
		if (Document.Text == textBox.Text)
			return;
	
		textBox.Text = Document.Text;
	}

	// File menu.

	// Print state.
	int numLinesPrinted;
	Font printFont;

	// Print.
	[Command.ExecuteHandler("File.Print")]
	protected void OnFilePrint(Command command, CommandEventArgs e) {
		printFont = new Font("Arial", 10);
		numLinesPrinted = 0;

		PrintDocument printDocument = new PrintDocument();
		printDocument.PrintPage += new PrintPageEventHandler(this.PrintPage);

		PrintDialog printDialog = new PrintDialog();
		printDialog.Document = printDocument;
		DialogResult result = printDialog.ShowDialog();

		if (result != DialogResult.Cancel)
			printDocument.Print();
	}

	// Print preview.
	[Command.ExecuteHandler("File.PrintPreview")]
	protected void OnFilePrintPreview(Command command, CommandEventArgs e) {
		printFont = new Font("Arial", 10);
		numLinesPrinted = 0;

		PrintDocument printDocument = new PrintDocument();
		printDocument.PrintPage += new PrintPageEventHandler(this.PrintPage);

		PrintPreviewDialog printPreviewDialog = new PrintPreviewDialog();
		printPreviewDialog.Document = printDocument;
		printPreviewDialog.ShowDialog();
	}


	// The PrintPage event is raised for each page to be printed.
	private void PrintPage(object sender, PrintPageEventArgs e) {
		// Calculate the number of lines per page.
		float linesPerPage = e.MarginBounds.Height / printFont.GetHeight(e.Graphics);

		// Print each line of the file.
		float yPos = 0;
		for (int count = 0; count < linesPerPage; count ++) {
			if (numLinesPrinted + count >= textBox.Lines.Length) {
				// We're all done.
				e.HasMorePages = false;
				numLinesPrinted += count;
				return;
			}

			string line = textBox.Lines[numLinesPrinted + count];
			yPos = e.MarginBounds.Top + (count * printFont.GetHeight(e.Graphics));
			e.Graphics.DrawString(line, printFont, Brushes.Black, e.MarginBounds.Left, yPos, new StringFormat());
		}

		// Print another page.
		e.HasMorePages = true;
		numLinesPrinted += (int) linesPerPage;
	}


	// Edit menu.

	// Undo.
	[Command.UpdateHandler("Edit.Undo")]
	protected void UpdateEditUndo(Command command, CommandEventArgs e) {
		command.Enabled = textBox.CanUndo;
	}

	[Command.ExecuteHandler("Edit.Undo")]
	protected void OnEditUndo(Command command, CommandEventArgs e) {
		textBox.Undo();
	}

	// Cut.
	[Command.UpdateHandler("Edit.Cut")]
	protected void UpdateEditCut(Command command, CommandEventArgs e) {
		command.Enabled = textBox.SelectionLength > 0;
	}

	[Command.ExecuteHandler("Edit.Cut")]
	protected void OnEditCut(Command command, CommandEventArgs e) {
		textBox.Cut();
	}

	// Copy.
	[Command.UpdateHandler("Edit.Copy")]
	protected void UpdateEditCopy(Command command, CommandEventArgs e) {
		command.Enabled = textBox.SelectionLength > 0;
	}

	[Command.ExecuteHandler("Edit.Copy")]
	protected void OnEditCopy(Command command, CommandEventArgs e) {
		textBox.Copy();
	}

	// Paste.
	[Command.UpdateHandler("Edit.Paste")]
	protected void UpdateEditPaste(Command command, CommandEventArgs e) {
		command.Enabled = Clipboard.GetDataObject().GetDataPresent(DataFormats.Text);
	}

	[Command.ExecuteHandler("Edit.Paste")]
	protected void OnEditPaste(Command command, CommandEventArgs e) {
		textBox.Paste();
	}

	// Select all.
	[Command.ExecuteHandler("Edit.SelectAll")]
	protected void OnEditSelectAll(Command command, CommandEventArgs e) {
		textBox.SelectAll();
	}
}