// StreamInfo.cs: Contributed by Paul Bartrum [paulbartrum@mail.com]
// Provides additional information about streams.

using System;
using System.IO;
using System.Net;

// TODO: Handle exceptions.

namespace Genghis {
	public class StreamInfo {
		Stream stream;
		long length;	// Stream length, in bytes (-1 if not known).
		
		public StreamInfo(Stream stream) {
			if (stream == null)
				throw new ArgumentNullException("stream");
			this.stream = stream;
			
			length = -1;
			// Ignore any exceptions.
			try { length = stream.Length; }
			catch { }
		}
		
		/// <summary>
		/// Gets the stream that this object contains information about.</summary>
		/// <value>
		/// The stream that this object contains information about.</value>
		public Stream Stream {
			get { return stream; }
		}
		
		/// <summary>
		/// Gets the URI of the stream.</summary>
		/// <value>
		/// The URI of the stream.</value>
		/// <remarks>Returns a null reference (<c>Nothing</c> in Visual Basic) if
		/// the stream cannot be accessed with a URI.</remarks>
		public virtual Uri Uri {
			get { return null; }
		}
		
		/// <summary>
		/// Gets the name of the stream, including the extension.</summary>
		/// <value>
		/// The name of the stream, including the extension.</value>
		/// <remarks>
		/// <para>This property is usually derived from the URI, if a URI is
		/// available.</para>
		/// <para>Returns a null reference (<c>Nothing</c> in Visual Basic) if
		/// the stream does not have a name.</para></remarks>
		public virtual string Name {
			get { return Uri == null ? null : Path.GetFileName(Uri.AbsolutePath); }
		}

		/// <summary>
		/// Gets the name of the stream, without the extension.</summary>
		/// <value>
		/// The name of the stream, without the extension.</value>
		/// <remarks>
		/// <para>This property is usually derived from the name, if the name
		/// is available.</para>
		/// <para>Returns a null reference (<c>Nothing</c> in Visual Basic) if
		/// the stream does not have a name.</para></remarks>
		public virtual string NameWithoutExtension {
			get { return Name == null ? null : Path.GetFileNameWithoutExtension(Name); }
		}

		/// <summary>
		/// Gets the extension of the stream.</summary>
		/// <value>
		/// The extension of the stream.</value>
		/// <remarks>
		/// <para>An extension always begins with a period.</para>
		/// <para>This property is usually derived from the name, if the name
		/// is available.</para>
		/// <para>Returns a null reference (<c>Nothing</c> in Visual Basic) if
		/// the stream does not have an extension.</para></remarks>
		public virtual string Extension {
			get { return Path.GetExtension(Name); }
		}
		
		/// <summary>
		/// Gets attributes indicating how the stream is stored.</summary>
		/// <value>
		/// Attributes indicating how the stream is stored.</value>
		/// <remarks>
		/// <para>Attributes are normally only available if the stream is a
		/// file.</para>
		/// <para>Returns <see cref="FileAttributes.Normal"/> if this information is
		/// not available.</para></remarks>
		public virtual FileAttributes Attributes {
			get { return FileAttributes.Normal; }
		}

		/// <summary>
		/// Gets the time at which the stream was created.</summary>
		/// <value>
		/// The time at which the stream was created.</value>
		/// <remarks>
		/// <para>The creation time is normally only available if the stream
		/// is a file.</para>
		/// <para>Returns <see cref="DateTime.MinValue"/> if this information
		/// is not available.</para></remarks>
		public virtual DateTime CreationTime {
			get { return DateTime.MinValue; }
		}

		/// <summary>
		/// Gets the time the stream was last accessed.</summary>
		/// <value>
		/// The time the stream was last accessed.</value>
		/// <remarks>
		/// <para>The last access time is normally only available if the stream
		/// is a file.</para>
		/// <para>Returns <see cref="DateTime.MinValue"/> if this information
		/// is not available.</para></remarks>
		public virtual DateTime LastAccessTime {
			get { return DateTime.MinValue; }
		}
		
		/// <summary>
		/// Gets the time the stream was last modified.</summary>
		/// <value>
		/// The time the stream was last modified.</value>
		/// <remarks>
		/// <para>Returns <see cref="DateTime.MinValue"/> if this information
		/// is not available.</para></remarks>
		public virtual DateTime LastWriteTime {
			get { return DateTime.MinValue; }
		}

		/// <summary>
		/// Gets the MIME type of the stream.</summary>
		/// <value>
		/// The MIME type of the stream.</value>
		/// <remarks>
		/// <para>Returns "binary/octet-stream" if this information is
		/// not available.</para></remarks>
		public virtual string ContentType {
			get { return "binary/octet-stream"; }
		}

		/// <summary>
		/// Gets the length of the stream, in bytes.</summary>
		/// <value>
		/// The length of the stream, in bytes.</value>
		/// <remarks>
		/// Returns -1 if this information is not available.</remarks>
		public virtual long Length {
			get {
				// This information should be available even if the stream is closed.
				return length;
			}
		}
	}
	
	public class FileStreamInfo : StreamInfo {
		string fileName;
		FileInfo fileInfo;
		
		public FileStreamInfo(Stream stream, string fileName) : base(stream) {
			if (fileName == null)
				throw new ArgumentNullException("fileName");
			this.fileName = fileName;
			
			fileInfo = new FileInfo(fileName);
		}

		/// <summary>
		/// Gets the path of the file.</summary>
		/// <value>
		/// The path of the file.</value>
		public string FullName {
			get { return fileName; }
		}

		/// <summary>
		/// Gets the URI of the file.</summary>
		/// <value>
		/// The URI of the stream, beginning with "file:".</value>
		public override Uri Uri {
			get { return new Uri(fileName); }
		}

		/// <summary>
		/// Gets the name of the file, including the extension.</summary>
		/// <value>
		/// The name of the stream, including the extension.</value>
		public override string Name {
			get { return Path.GetFileName(fileName); }
		}

		/// <summary>
		/// Gets attributes indicating how the file is stored.</summary>
		/// <value>
		/// Attributes indicating how the stream is stored.</value>
		public override FileAttributes Attributes {
			get { return fileInfo.Attributes; }
		}

		/// <summary>
		/// Gets the time at which the file was created.</summary>
		/// <value>
		/// The time at which the file was created.</value>
		public override DateTime CreationTime {
			get { return fileInfo.CreationTime; }
		}

		/// <summary>
		/// Gets the time the file was last accessed.</summary>
		/// <value>
		/// The time the file was last accessed.</value>
		/// <remarks>
		/// <para>Returns <see cref="DateTime.MinValue"/> if this information
		/// is not available.</para></remarks>
		public override DateTime LastAccessTime {
			get { return fileInfo.LastAccessTime; }
		}

		/// <summary>
		/// Gets the time the file was last modified.</summary>
		/// <value>
		/// The time the file was last modified.</value>
		public override DateTime LastWriteTime {
			get { return fileInfo.LastWriteTime; }
		}

		/// <summary>
		/// Gets the length of the file, in bytes.</summary>
		/// <value>
		/// The length of the file, in bytes.</value>
		public override long Length {
			get {
				if (base.Length != -1)
					return base.Length;
				return fileInfo.Length;
			}
		}
	}
	
	public class WebStreamInfo : StreamInfo {
		WebResponse webResponse;
		
		public WebStreamInfo(WebResponse webResponse) : base(webResponse.GetResponseStream()) {
			if (webResponse == null)
				throw new ArgumentNullException("webResponse");
			this.webResponse = webResponse;
		}

		/// <summary>
		/// Gets the <see cref="WebResponse"/> this object was constructed
		/// from.</summary>
		/// <value>
		/// The <see cref="WebResponse"/> this object was constructed from.</value>
		public WebResponse WebResponse {
			get { return webResponse; }
		}
		
		/// <summary>
		/// Gets the URI of the internet resource.</summary>
		/// <value>
		/// The URI of the internet resource.</value>
		public override Uri Uri {
			get { return webResponse.ResponseUri; }
		}

		/// <summary>
		/// Gets the time the internet resource was last modified.</summary>
		/// <value>
		/// The time the internet resource was last modified.</value>
		public override DateTime LastWriteTime {
			get {
				string lastModified = webResponse.Headers["Last-Modified"];
				if (lastModified == null)
					return DateTime.MinValue;
				return DateTime.Parse(lastModified);
			}
		}

		/// <summary>
		/// Gets the MIME type of the internet resource.</summary>
		/// <value>
		/// The MIME type of the internet resource.</value>
		public override string ContentType {
			get { return webResponse.Headers["Content-Type"]; }
		}

		/// <summary>
		/// Gets the length of the internet resource, in bytes.</summary>
		/// <value>
		/// The length of the internet resource, in bytes.</value>
		public override long Length {
			get {
				string contentLength = webResponse.Headers["Content-Length"];
				if (contentLength == null)
					return -1;
				return Int64.Parse(contentLength);
			}
		}
	}
	
	public class NamedStreamInfo : StreamInfo {
		string name;

		public NamedStreamInfo(Stream stream, string name) : base(stream) {
			this.name = name;
		}

		/// <summary>
		/// Gets the name of the stream, including the extension.</summary>
		/// <value>
		/// The name of the stream, including the extension.</value>
		public override string Name {
			get { return name; }
		}
	}
	
	/*
	public class StreamInfoUnitTests : TestCase { 
		protected override void SetUp() {
		}
		
		public static ITest Suite {
			get {
				return new TestSuite(typeof(StreamInfoUnitTests));
			}
		}
		
		public void TestFileStreamInfo() {
			string path = @"C:\Documents and Settings\Paul\My Documents\Projects\Genghis\SimplePad\Notes.txt";
			Console.WriteLine("Testing FileStreamInfo on the file: " + path);
			StreamInfo streamInfo = new FileStreamInfo(File.OpenRead(path), path);
			Test(streamInfo);
			Console.WriteLine("");
			
			string name = @"Notes.txt";
			Console.WriteLine("Testing NamedStreamInfo on the file: " + path);
			streamInfo = new NamedStreamInfo(File.OpenRead(path), name);
			Test(streamInfo);
			Console.WriteLine("");
			
			string url = "http://www.google.co.nz/microsoft.html";
			WebRequest webRequest = WebRequest.Create(url);
			WebResponse webResponse = webRequest.GetResponse();
			Console.WriteLine("Testing WebStreamInfo on the URL: " + url);
			streamInfo = new WebStreamInfo(webResponse);
			Test(streamInfo);
			webResponse.Close();
		}
		
		public static void Test(StreamInfo streamInfo) {
			Console.WriteLine("Stream:                 " + streamInfo.Stream);
			Console.WriteLine("URI:                    " + streamInfo.Uri);
			Console.WriteLine("Name:                   " + streamInfo.Name);
			Console.WriteLine("Name without extension: " + streamInfo.NameWithoutExtension);
			Console.WriteLine("Extension:              " + streamInfo.Extension);
			Console.WriteLine("Attributes:             " + streamInfo.Attributes);
			Console.WriteLine("Creation time:          " + streamInfo.CreationTime);
			Console.WriteLine("Last access time:       " + streamInfo.LastAccessTime);
			Console.WriteLine("Last write time:        " + streamInfo.LastWriteTime);
			Console.WriteLine("Content type:           " + streamInfo.ContentType);
			Console.WriteLine("Length:                 " + streamInfo.Length);
		}
	}*/
}