using System;
using System.Collections;
using System.Diagnostics;
using System.Windows.Forms;
using Genghis;

namespace Genghis.DocView {

	public class ViewContainer : Form, ICommandTarget {
		ViewCollection views;

		Hashtable executeHandlers;
		Hashtable updateHandlers;

		public ViewContainer() {
			views = new ViewCollection();
			views.ItemsAdded += new ViewCollectionEventHandler(OnViewsChanged);
			views.ItemsRemoved += new ViewCollectionEventHandler(OnViewsChanged);
			
			executeHandlers = Command.GetExecuteHandlers(this);
			updateHandlers = Command.GetUpdateHandlers(this);
		}
		
		// Used to hook up command event handlers.
		public Hashtable ExecuteHandlers {
			get { return executeHandlers; }
		}
		public Hashtable UpdateHandlers {
			get { return updateHandlers; }
		}
		
		public ViewCollection Views {
			get { return views; }
		}
		
		/// <summary>
		/// A view was added or removed from the Views collection.</summary>
		protected virtual void OnViewsChanged(object sender, ViewCollectionEventArgs e) {
			// Update document of each view.
			if (e.Added) {
				foreach (View view in e.Items) {
					// A view was added to this container.
					view.Dock = DockStyle.Fill;
					this.Controls.Add(view);
					
					// Make sure we receive title updates so we can update the window title.
					view.Document.TitleChanged += new EventHandler(OnTitleChanged);
					view.Document.ModifiedChanged += new EventHandler(OnTitleChanged);
					OnTitleChanged(this, EventArgs.Empty);	// Invoke TitleChanged event.
				}
			}
		}
		
		protected override void OnClosing(System.ComponentModel.CancelEventArgs e) {
			base.OnClosing(e);
			if (e.Cancel)
				return;
			
			// Hook into the closing event so we can offer a "Do you want to save?" dialog.
			// NOTE: Do not do any actual closing of any resources in this method.
			
			// Calculate the number of views for each document AFTER this window has been closed.
			
			// Map document to number of views (int).
			Hashtable documentViewCount = new Hashtable(1);
			foreach (View view in views) {
				if (view.Document != null) {
					if (documentViewCount[view.Document] == null)
						documentViewCount[view.Document] = view.Document.Views.Count;
					int viewCount = (int) documentViewCount[view.Document];
					documentViewCount[view.Document] = viewCount - 1;
					Debug.Assert((int) documentViewCount[view.Document] >= 0);
				}
			}
			
			// For each document that has no views after this window has been closed
			// and has been modified, display a dialog asking whether the user wants
			// to save changes.
			
			foreach (DictionaryEntry entry in documentViewCount) {
				if ((int) entry.Value == 0) {
					// No more views in the document.
					Document unsavedDocument = (Document) entry.Key;
					if (unsavedDocument.Modified) {
						DialogResult result = MessageBox.Show("Save changes to " + unsavedDocument.Title + "?",
							"SimplePad", MessageBoxButtons.YesNoCancel, MessageBoxIcon.Warning);
						
						if (result == DialogResult.Cancel) {
							e.Cancel = true;
							return;
						}
						else if (result == DialogResult.Yes) {
							// TODO: fix.
							unsavedDocument.OnFileSave(null, null);
						}
					}
				}
			}
		}
		
		protected override void OnClosed(EventArgs e) {
			// Close views and document.
			foreach (View view in views) {
				view.Close();
			}
		}

		private void OnTitleChanged(object sender, EventArgs e) {
			if (views.Count > 0) {
				if (views[0].Document.Modified)
					this.Text = views[0].Document.Title + " *";
				else
					this.Text = views[0].Document.Title;
			}
		}

		// File menu.

		// Close.
		[Command.ExecuteHandler("File.Close")]
		protected void OnFileClose(Command command, CommandEventArgs e) {
			this.Close();
		}
	}
}