// ContainerContainerStatusBar.cs: Contributed by Ethan Smith [ethan.smith@pobox.com]
// Status bar that can contain other controls, including design-time drag-n-drop support
#region Copyright  2002-2003 The Genghis Group
/*
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from the
 * use of this software.
 * 
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, subject to the following restrictions:
 * 
 * 1. The origin of this software must not be misrepresented; you must not claim
 * that you wrote the original software. If you use this software in a product,
 * an acknowledgment in the product documentation is required, as shown here:
 * 
 * Portions Copyright  2002-2003 The Genghis Group (http://www.genghisgroup.com/).
 * 
 * 2. No substantial portion of the source code of this library may be redistributed
 * without the express written permission of the copyright holders, where
 * "substantial" is defined as enough code to be recognizably from this library. 
*/
#endregion
#region History
/// 1.0.0.0: Initial Version
/// 1.0.0.1: Added Drag & Drop design time support
#endregion
#region To Do
/// To do:
/// Filter HostedControl property pick list to only include controls that are 
///     children of the parent ContainerStatusBar.
#endregion

using System;
using System.ComponentModel;
using System.Drawing;
using WinForms = System.Windows.Forms;
using System.Collections;
using Genghis.Windows.Forms.Design;

namespace Genghis.Windows.Forms
{
    /// <summary>
    /// A substitute for the System.Windows.Forms.ContainerStatusBar
    /// that allows any control to be hosted in a status bar panel.
    /// </summary>
    /// <remarks>
    /// Provides full designer support for adding and customizing panels.
    /// 
    /// To use: Add the <c>Genghis.Windows.Forms.ContainerStatusBar</c> to your toolbox.
    /// From the toolbox, add the ContainerStatusBar to your form and use as a normal
    /// ContainerStatusBar. To host a control, the control must be a child of the 
    /// status bar. It can then be set as the value for the HostedControl
    /// property on one of the status bar's panels.
    /// </remarks>
    [Designer(typeof(ContainerStatusBarDesigner))]
    public class ContainerStatusBar : WinForms.StatusBar
    {
        #region Instance variables
        private StatusBarPanelCollection _InternalPanels;
        #endregion
        
          #region Properties
        
        /// <summary>
        /// Typed collection of <see cref="ContainerStatusBarpanel"/>s
        /// </summary>
        /// <remarks>Provides the same functionality as 
        /// <seealso cref="System.Windows.Forms.ContainerStatusBar.Panels"/>
        /// except that it supports <see cref="StatbarPanel"/> instead of
        /// <see cref="System.Windows.Forms.ContainerStatusBarpanel"/>
        /// </remarks>
        [Localizable(true),
        DesignerSerializationVisibility(DesignerSerializationVisibility.Content),
        MergableProperty(false),
        Category("Appearance"),
        Description("The panels in the status bar.")]
        public new StatusBarPanelCollection Panels
        {
            get
            {
                if (this._InternalPanels == null)
                {
                    this._InternalPanels = 
                        new StatusBarPanelCollection(this,base.Panels);
                }
                return _InternalPanels;
            }
        }
        #endregion

        #region Event Handlers
        /// <summary>
        /// Override standard processing of removing child controls from the ContainerStatusBar.
        /// Ensures that no panels host the control that has been removed.
        /// </summary>
        /// <remarks>Find the panel hosting the control and set it's hosted control
        /// property to null.
        /// </remarks>
        /// <param name="e"></param>
        protected override void OnControlRemoved(WinForms.ControlEventArgs e)
        {
            base.OnControlRemoved(e);

            if(e!=null && e.Control!= null)
            {
                ContainerStatusBarPanel HostingPanel = this.GetHostingPanel(e.Control);
                if(HostingPanel!=null)
                {
                    HostingPanel.HostedControl=null;
                }
            }
        }
        /// <summary>
        /// Handles drawing of ContainerStatusBarpanles with a value for their NestedControl
        /// Property.
        /// </summary>
        /// <remarks>Calls base implementation upon completion.</remarks>
        /// <param name="e"></param>
        protected override void OnDrawItem(WinForms.StatusBarDrawItemEventArgs e)
        {
            ContainerStatusBarPanel panel = e.Panel as ContainerStatusBarPanel;
            if(panel != null)
            {
                if( panel.Style == WinForms.StatusBarPanelStyle.OwnerDraw
                    && panel.HostedControl != null)
                {
                    /// Resize the control to the panel bounds
                    /// and adjust for padding
                    Rectangle paddedBounds = e.Bounds;
                    paddedBounds.X+= panel.ControlPaddingLeft;
                    paddedBounds.Y+= panel.ControlPaddingTop;
                    paddedBounds.Width-= panel.ControlPaddingLeft+panel.ControlPaddingRight;
                    paddedBounds.Height-= panel.ControlPaddingTop+panel.ControlPaddingBottom;
                    if(!paddedBounds.Equals(panel.HostedControl.Bounds))
                        panel.HostedControl.Bounds = paddedBounds;
                }
            }
            /// placed after our drawing code so that event notifications don't
            /// go out till we've done our drawing.
            base.OnDrawItem(e);
        }
        #endregion

        #region Utilities
        /// <summary>
        /// Examines the exisitng panels hosted by the ContainerStatusBar to determine
        /// if any of them already host the control. Used in design mode to 
        /// ensure that a control is not placed on more than one panel.
        /// </summary>
        /// <param name="hostedControl"></param>
        /// <returns>The <c>ContainerStatusBarPanel hosting the control or null if the
        /// control is not hosted on any panels.</c></returns>
        public ContainerStatusBarPanel GetHostingPanel(WinForms.Control hostedControl)
        {
            foreach(ContainerStatusBarPanel panel in this.Panels)
            {
                if(panel.HostedControl == hostedControl)
                    return panel;
            }
            return null;
        }

        private void InitializeComponent()
        {
            // 
            // ContainerStatusBar
            // 
            this.AllowDrop = true;

        }

        /// <summary>
        /// Determines the <c>ContainerStatusBarPanel</c> that is under the point
        /// </summary>
        /// <param name="p">The point in the ContainerStatusBar's coordinate system</param>
        /// <returns>The panel under the point, or null if the point is out 
        /// of range.</returns>
        public ContainerStatusBarPanel GetPanelAtPoint(Point p)
        {
            if(p.X<0 ||p.X>this.Width || p.Y<0 || p.Y>this.Height)
                return null;
            int CumulativeWidth=0;
            foreach(ContainerStatusBarPanel panel in this.Panels)
            {
                CumulativeWidth+=panel.Width;
                if(p.X<=CumulativeWidth)
                    return panel;
            }
            return null;
        }
        #endregion

        #region Nested Class
        /// <summary>
        /// Typed collection of ContainerStatusBarpanels.
        /// </summary>
        /// <remarks>
        /// Delegates storgae to the base StatsuBar's panels collection, but 
        /// performs type checking and type conversions to ensure that item 
        /// are ContainerStatusBarpanels, not System.Windows.Forms.ContainerStatusBarPanels.
        /// </remarks>
        public new class StatusBarPanelCollection: ICollection,IList,IEnumerable
        {
            #region Instance Variables
            private WinForms.StatusBar.StatusBarPanelCollection _RealPanels;
            private ContainerStatusBar _Owner;
            #endregion

            #region Constructor
            internal StatusBarPanelCollection(ContainerStatusBar owner
                ,WinForms.StatusBar.StatusBarPanelCollection realPanels)
            {
                _Owner = owner;
                _RealPanels = realPanels;
            }
            #endregion

            #region Properties
            /// <summary>
            /// The number of panels in the collection
            /// </summary>
            public virtual int Count
            {
                get{return _RealPanels.Count;} 
            }
            public virtual bool IsReadOnly
            {
                get{return _RealPanels.IsReadOnly;}
            }
            public virtual ContainerStatusBarPanel this[int index]
            {
                get{return (ContainerStatusBarPanel) _RealPanels[index];}
                set{_RealPanels[index] = value;}
            }
            #endregion
            
            #region Typed Collection Methods
            /// Allow adding and remving ContainerStatusBarPanels from the underlying 
            /// collection. all type checking is done base on function argument
            /// types.
            public virtual ContainerStatusBarPanel Add(string text)
            {
                ContainerStatusBarPanel panel = new ContainerStatusBarPanel();
                panel.Text = text;
                Add(panel);
                return panel;
            }
            public virtual int Add(ContainerStatusBarPanel panel)
            {
                return _RealPanels.Add(panel);
            }
            public virtual void AddRange(ContainerStatusBarPanel[] panels)
            {
                _RealPanels.AddRange(panels);
            }
            public virtual void Clear()
            {
                _RealPanels.Clear();
            }
            public bool Contains(ContainerStatusBarPanel panel)
            {
                return _RealPanels.Contains(panel);
            }
            public virtual IEnumerator GetEnumerator()
            {
                return _RealPanels.GetEnumerator();
            }
            public int IndexOf(ContainerStatusBarPanel panel)
            {
                return _RealPanels.IndexOf(panel);
            }
            public virtual void Insert(int index, ContainerStatusBarPanel panel)
            {
                _RealPanels.Insert(index,panel);
            }
            public virtual void Remove(ContainerStatusBarPanel panel)
            {
                _RealPanels.Remove(panel);
            }
            public virtual void RemoveAt(int index)
            {
                _RealPanels.RemoveAt(index);
            }
            #endregion

            #region Untyped Collection Methods
            void ICollection.CopyTo(Array dest, int index) 
            {
                ((ICollection)_RealPanels).CopyTo(dest, index);
            }
            bool ICollection.IsSynchronized
            {
                get{return ((ICollection)_RealPanels).IsSynchronized;}
            }
            object ICollection.SyncRoot 
            {
                /// haven't thought about this much,but to be safe we won't 
                /// delegate this one
                get{return this;}
            }
            int IList.Add(object value) 
            {
                if (value as ContainerStatusBarPanel != null) 
                {
                    return this.Add((ContainerStatusBarPanel) value);
                }
                throw BuildBadTypeException("value");
            }
            bool IList.Contains(object value) 
            {
                if (value as ContainerStatusBarPanel != null) 
                {
                    return this.Contains((ContainerStatusBarPanel) value);
                }
                throw BuildBadTypeException("value");
            }
            int IList.IndexOf(object value) 
            {
                if (value as ContainerStatusBarPanel != null) 
                {
                    return this.IndexOf((ContainerStatusBarPanel) value);
                }
                throw BuildBadTypeException("value");
            }
            void IList.Insert(int index, object value)
            {
                if (value as ContainerStatusBarPanel != null) 
                {
                    this.Insert(index, (ContainerStatusBarPanel) value);
                    return;
                }
                throw BuildBadTypeException("value");
            }
            bool IList.IsFixedSize 
            {
                get{return ((IList)_RealPanels).IsFixedSize;}
            }
            void IList.Remove(object value)
            {
                if (value as ContainerStatusBarPanel != null) 
                {
                    this.Remove((ContainerStatusBarPanel) value);
                }
                throw BuildBadTypeException("value");
            }
            object IList.this[int index] 
            {
                get{return this[index];}
                set
                {
                    if (value as ContainerStatusBarPanel != null) 
                    {
                        this[index]=(ContainerStatusBarPanel)value;
                    }
                    throw BuildBadTypeException("value");
                }
            }
            private ArgumentException BuildBadTypeException(string parameterName)
            {
                return new ArgumentException("Only instances of '"
                    + typeof(ContainerStatusBarPanel).FullName
                    + "' or derived types can be added"
                    , parameterName);
            }
            #endregion
        }
        #endregion
    }

    /// <summary>
    /// A modified ContainerStatusBarPanel that adds support for a hosted Control
    /// </summary>
    public class ContainerStatusBarPanel: WinForms.StatusBarPanel
    {
        #region Instance variables
        private WinForms.Control _HostedControl;        //Hosted Control reference
        private int _ControlPaddingLeft=0;              //Property storage
        private int _ControlPaddingTop=0;               //Property storage
        private int _ControlPaddingRight=0;             //Property storage
        private int _ControlPaddingBottom=0;            //Property storage
        private bool _Initializing=false;               //Supports System.ComponentModel.ISupportInitialize
        #endregion

        #region Properties
        /// <summary>
        /// The <see cref="StatuBar"/> control hosting the panel.
        /// </summary>
        /// <remarks>
        /// You can use this property to determine the <c>ContainerStatusBar</c>
        /// control that a <c>ContainerStatusBarPanel</c> object is 
        /// contained within. <c>ContainerStatusBarPanel</c> objects are added to a 
        /// <c>ContainerStatusBar</c> control by calling the Add method of the 
        /// ContainerStatusBar.Panels property. The ContainerStatusBar.Panels property is a 
        /// collection of a <c>ContainerStatusBarPanel</c> items.
        /// </remarks>
        [Browsable(false)]
        public new ContainerStatusBar Parent
        {
            get{return base.Parent as ContainerStatusBar;}
        }

        /// <summary>
        /// Specifies the control that will be displayed in the panel. 
        /// <see cref="ContainerStatusBar"/>.
        /// </summary>
        [Localizable(false),
        MergableProperty(false),
        Category("Appearance"),
        Description("Specifies the control that will be displayed in the panel."),
        DefaultValue(null)]
        public WinForms.Control HostedControl
        {
            get{return _HostedControl; }
            set
            {
                if(value == _HostedControl)
                    return;
                if(value != null && this.Parent != null )
                {
                    /// Make sure the the hosted control a child of the
                    /// panel's parent status bar.
                    if(!this.Parent.Controls.Contains(value))
                    {
                        throw new ArgumentException("The HostedControl property may only be set to one of the panel's parent status bar's chid controls.");
                    }
                    /// Make sure the control is not already hosted on 
                    /// another panel
                    ContainerStatusBarPanel hostingPanel = Parent.GetHostingPanel(value);
                    if(hostingPanel != null)
                    {
                        //TODO: move design time code to designer
                        if(this.Site!=null && this.Site.DesignMode)
                        {
                            hostingPanel.HostedControl=null;
                        }
                        else
                            throw new ArgumentException("The control '" + value.Name +"' is being hosted on panel '"
                                +Parent.GetHostingPanel(value).Text 
                                +"'. You must remove it from the existing panel before it can be hosted on this panel.");
                    }
                }
                this._HostedControl = value;
                this.OnHostedControlChanged();
            }
        }

        // Wish I had templates!!!!!

        /// <summary>
        /// The number of pixels to pad betwen the left edge of the panel and
        /// the left edge of the hosted control.
        /// </summary>
        [Localizable(false),
        Category("Layout"),
        Description("Number of pixels from the left edge of the panel."),
        DefaultValue(0)]
        public int ControlPaddingLeft
        {
            get{return _ControlPaddingLeft;}
            set{SetControlPadding(ref _ControlPaddingLeft,value);}
        }
        /// <summary>
        /// The number of pixels to pad betwen the top edge of the panel and
        /// the top edge of the hosted control.
        /// </summary>
        [Localizable(false),
        Category("Layout"),
        Description("Number of pixels from the top edge of the panel."),
        DefaultValue(0)]
        public int ControlPaddingTop
        {
            get{return _ControlPaddingTop;}
            set{SetControlPadding(ref _ControlPaddingTop,value);} 
        }
        /// <summary>
        /// The number of pixels to pad betwen the left edge of the panel and
        /// the left edge of the hosted control.
        /// </summary>
        [Localizable(false),
        Category("Layout"),
        Description("Number of pixels from the right edge of the panel."),
        DefaultValue(0)]
        public int ControlPaddingRight
        {
            get{return _ControlPaddingRight;}
            set{SetControlPadding(ref _ControlPaddingRight,value);}
        }
        /// <summary>
        /// The number of pixels to pad betwen the bottom edge of the panel and
        /// the bottom edge of the hosted control.
        /// </summary>
        [Localizable(false),
        Category("Layout"),
        Description("Number of pixels from the bottom edge of the panel."),
        DefaultValue(0)]
        public int ControlPaddingBottom
        {
            get{return _ControlPaddingBottom;}
            set{SetControlPadding(ref _ControlPaddingBottom,value);}
        }
        #endregion
        
        #region Helpers
        private void Refresh()
        {
            if( this.Parent != null && !this._Initializing )
                this.Parent.Refresh();
        }

        /// <summary>
        /// Applies the value to the particular control padding vairable.
        /// </summary>
        /// <param name="controlPaddingVar">A reference to one of the control
        /// padding instance variables.</param>
        /// <param name="value">The the padding value to apply</param>
        private void SetControlPadding(ref int controlPaddingVar, int value)
        {
            // Still wish I had templates
            if(controlPaddingVar==value)
                return;
            controlPaddingVar = value;
            this.OnPaddingChanged();
        }
        #endregion

        #region Event Handlers
        /// <summary>
        /// What to do when the value of the HostControl has chnaged
        /// We do not fire public events for this event
        /// </summary>
        protected virtual void OnHostedControlChanged()
        {
            if(this.Parent != null)
            {
                if(this.HostedControl!= null )
                {
                    /// Apply reasonable default properties
                    /// Trying not to be overly helpfull, as that
                    /// can end up getting in the way.
                    /// TODO: Move design time code into designer
                    if(this.Site !=null && this.Site.DesignMode)
                    {
                        this.Style = WinForms.StatusBarPanelStyle.OwnerDraw;
                    }
                    /// Don't want Winforms layout getting in the way
                    this.HostedControl.Dock = WinForms.DockStyle.None;
                    /// Makes dragging repaint smoother
                    this.HostedControl.Anchor = WinForms.AnchorStyles.Bottom
                        | WinForms.AnchorStyles.Left;
                }
                this.Refresh();
            }
        }
        /// <summary>
        /// What to do when any of the padding values have changed.
        /// </summary>
        /// <remarks>Public events are not fired for this event.</remarks>
        protected void OnPaddingChanged()
        {
            this.Refresh();
        }
        #endregion

        #region ISupportsInitialize
        /// <summary>
        /// Override base class implementation so we can trap the notification.
        /// </summary>
        /// <remarks>Calls into the base implementation.
        /// <seealso cref="ISupportsInitialize"/></remarks>
        public new void BeginInit() 
        {
            base.BeginInit();
            this._Initializing = true;
        }
        /// <summary>
        /// Override base class implementation so we can trap the notification.
        /// </summary>
        /// <remarks>Calls into the base implementation.
        /// <seealso cref="ISupportsInitialize"/></remarks>
        public new void EndInit() 
        {
            this._Initializing = false;
            base.EndInit();
        }
        #endregion
    }
}

namespace Genghis.Windows.Forms.Design
{
    using System;
    using System.ComponentModel;
    using System.ComponentModel.Design;
    using System.Drawing;
    using System.Reflection;
    using WinForms = System.Windows.Forms;

    /// <summary>
    /// Designer that provides additional cusomization to the Genghis
    /// StatuBar, when it is in deisgn mode in VS.NET.
    /// </summary>
    /// <remarks>
    /// Provide Drag & Drop editing and inteligent property default values.
    /// </remarks>
    public class ContainerStatusBarDesigner : WinForms.Design.ParentControlDesigner
    {
        #region Initialization
        /// <summary>
        /// Prepares the designer to view, edit, and design ContainerStatusBar
        /// </summary>
        /// <param name="component">The component being designed. This
        /// should be a ContainerStatusBar, but no checks are made for this. if
        /// this designer is used on another control, not inherited from
        /// ContainerStatusBar, it will crash.</param>
        public override void Initialize(IComponent component) 
        {
            base.Initialize(component);
            this.DrawGrid = false;
            this.Control.ControlAdded+=
                new WinForms.ControlEventHandler(this.OnChildControlAdded);
        }
        #endregion

        #region Properties
 
        /// <summary>
        /// A reference to the ContainerStatusBar being designed.
        /// </summary>
        public ContainerStatusBar ContainerStatusBar
        {
            get
            {return ((ContainerStatusBar)this.Control);}
        }

        #endregion
        
        #region Event Handlers
        /// <summary>
        /// Attempts to host the control on the panel on which it was dropped.
        /// </summary>
        /// <param name="e"></param>
        protected override void OnDragDrop(WinForms.DragEventArgs e)
        {
            base.OnDragDrop(e);     // perform standard processing

            if( e.Effect == WinForms.DragDropEffects.Move
                || e.Effect == WinForms.DragDropEffects.Copy)
            {
                /// Get Dropped Control
                WinForms.Control[] droppedControls = ExtractDroppedControls(e);
                if(droppedControls.Length==1)
                {
                    /// Get the panel that the object was droped on
                    ContainerStatusBarPanel dropPanel = this.ContainerStatusBar.GetPanelAtPoint(GetLocalPoint(e.X,e.Y));
                    ApplyControlToPanel(droppedControls[0],dropPanel);
                }
                else
                {
                    //fill empty panels with dropped controls
                    ApplyControlsToPanels(droppedControls);
                }
            }
        }
        /// <summary>
        /// Overrides standard processing of adding child controls to the ContainerStatusBar.
        /// Setts appropriate default properties.
        /// </summary>
        private void OnChildControlAdded(object sender, WinForms.ControlEventArgs e)
        {
            if(e!=null && e.Control!= null)
            {
                if(e.Control.Left<0)
                    e.Control.Left=0;
                if(e.Control.Top<0)
                    e.Control.Top=0;
                e.Control.BringToFront();
            }
        }
 
        #endregion

        #region Designer Overrides

        protected override IComponent[] CreateToolCore(
            System.Drawing.Design.ToolboxItem tool,
            int x, int y, int width, int height,
            bool hasLocation, bool hasSize)
        {
            IComponent[] components = base.CreateToolCore(
                tool,x,y,width,height,hasLocation,hasSize);
            if(!hasSize)
            {
                // User requested creation at a particular point
                if( hasLocation
                    && components.Length>0
                    && components[0] is WinForms.Control)
                {

                    ContainerStatusBarPanel panel = this.ContainerStatusBar.GetPanelAtPoint(
                        GetLocalPoint(x,y));
                    if(panel!=null)
                    {
                        //System.Diagnostics.Debugger.Break();
                        panel.HostedControl = (WinForms.Control) components[0];
                        //TODO: Clean up control focus rectangle
                    }
                }
            }
            return components;
        }
        #endregion

        #region Helper Methods
        private Point GetLocalPoint(int x, int y)
        {
            return this.ContainerStatusBar.PointToClient(new Point(x,y));
                    
        }
        /// <summary>
        /// Extracts the dropped controls from the DropEvent data.
        /// </summary>
        /// <param name="e"></param>
        /// <returns>An array of the dropped Controls.</returns>
        private WinForms.Control[] ExtractDroppedControls(WinForms.DragEventArgs e)
        {
            WinForms.Control[] droppedControls = new WinForms.Control[0];
            if(e.Data.GetType().FullName.Equals("System.Windows.Forms.Design.OleDragDropHandler+ComponentDataObjectWrapper"))
            {
                try
                {
                    /// Winforms designer uses the internal type 
                    /// System.Windows.Forms.Design.OleDragDropHandler.ComponentDataObjectWrapper
                    /// to wrap dragdrop controls in the designer. To bypass the
                    /// internal security protection, use reflection.
                    /// This is undocumented and brittle, but its only purpose is
                    /// to provide design time drag & drop editing, so if it breaks,
                    /// it is not a big deal.
                    PropertyInfo prop = e.Data.GetType().GetProperty("InnerData");
                    if(prop != null)
                    {
                        //use reflection to get around private objects
                        object data = prop.GetValue(e.Data,null);
                        prop = data.GetType().GetProperty("Components");
                        if(prop!=null)
                        {
                            object[] droppedComponents
                                = (object[])(prop.GetValue(data,null));
    
                            System.Collections.ArrayList droppedControlsList
                                = new System.Collections.ArrayList(
                                droppedComponents.Length);
                            foreach(object o in droppedComponents)
                            {
                                if(o is WinForms.Control)
                                    droppedControlsList.Add(o);
                            }
                            droppedControls = new WinForms.Control
                                [droppedControlsList.Count];
                            droppedControlsList.CopyTo(droppedControls);
                        }
                    }
                }
                catch(AmbiguousMatchException){}
                catch(ArgumentException){}
                catch(TargetException){}
                catch(TargetParameterCountException){}
                catch(MethodAccessException){}
            }
           
            return droppedControls;
        }

        /// <summary>
        /// Sets the panel's Hosted control property to the supplied control.
        /// If the control is currently hosted on another panel, swithces the
        /// controls between the panels. 
        /// </summary>
        /// <param name="control">The control to host</param>
        /// <param name="panel">The panel on which to host the control</param>
        private void ApplyControlToPanel(WinForms.Control control, ContainerStatusBarPanel panel)
        {
            if(control != null && panel != null)
            {
                if(panel.HostedControl != null)
                {
                    if(panel.HostedControl==control)
                        return;
                        
                    /// If Control is hosted on another panel, switch the controls
                    ContainerStatusBarPanel originalPanel = 
                        this.ContainerStatusBar.GetHostingPanel(control);
                    if(originalPanel!=null)
                    {
                        originalPanel.HostedControl=panel.HostedControl;
                    }
                }
                panel.HostedControl=control;
            }
        }
        /// <summary>
        /// Fills empty panels with Controls from the supplied array.</summary>
        /// <param name="controls">The controls to host</param>
        private void ApplyControlsToPanels(WinForms.Control[] controls)
        {
            foreach(WinForms.Control control in controls)
            {
                ContainerStatusBarPanel existingPanel=null;
                ContainerStatusBarPanel emptyPanel=null;

                foreach(ContainerStatusBarPanel panel in this.ContainerStatusBar.Panels)
                {
                    if(panel.HostedControl==null && emptyPanel== null)
                        emptyPanel=panel;
                    else if(panel.HostedControl == control)
                    {
                        existingPanel = panel;
                        break;
                    }
                }
                if(existingPanel==null & emptyPanel!=null)
                {
                    emptyPanel.HostedControl=control;
                }
            }
        }
        #endregion
    }
}
