// FileDocument.cs: Contributed by Chris Sells [csells@sellsbrothers.com]
// A helper for dirty bit/current file name management
#region Copyright  2002-2003 The Genghis Group
/*
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from the
 * use of this software.
 * 
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, subject to the following restrictions:
 * 
 * 1. The origin of this software must not be misrepresented; you must not claim
 * that you wrote the original software. If you use this software in a product,
 * an acknowledgment in the product documentation is required, as shown here:
 * 
 * Portions Copyright  2002-2003 The Genghis Group (http://www.genghisgroup.com/).
 * 
 * 2. No substantial portion of the source code of this library may be redistributed
 * without the express written permission of the copyright holders, where
 * "substantial" is defined as enough code to be recognizably from this library. 
*/
#endregion
#region Limitations
// -Doesn't support multiple views of the same file document
#endregion

using System;
using System.IO;
using System.ComponentModel;
using System.ComponentModel.Design;
using System.Diagnostics;
using System.Windows.Forms;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;

namespace Genghis.Windows.Forms
{
    public class FileDocument : Component
    {
        public FileDocument() : this(null)
        {
        }

        public FileDocument(IContainer container)
        {
            // Integrate with container
            if( container != null ) container.Add(this);

            // Required for Designer support
            InitializeComponent();
        }

    #region Windows Form Designer generated code
        private void InitializeComponent()
        {
            this._openFileDialog = new System.Windows.Forms.OpenFileDialog();
            this._saveFileDialog = new System.Windows.Forms.SaveFileDialog();
            // 
            // _saveFileDialog
            // 
            this._saveFileDialog.FileName = "doc1";

        }
    #endregion

    #region Public Events
        public delegate void ChangedEventHandler(object sender, EventArgs e);
        public event ChangedEventHandler Changed;
    #endregion

    #region Public Properties
        // TODO: Hide from designer
        [DefaultValue(null)]
        public object Data
        {
            get
            {
                if( _master != null ) return _master.Data;
                return _data;
            }

            set 
            {
                if( _master != null ) { _master.Data = value; return; }
                _data = value;
            }
        }

        // TODO: Check for data to expose Changed event to update this automatically
        [DefaultValue(false)]
        public bool Dirty
        {
            get
            {
                if( _master != null ) return _master.Dirty;
                return _dirty;
            }

            set 
            {
                if( _master != null ) { _master.Dirty = value; return; }

                Debug.Assert(this.Site != null);
                if( _dirty == value ) return;
                _dirty = value;
                if( Changed != null ) Changed(this, EventArgs.Empty);
                SetCaption();
            }
        }

        // TODO: Hide from designer
        [DefaultValue(null)]
        public string FileName
        {
            get
            {
                if( _master != null ) return _master.FileName;
                return _fileName;
            }

            set
            {
                if( _master != null ) { _master.FileName = value; return; }

                _fileName = value;
                SetCaption();
            }
        }

        // TODO: Let non-empty settings specific override shared settings
        //    public OpenFileDialog OpenFileDialog
        //    {
        //        get { return _openFileDialog; }
        //        set { _openFileDialog = value; }
        //    }
        //
        //    public SaveFileDialog SaveFileDialog
        //    {
        //        get { return _saveFileDialog; }
        //        set { _saveFileDialog = value; }
        //    }

        [DefaultValue(null)]
        public string DefaultExt
        {
            get
            {
                if( _master != null ) return _master.DefaultExt;
                return _defaultExt;
            }

            set
            {
                if( _master != null ) { _master.DefaultExt = value; return; }

                _defaultExt = value;
                if( Empty(_defaultExt) ) return;

                // Drop leading dot
                if( _defaultExt.StartsWith(".") ) _defaultExt = _defaultExt.Substring(1);

                // TODO: Make this work in designer mode
                // Set default filter
                if( Empty(_filter) && !DesignMode )
                {
                    _filter = string.Format("{0} files (*.{1})|*.{1}|All files|*.*", Application.ProductName, _defaultExt);
                }
            }
        }

        [DefaultValue(null)]
        public string Filter
        {
            get
            {
                if( _master != null ) return _master.Filter;
                return _filter;
            }

            set
            {
                if( _master != null ) { _master.Filter = value; return; }
                _filter = value;
            }
        }

        // For use with multiple views/document. We want each form to be able
        // to have a file document to support the designer (mostly for event handling),
        // but we also want centralized access to the "master" file document and data
        // NOTE: Implementation (almost) always uses properties and not fields to take master into account
        // TODO: Hide from designer
        // TODO: Finish
        //    [DefaultValue(null)]
        //    public FileDocument MasterFileDocument
        //    {
        //        get { return _master; }
        //        set
        //        {
        //            if( _master != null ) _master.Changed -= new FileDocument.ChangedEventHandler(_master_Changed);
        //            _master = value;
        //            if( _master != null ) _master.Changed += new FileDocument.ChangedEventHandler(_master_Changed);
        //        }
        //    }

        // From Ian Griffiths (thanks, Ian!)
        public Form HostingForm
        {
            get 
            {
                if( (_host == null) && this.DesignMode ) 
                {
                    // See if we're being hosted in VS.NET (or something similar)
                    IDesignerHost designer = this.GetService(typeof(IDesignerHost)) as IDesignerHost;
                    if( designer != null ) _host = designer.RootComponent as Form;
                }

                return _host;
            }

            set
            {
                // Detach from host's Closing event
                if( _host != null ) _host.Closing -= new CancelEventHandler(_host_Closing);

                // Cache the hosting form
                _host = value;

                // Attach to host's Closing event (to handle dirty bit)
                if( _host != null ) _host.Closing += new CancelEventHandler(_host_Closing);

                // Set the host's caption
                SetCaption();
            }
        }

        [DefaultValue(null)]
        public MenuItem FileOpenMenuItem
        {
            get { return _fileOpenMenuItem; }
            set
            {
                // Unhook from the old one
                if( _fileOpenMenuItem != null ) _fileOpenMenuItem.Click -= new System.EventHandler(_fileOpenMenuItem_Click);

                // Hook to the new one
                _fileOpenMenuItem = value;
                if( _fileOpenMenuItem != null ) _fileOpenMenuItem.Click += new System.EventHandler(_fileOpenMenuItem_Click);
            }
        }

        [DefaultValue(null)]
        public MenuItem FileSaveMenuItem
        {
            get { return _fileSaveMenuItem; }
            set
            {
                // Unhook from the old one
                if( _fileSaveMenuItem != null ) _fileSaveMenuItem.Click -= new System.EventHandler(_fileSaveMenuItem_Click);

                // Hook to the new one
                _fileSaveMenuItem = value;
                if( _fileSaveMenuItem != null ) _fileSaveMenuItem.Click += new System.EventHandler(_fileSaveMenuItem_Click);
            }
        }

        [DefaultValue(null)]
        public MenuItem FileSaveAsMenuItem
        {
            get { return _fileSaveAsMenuItem; }
            set
            {
                // Unhook from the old one
                if( _fileSaveAsMenuItem != null ) _fileSaveAsMenuItem.Click -= new System.EventHandler(_fileSaveAsMenuItem_Click);

                // Hook to the new one
                _fileSaveAsMenuItem = value;
                if( _fileSaveAsMenuItem != null ) _fileSaveAsMenuItem.Click += new System.EventHandler(_fileSaveAsMenuItem_Click);
            }
        }

        [DefaultValue(null)]
        public MenuItem FileCloseMenuItem
        {
            get { return _fileCloseMenuItem; }
            set
            {
                // Unhook from the old one
                if( _fileCloseMenuItem != null ) _fileCloseMenuItem.Click -= new System.EventHandler(_fileCloseMenuItem_Click);

                // Hook to the new one
                _fileCloseMenuItem = value;
                if( _fileCloseMenuItem != null ) _fileCloseMenuItem.Click += new System.EventHandler(_fileCloseMenuItem_Click);
            }
        }

        // TODO: Toolbar buttons, too
        //    public ToolBarButton FileOpenButton
        //    {
        //    }
    #endregion

    #region Public Methods
        public bool New()
        {
            // Check to see if we can close
            if( !Close() ) return false;

            // Reset
            Data = null;
            Dirty = false;
            FileName = null;
            SetCaption();

            return true;
        }

        public bool Open()
        {
            return Open(null);
        }

        public bool Open(string newFileName)
        {
            // Check if we can close current file
            if( !Close() ) return false;

            // Get a file name
            if( Empty(newFileName) )
            {
                // Set up dialog
                _openFileDialog.DefaultExt = DefaultExt;
                _openFileDialog.Filter = Filter;

                // Get the file to open
                DialogResult res = _openFileDialog.ShowDialog(_host);
                if( res != DialogResult.OK ) return false;

                newFileName = _openFileDialog.FileName;
            }

            // Read the data
            try
            {
                Read(newFileName);
            }
            catch( Exception e )
            {
                MessageBox.Show(_host, "Can't open file: " + newFileName + "\r\n" + e.Message, Application.ProductName);
                return false;
            }

            // Clear dirty bit, cache the file name and set the caption
            Dirty = false;
            FileName = newFileName;
            SetCaption();
 
            // Return success
            return true;
        }

        public bool Save()
        {
            return Save(false);
        }

        public bool SaveAs()
        {
            return Save(true);
        }

        public bool Close()
        {
            // It's all about the dirty bit...
            if( !Dirty ) return true;

            DialogResult res = MessageBox.Show(_host, "Save changes?", Application.ProductName, MessageBoxButtons.YesNoCancel);
            switch( res )
            {
                case DialogResult.Cancel:   return false;
                case DialogResult.Yes:      return Save();
                case DialogResult.No:       return true;
                default:                    Debug.Assert(false); return false;
            }
        }

    #endregion

    #region Implementation
        bool Empty(string s) { return ((s == null) || (s == string.Empty)); }

        protected bool ShowExtension
        {
            get
            {
                // TODO: Check the shell setting
                return true;
            }
        }

        protected string DefaultFileName
        {
            get { return "Untitled" + (Empty(DefaultExt) ? "" : "." + DefaultExt); }
        }

        protected string CaptionFileName
        {
            get
            {
                string fileName = FileName;
                if( Empty(fileName) ) fileName = DefaultFileName;
                return (ShowExtension ? Path.GetFileName(fileName) : Path.GetFileNameWithoutExtension(fileName));
            }
        }

        protected void SetCaption()
        {
            if( _host == null ) return;

            // Format = "<fileName>*"
            if( _host.IsMdiChild )
            {
                _host.Text = string.Format("{0}{1}", CaptionFileName, (Dirty ? "*" : ""));
            }
                // Format = "<appName> - <fileName>*"
            else
            {
                _host.Text = string.Format("{2} - {0}{1}", CaptionFileName, (Dirty ? "*" : ""), Application.ProductName);
            }
        }

        protected bool Save(bool saveAs)
        {
            // Get the file name
            string   newFileName = FileName;
            if( saveAs || Empty(newFileName) )
            {
                // Set up dialog
                _saveFileDialog.DefaultExt = DefaultExt;
                _saveFileDialog.Filter = Filter;
                if( !Empty(newFileName) )
                {
                    _saveFileDialog.InitialDirectory = Path.GetDirectoryName(newFileName);
                    _saveFileDialog.FileName = Path.GetFileName(newFileName);
                }
                else
                {
                    _saveFileDialog.FileName = DefaultFileName;
                }

                DialogResult   res = _saveFileDialog.ShowDialog(_host);
                if( res != DialogResult.OK ) return false;
                newFileName = _saveFileDialog.FileName;
            }

            // Write the data
            try
            {
                Write(newFileName);
            }
            catch( Exception e )
            {
                MessageBox.Show(_host, "Can't save file: " + newFileName + "\r\n" + e.Message, Application.ProductName);
                return false;
            }

            // Clear the dirty bit, cache the new file name and set the caption
            Dirty = false;
            FileName = newFileName;
            SetCaption();

            // Return success
            return true;
        }

        // TODO: Allow writing to the raw stream instead of mandating formatter-based serialization
        protected void Write(string fileName)
        {
            IFormatter  formatter = new BinaryFormatter();
            Stream      stream = new FileStream(fileName, FileMode.Create, FileAccess.Write);
            formatter.Serialize(stream, Data);
            stream.Close();
        }

        // TODO: Allow reading from the raw stream instead of mandating formatter-based serialization
        protected void Read(string fileName)
        {
            IFormatter  formatter = new BinaryFormatter();
            Stream      stream = new FileStream(fileName, FileMode.Open, FileAccess.Read);
            Data = formatter.Deserialize(stream);
            stream.Close();
        }

        private object          _data = null;
        private bool            _dirty = false;
        private string          _fileName = "";
        private Form            _host = null;
        private string          _defaultExt = null;
        private string          _filter = null;
        private FileDocument    _master = null;

        private void _host_Closing(object sender, CancelEventArgs e)
        {
            if( !Close() ) e.Cancel = true;
        }

        private void _master_Changed(object sender, EventArgs e)
        {
            SetCaption();
            if( Changed != null ) Changed(sender, e);
        }

        private void _fileOpenMenuItem_Click(object sender, EventArgs e)
        {
            Open();
        }

        private void _fileSaveMenuItem_Click(object sender, EventArgs e)
        {
            Save();
        }

        private void _fileSaveAsMenuItem_Click(object sender, EventArgs e)
        {
            SaveAs();
        }

        private void _fileCloseMenuItem_Click(object sender, EventArgs e)
        {
            if( _host == null ) return;
            if( Close() ) _host.Close();
        }

        private MenuItem _fileOpenMenuItem = null;
        private MenuItem _fileSaveMenuItem = null;
        private MenuItem _fileSaveAsMenuItem = null;
        private MenuItem _fileCloseMenuItem = null;

        // TODO: Because each FileDocument gets their own dialogs,
        // they'll each have their own directories instead of a shared one.
        // Is that OK? Worth fixing?
        private System.Windows.Forms.OpenFileDialog _openFileDialog;
        private System.Windows.Forms.SaveFileDialog _saveFileDialog;
    #endregion
    }
}













