// FileSearchEngine.cs - Contributed by Mike Marshall (mmarshall@mn.rr.com)
// This class implements a simple text search engine and fires events
// when text is found, complete with file, line and column information
#region Copyright  2002-2003 The Genghis Group
/*
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from the
 * use of this software.
 * 
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, subject to the following restrictions:
 * 
 * 1. The origin of this software must not be misrepresented; you must not claim
 * that you wrote the original software. If you use this software in a product,
 * an acknowledgment in the product documentation is IsRequired, as shown here:
 * 
 * Portions Copyright  2002-2003 The Genghis Group (http://www.genghisgroup.com/).
 * 
 * 2. No substantial portion of the source code of this library may be redistributed
 * without the express written permission of the copyright holders, where
 * "substantial" is defined as enough code to be recognizably from this library. 
*/
#endregion
#region History
// 04/15/03: Initial Version
#endregion
#region Limitations/To Do
/// To do:
/// - Files with extremely long lines could pose a performance problem. Ideally
///   we should read a fixed # of bytes in rather than waiting for a \r\n    
/// - Ideally we should skip binary files, but what criteria do we use to exclude
///   binary files that 1) reliably distinguish them from text files and
///   2) Don't cause significant performance degradation?
#endregion

using System;
using System.Drawing;
using System.IO;
using System.Threading;
using System.Collections;

namespace Genghis
{
    public delegate void StartFileHandler(string strFilePath);
    public delegate void SearchHitHandler(SearchHitArgs args);
    public delegate void EndFileHandler(string strFilePath);
    public delegate void StartProcessingHandler(object sender);
    public delegate void EndProcessingHandler(object sender);
    public delegate void SearchExceptionHandler(SearchExceptionArgs args);

    public struct SearchHitArgs
    {
        public SearchHitArgs(string strPath, long ulLine, long ulColumn, string strText)
        {
            FilePath = strPath;
            LineNumber = ulLine;
            ColumnNumber = ulColumn;
            LineText = strText;
        }

        public string FilePath;
        public long LineNumber;
        public long ColumnNumber;
        public string LineText;
    }

    public struct SearchExceptionArgs
    {
        public SearchExceptionArgs(string strPath, Exception ex)
        {
            FilePath = strPath;
            SourceException = ex;
        }

        public string FilePath;
        public Exception SourceException;
    }
    
    /// <summary>
    /// Enumeration whose members represent a particular search state
    /// </summary>
    public enum SearchState
    {
        /// <summary>
        /// The search is stopped. It either never started or has completed searching
        /// files
        /// </summary>
        Stopped,

        /// <summary>
        /// The search is running
        /// </summary>
        Running,

        /// <summary>
        /// The user has requested that the search stop but existing file searches
        /// need to complete before the search is stopped completely
        /// </summary>
        UserInterrupt
    }

    /// <summary>
    /// ISearchState is a lightweight interface that provides search status information
    /// </summary>
    public interface ISearchState
    {
        /// <summary>
        /// The State property is read-only and provides the state of the search
        /// in the form of a <see cref="SearchState"/> enumeration member
        /// </summary>
        SearchState State
        {
            get;          
        }
    }


    internal class FileSearchWorker
    {
        #region Private Fields
        private SearchHitHandler m_hitHandler;
        private string m_strPath;
        private string m_strExpr;
        private bool m_bRegEx;
        private bool m_bCaseSensitive;
        private bool m_bWholeWord;
        private ISearchState m_searchState;
        #endregion

        #region Constructors
        public FileSearchWorker(string strPath, string strExpr, bool bRegEx, bool bCaseSens, bool bWholeWord, SearchHitHandler hit, ISearchState searchState)
        {
            m_strPath = strPath;
            m_strExpr = strExpr;
            m_bRegEx = bRegEx;
            m_hitHandler = hit;
            m_bCaseSensitive = bCaseSens;
            m_bWholeWord = bWholeWord;
            m_searchState = searchState;
        }
        #endregion

        #region Public Delegates
        public StartFileHandler StartFile;
        public SearchExceptionHandler SearchException;
        public EndFileHandler EndFile;
        #endregion

        #region Public Methods
        public void Worker(object oState)
        {
            StreamReader sr = null;
            try
            {
                sr = File.OpenText(m_strPath);
            }
            catch(Exception ex)
            {
                if (SearchException != null)
                    SearchException(new SearchExceptionArgs(m_strPath, ex));
            }

            if (sr != null)
            {
                System.Text.RegularExpressions.Regex regEx;

                // We may need to change the search text to accomodate the
                // search options that were selected
                string strRawExpr = Preprocess();

                try
                {
                    // RegEx was the easiest way that I could find to do a
                    // case insensitive match
                    if (m_bRegEx || !m_bCaseSensitive || m_bWholeWord)
                    {
                        System.Text.RegularExpressions.RegexOptions opts = System.Text.RegularExpressions.RegexOptions.Compiled;
                    
                        if (!m_bCaseSensitive)
                        {
                            opts |= System.Text.RegularExpressions.RegexOptions.IgnoreCase;
                        }

                        regEx = new System.Text.RegularExpressions.Regex(strRawExpr, opts);
                    }
                    else
                        regEx = null;
                }
                catch(Exception ex)
                {
                    if (SearchException != null)
                        SearchException(new SearchExceptionArgs(m_strPath, ex));
                    return;
                }

                if (StartFile != null)
                    StartFile(m_strPath);

                long lLine = 0;
                int iIndex = -1;
                while (sr.Peek() > -1)
                {
                    try
                    {
                        string strLine = sr.ReadLine();
                        lLine++;

                        if (m_bRegEx || !m_bCaseSensitive || m_bWholeWord)
                        {                        
                            System.Text.RegularExpressions.Match matchedText;
                            matchedText = regEx.Match(strLine);
                            if (matchedText.Success)
                            {
                                if (m_hitHandler != null)
                                    m_hitHandler(new SearchHitArgs(m_strPath, lLine, (long) matchedText.Index, strLine));
                            }
                        }
                        else
                        {
                            iIndex = strLine.IndexOf(m_strExpr);
                            if (iIndex > -1)
                            {
                                if (m_hitHandler != null)
                                    m_hitHandler(new SearchHitArgs(m_strPath, lLine, (long) iIndex, strLine));
                            }
                        }                    
                    }
                    catch(Exception ex)
                    {
                        if (SearchException != null)
                            SearchException(new SearchExceptionArgs(m_strPath, ex));
                    }
                }

                if (EndFile != null)
                    EndFile(m_strPath);
                sr.Close();
                
            }            
        }
        #endregion

        #region Protected Methods
        protected string Preprocess()
        {
            string strRaw = m_strExpr;

            if (m_bWholeWord)
            {
                // I'm not sure that this catches whole words that occur
                // at the beginning of lines
                strRaw = "\\W" + m_strExpr + "\\W";
            }

            return strRaw;
        }
        #endregion
    }

    /// <summary>
    /// These are search flags that can be combined to affect the behavior
    /// of the search.  They can all be set by individual properties of the
    /// <see cref="FileSearchEngine"/> class, and may be extended in the future
    /// </summary>
    [Flags]
    public enum SearchType
    {
        /// <summary>
        /// No advanced settings are set
        /// </summary>
        None = 0,
        /// <summary>
        /// Asynchronous Searches
        /// </summary>
        Asynchronous = 0x1,
        /// <summary>
        /// Search directories recursively
        /// </summary>
        Recursive = 0x2,
        /// <summary>
        /// Expression property can be a regular expression
        /// </summary>
        UseRegEx = 0x4,
        /// <summary>
        /// Match should be case sensitive
        /// </summary>
        MatchCase = 0x8,
        /// <summary>
        /// Match should only find whole word matches
        /// </summary>
        WholeWord = 0x10
    };

    /// <summary>
    /// <para>The FileSearchEngine component can be used to do a synchronous or asynchronous text
    /// search of files in a given directory and subdirectories.
    /// </para>
    /// <para>
    /// The component can be created via the .NET design environment or programmatically, and supports
    /// standard "Find in Files" type functionality (recursive search, regular expressions, etc)
    /// </para>
    /// </summary>
    /// 
    //[ToolboxBitmap(@"FileSearchEngine.gif")]
    public class FileSearchEngine : System.ComponentModel.Component, ISearchState
    {
        #region Private Fields
        private string m_strDir;
        private string m_strExts;
        private SearchType m_st;
        private string m_strExpr;
        private Queue m_queueDirs;
        private SearchState m_state;
        #endregion

        #region Constructors
        /// <summary>
        /// Instantiates an instance of a FileSearchEngine object
        /// </summary>
        public FileSearchEngine()
        {
            //InitializeComponents();
            m_queueDirs = new Queue();
            m_st = SearchType.MatchCase;
            m_state = SearchState.Stopped;
        }
        #endregion

        #region Events
        /// <summary>
        /// StartProcessing is the first event fired for a search and
        /// indicates that the search has started
        /// </summary>
        public event StartProcessingHandler StartProcessing;

        /// <summary>
        ///  StartFile is fired when the engine begins to search an
        ///  individual file
        /// </summary>
        public event StartFileHandler StartFile;

        /// <summary>
        ///  SearchHit is fired when a matching line is found
        ///  in a file during a search
        /// </summary>
        public event SearchHitHandler SearchHit;

        /// <summary>
        ///  EndFile is fired after the engine has finished
        ///  searching an individual file
        /// </summary>
        public event EndFileHandler EndFile;

        /// <summary>
        /// EndProcessing is fired when the entire search has
        /// completed
        /// </summary>
        public event EndProcessingHandler EndProcessing;

        /// <summary>
        /// SearchException is fired when an exception occurs during the
        /// course of a search
        /// </summary>
        public event SearchExceptionHandler SearchException;

        #endregion

        /// <summary>
        /// The directory where the search starts
        /// </summary>
        public string StartDirectory
        {
            get
            {
                return m_strDir;
            }
            set
            {
                m_strDir = value;
            }
        }

        /// <summary>
        /// Only files whose names match this filter string will be searched.
        /// The format of the string is the same as the one that is used with
        /// the GetFiles method of <see cref="System.IO.DirectoryInfo"/>
        /// </summary>
        public string WildCard
        {
            get
            {
                return m_strExts;
            }
            set
            {
                m_strExts = value;
            }
        }

        /// <summary>
        /// The search text.  This can be any valid string, and if the <see cref="Recursive"/>
        /// property is set to true, it can be any valid regular expression
        /// according to the grammar for <see cref="RegEx"/>
        /// </summary>
        public string Expression
        {
            get
            {
                return m_strExpr;
            }
            set
            {
                m_strExpr = value;
            }
        }

        /// <summary>
        /// Indicates whether or not the search will be done asynchronously. If
        /// set to true, <see cref="Search"/> will return immediately when called
        /// and events will be fired from a worker thread.  This is important to
        /// know when displaying the results in a Windows Forms control.  For 
        /// more Information on this see <see cref="System.Windows.Forms.Control.Invoke"/>
        /// </summary>
        public bool Asynchronous
        {
            get
            {
                return ((m_st & SearchType.Asynchronous) == SearchType.Asynchronous);
            }
            set
            {
                if (value)
                    m_st |= SearchType.Asynchronous;
                else
                    m_st &= ~SearchType.Asynchronous;
            }
        }

        /// <summary>
        /// Indicates whether or not the search will recurse subdirectories during
        /// its execution
        /// </summary>
        public bool Recursive
        {
            get
            {
                return ((m_st & SearchType.Recursive) == SearchType.Recursive);
            }
            set
            {
                if (value)
                    m_st |= SearchType.Recursive;
                else
                    m_st &= ~SearchType.Recursive;
            }
        }

        /// <summary>
        /// If set to true, the <see cref="Expression"/> property can be any valid
        /// regular expression.  For details on expression grammar see
        /// <see cref="System.Text.RegularExpressions.RegEx"/>
        /// </summary>
        public bool UseRegEx
        {
            get
            {
                return ((m_st & SearchType.UseRegEx) == SearchType.UseRegEx);
            }
            set
            {
                if (value)
                    m_st |= SearchType.UseRegEx;
                else
                {                    
                    m_st &=  (SearchType)(~((int)SearchType.UseRegEx));
                }
            }
        }

        /// <summary>
        /// Indicates whether or not to do a case sensitive search
        /// </summary>
        public bool MatchCase
        {
            get
            {
                return ((m_st & SearchType.MatchCase) == SearchType.MatchCase);
            }
            set
            {
                if (value)
                    m_st |= SearchType.MatchCase;
                else
                    m_st &=  (SearchType)(~((int)SearchType.MatchCase));
            }
        }
    
        /// <summary>
        /// Indicates whether or not only complete word matches 
        /// are found
        /// </summary>
        public bool WholeWord
        {
            get
            {
                return ((m_st & SearchType.WholeWord) == SearchType.WholeWord);
            }
            set
            {
                if (value)
                    m_st |= SearchType.WholeWord;
                else
                    m_st &= ~SearchType.WholeWord;
            }
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="strDir">A string containing the full path of the initial search directory</param>
        /// <param name="strExts">A string specifying the file filter.  See <see cref="WildCard"/></param>
        /// <param name="strExpr">A string specifying the text or expression to search for</param>
        /// <param name="st">A <see cref="SearchType"/> containing the search option flags</param>
        public void Search(string strDir, string strExts, string strExpr, SearchType st)
        {
            StartDirectory = strDir;
            WildCard = strExts;
            Expression = strExpr;
            m_st = st;

            Search();            
        }

        /// <summary>
        /// Starts the search using the current object properties
        /// </summary>
        public void Search()
        {
            if (this.Asynchronous)
            {
                ThreadStart tsHelper = new ThreadStart(this.SearchWorker);
                Thread thWorker = new Thread(tsHelper);
                thWorker.Start();
            }
            else
            {                
                SearchWorker();
            }
        }

        public void Stop()
        {
            lock(this)
            {
                m_state = SearchState.UserInterrupt;
            }
        }

        protected void SearchWorker()
        {
            if (StartProcessing != null)
                StartProcessing(this);
            
            // Protects the state variable from corruption
            lock(this)
            {
                m_state = SearchState.Running;
            }

            ExecuteSearch();
            if (EndProcessing != null)
                EndProcessing(this);

            // Protected the state variable from corruption
            lock(this)
            {
                m_state = SearchState.Stopped;
            }
        }

        private void ExecuteSearch()
        {
            if (!Directory.Exists(m_strDir))
            {
                // More detailed info necessary
                throw new FileNotFoundException();
            }
            else
            {
                // Breadth First Search.  Throw the first directory on the
                // queue and more will be added during the search if the
                // Recursive property is set.
                m_queueDirs.Enqueue(m_strDir);

                
                while (m_queueDirs.Count > 0)
                {
                    SearchDirectory((string)m_queueDirs.Dequeue(), this.Recursive);    
                }
    
            }
        }

        protected void SearchDirectory(string strDir, bool bRec)
        {
            if (Directory.Exists(strDir))
            {
                DirectoryInfo dirInfo = new DirectoryInfo(strDir);
                FileSearchWorker fsw = null;

                // Search all the files in the directory first

                FileInfo[] infos = null;

                // The directory may be inaccessible for whatever reason, so trap
                // that here
                try
                {
                    infos = dirInfo.GetFiles(m_strExts);
                }
                catch(Exception ex)
                {
                    if (SearchException != null)
                        SearchException(new Genghis.SearchExceptionArgs(strDir, ex));
                    return;
                }
                
                foreach (FileInfo fsi in infos)
                {
                    // We'll use a worker object instead of a member function
                    // just in case we want to use a ThreadPool based implementation
                    // in the future
                    if (State != SearchState.UserInterrupt)
                    {
                        fsw = new FileSearchWorker(fsi.FullName, m_strExpr, UseRegEx, MatchCase, WholeWord, SearchHit, (ISearchState) this);
                        fsw.StartFile += this.StartFile;
                        fsw.EndFile += this.EndFile;                    
                        fsw.SearchException += this.SearchException;
                        fsw.Worker(null);
                        fsw = null;
                    }
                    else
                    {
                        break;
                    }
                }
                
                if (bRec && (State != SearchState.UserInterrupt))
                {
                    // If Recursive is set, add sub-directories for processing later
                    foreach (DirectoryInfo dInfo in dirInfo.GetDirectories())
                    {
                        m_queueDirs.Enqueue(dInfo.FullName);
                    }
                }
            }
        }

       
        #region Implementation of ISearchState
        /// <summary>
        /// Returns a <see cref="SearchState"/> value indicating the current
        /// state of the search
        /// </summary>
        public Genghis.SearchState State
        {
            get
            {
                // Protects the member variable from being corrupted
                lock(this)
                {
                    return m_state;
                }
                
            }            
        }
        #endregion
    }
}
