// HandleCollector.cs: Contributed by Ethan J. Brown (ethan_j_brown@hotmail.com)
// Connects to the existing HandleCollector shared within System.Windows.Forms
// for the purpose of unifying Win32 resource management amongst the classes in that
// namespace and your own code
// Initiates a handle collection when resources hit an expense or count threshold

#region Copyright  2002-2003 The Genghis Group
/*
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from the
 * use of this software.
 * 
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, subject to the following restrictions:
 * 
 * 1. The origin of this software must not be misrepresented; you must not claim
 * that you wrote the original software. If you use this software in a product,
 * an acknowledgment in the product documentation is IsRequired, as shown here:
 * 
 * Portions Copyright  2002-2003 The Genghis Group (http://www.genghisgroup.com/).
 * 
 * 2. No substantial portion of the source code of this library may be redistributed
 * without the express written permission of the copyright holders, where
 * "substantial" is defined as enough code to be recognizably from this library. 
*/
#endregion
#region Features
/*
 * - Provides a wrapper for all of the existing methods found in the sealed HandleCollector class
 * - Additionally modifies methods to be more type-safe in regards to using the correct handle
 * type when adding/removing handles from the collector
 * - Provides additional statistics and information on existing registered types
 * - Throws exceptions if the HandleCollector is used before winforms has initialized it's own copy
*/
#endregion
#region History
/*
 * [8/2/2002]    First Release
 * 
*/
#endregion
#region TODO
/*
 * - Nice up the thrown exceptions (HandleTypeNotRegisteredException, WindowsFormsNotInitializedException) with a little extra information
 */
#endregion


using System;
using System.Windows.Forms;

namespace Genghis.Windows.Forms
{
    /// <summary>
    /// Represents the method prototype a delegate must conform to sink events when handles
    /// are added or removed from the <see cref="HandleCollector"/>
    /// </summary>
    public delegate void HandleChangeEventHandler(string handleType, System.IntPtr handleValue, int currentHandleCount) ;
    
    /// <summary>
    /// Represents an exception that can be thrown by the HandleCollector, when a type
    /// is used that has not been registered first
    /// </summary>
    public class HandleTypeNotRegisteredException : System.ApplicationException
    {
        private readonly int type;
        /// <summary>
        /// Constructs a new exception that contains the handle type that has not been registered
        /// </summary>
        /// <param name="type"></param>
        public HandleTypeNotRegisteredException(int type)
        {
            this.type = type;
            //base.Message = "There is no registered handle type corresponding to " + type.ToString();
        }

        /// <summary>
        /// The type that was not registered before use
        /// </summary>
        public int Type
        {
            get { return type; }
        }
    }


    /// <summary>
    /// Represents an exception that can be thrown by the HandleCollector, when registering
    /// handle types has been attempted before windows forms has initiated, and registered
    /// it's own types
    /// </summary>
    public class WindowsFormsNotInitializedException : System.ApplicationException
    {
        //TODO: maybe add more to this
        /// <summary>
        /// Constructs a new WindowsFormsNotInitializedException
        /// </summary>
        public WindowsFormsNotInitializedException()
        {
        }
    }


    /// <summary>
    /// Summary description for GenghisHandleCollector.
    /// </summary>
    public class HandleCollector
    {
        private static System.Reflection.Assembly winforms = null;
        private static System.Type handleCollector = null;
        private static System.Reflection.MethodInfo add = null;
        private static System.Reflection.MethodInfo registerType = null;
        private static System.Reflection.MethodInfo remove = null;
        private static System.Reflection.EventInfo handleAdded = null;
        private static System.Reflection.EventInfo handleRemoved = null;
        private static System.Reflection.FieldInfo handleTypes = null;
        private static System.Reflection.FieldInfo handleTypeCount = null;
        
        private static System.Type handleType = null;
        private static System.Reflection.FieldInfo    handleTypeName = null;
        private static System.Reflection.FieldInfo    handleTypeThreshHold = null;
        private static System.Reflection.FieldInfo    handleTypeDeltaPercent = null;

        private static System.Collections.Hashtable handleNameType = null;
        private static System.Collections.Hashtable handles = null;

        private static System.Delegate handleAddedDelegate = null;
        private static System.Delegate handleRemovedDelegate = null;

        private static bool useStrongNamedAssembly = true;
        

        /// <summary>
        /// Occurs whenever a new handle has been added to the handle collector
        /// </summary>
        public static event HandleChangeEventHandler HandleAdded = null;
        
        /// <summary>
        /// Occurs whenever an existing handle has been removed from the handle collector
        /// </summary>
        public static event HandleChangeEventHandler HandleRemoved = null;

        
        /// <summary>
        /// Our class performs some static initialization, to reflect on the <see cref="System.Windows.Forms"/> assembly
        /// and retreive method, field and event info for the HandleCollector class
        /// </summary>
        static HandleCollector()
        {
            InitializeReflection(false);
        }
        
        /// <summary>
        /// Adds the given handle to the handle collector.  The handle is added to the <see cref="System.Windows.Forms"/> sealed
        /// HandleCollector class so as to consolidate handle tracking between an external class and the internals within that namespace.
        /// The garbage collector will automatically determine when/if the handles need to be collected.
        /// </summary>
        /// <param name="handle">A <see cref="System.IntPtr"/> referencing a handle of the specified type</param>
        /// <param name="type">A <see cref="RegisteredHandleType"/> representing the type of the handle</param>
        /// <returns>The handle that was passed into the Add method</returns>
        /// <exception cref="HandleTypeNotRegisteredException">Occurs when the specified type has not been registered</exception>
        public static IntPtr Add(IntPtr handle, RegisteredHandleType type)
        {
            try
            {
                return (IntPtr)(add.Invoke(null, new object[] {handle, type.Type}));
            }
            catch (System.Reflection.TargetInvocationException)
            {
                //there's really only one problem that we could have here
                //internally, the HandleCollector doesn't check for a valid index into the array
                //in other words - our type hasn't been registered
                throw new HandleTypeNotRegisteredException(type.Type);
            }
        }

        /// <summary>
        /// Registers a new type of handle with the handle collector.
        /// </summary>
        /// <param name="typeName">String name of the handle type</param>
        /// <param name="expense">An expense in the range of 0 - 100.  0 is the least expensive handle, and 100 is the most expensive</param>
        /// <param name="initialThreshold">Number of handles that must be added to the handle collector before garbage collection is initiated</param>
        /// <returns>A <see cref="RegisteredHandleType"/> representing the handles type, that must be used with the Add and Remove methods, or null if one could not be created</returns>
        /// <exception cref="System.ArgumentOutOfRangeException">Occurrs when an expense is not between 0 and 100, or the initialThreshold is set below 0</exception>
        /// <exception cref="WindowsFormsNotInitializedException">Occurrs any time the HandleCollector is used before <see cref="System.Windows.Forms"/> has initialized it's internal HandleCollector</exception>
        /// <remarks>
        ///    Within <see cref="System.Windows.Forms"/>, the following types are already registered:
        ///    
        ///    Accelerator
        ///    Cursor
        ///    EnhancedMetaFile
        ///    Find
        ///    GDI
        ///    HDC
        ///    Icon
        ///    Kernel
        ///    Menu
        ///    Window
        ///    
        ///    To retreive the current <see cref="RegisteredHandleType"/> representing the handles type, use the <see cref="HandleCollector.GetRegisteredHandleType"/> method
        ///    or examine the <see cref="HandleCollector.RegisteredHandleTypes"/> collection
        ///    
        ///    If you attempt to add a typeName that has already been registered, the specified expense and initialThreshold values are ignored, and
        ///    the ones that are already registered are used
        ///    </remarks>
        public static RegisteredHandleType RegisterType(string typeName, int expense, int initialThreshold)
        {
            try
            {
                if ((expense < 0) || (expense > 100))
                    throw new System.ArgumentOutOfRangeException("expense", expense, "The expense must be greater than or equal to 0 and less than or equal to 100.");
                if (initialThreshold < 1)
                    throw new System.ArgumentOutOfRangeException("initialThreshold", initialThreshold, "The initialThreshHold must be greater than 0.");
                
                //if we can't find it, kick off reflecting the registered types
                if (!handleNameType.ContainsKey(typeName))
                    ReflectHandleTypeArray();

                if (0 == handleNameType.Count)
                    throw new WindowsFormsNotInitializedException();

                //shortcut and give them back the RegisteredType if its already registered
                if (handleNameType.ContainsKey(typeName))
                    return (RegisteredHandleType)handleNameType[typeName];

                //create a new handle type and store it
                int type = (int)(registerType.Invoke(null, new object[] {typeName, expense, initialThreshold}));
                RegisteredHandleType rht = new RegisteredHandleType(typeName, type, expense, initialThreshold);
                handleNameType.Add(typeName, rht);
                return rht;
            }
            catch (System.Reflection.TargetInvocationException)
            {
                //string mess = e.Message;
            }
            finally
            {}
            return null;
        }

        /// <summary>
        ///     Removes the given handle from the handle collector.  This handle will be removed from the <see cref="System.Windows.Forms"/> sealed
        /// HandleCollector class.  Therefore it will not be included in the garbage collection kick-off algorithm.
        /// </summary>
        /// <param name="handle">A <see cref="System.IntPtr"/> referencing a handle of the specified type</param>
        /// <param name="type">A <see cref="RegisteredHandleType"/> representing the type of the handle</param>
        /// <returns>The handle that was passed into the Remove method</returns>
        /// <exception cref="HandleTypeNotRegisteredException">Occurs when the specified type has not been registered</exception>
        public static IntPtr Remove(IntPtr handle, RegisteredHandleType type)
        {
            try
            {
                return (IntPtr)(remove.Invoke(null, new object[] {handle, type.Type}));
            }
            catch (System.Reflection.TargetInvocationException)
            {
                //there's really only one problem that we could have here
                //internally, the HandleCollector doesn't check for a valid index into the array
                //in other words - our type hasn't been registered
                throw new HandleTypeNotRegisteredException(type.Type);
            }
        }

        /// <summary>
        /// Occurs whenever a new handle has been added to the handle collector
        /// </summary>
        /// <param name="handleType">The string representation of the handle type</param>
        /// <param name="handleValue">The actual handle resource</param>
        /// <param name="currentHandleCount">The number of handles of that type registered</param>
        private static void OnHandleAdded(string handleType, IntPtr handleValue, int currentHandleCount) 
        {
            // When a handle is added, we increment the count based on the type
            if (handles.Contains(handleType)) 
            {
                lock (handles.SyncRoot)
                {    
                    int count = (int)handles[handleType];
                    handles[handleType] = ++count;
                }
            }
            else
            {
                lock (handles.SyncRoot)
                    handles.Add(handleType, 1);
            }

            if (!handleNameType.ContainsKey(handleType))
                ReflectHandleTypeArray();

            //invoke the delegates
            if (null != HandleAdded)
                HandleAdded(handleType, handleValue, currentHandleCount);
        }

        /// <summary>
        /// Occurs whenever a new handle has been removed from the handle collector
        /// </summary>
        /// <param name="handleType">The string representation of the handle type</param>
        /// <param name="handleValue">The actual handle resource</param>
        /// <param name="currentHandleCount">The number of handles of that type registered</param>
        private static void OnHandleRemoved(string handleType, IntPtr handleValue, int currentHandleCount) 
        {
            // When a handle is removed, we decrement the count based on the type
            if (handles.Contains(handleType)) 
            {
                lock (handles.SyncRoot)
                {
                    int count = (int)handles[handleType];
                    handles[handleType] = --count;
                }
            }
            else
            {
                lock (handles.SyncRoot)
                    handles.Add(handleType, -1);
            }

            //this handle type isn't present in our hashtable - check it out
            if (!handleNameType.ContainsKey(handleType))
                ReflectHandleTypeArray();

            //invoke the delegates
            if (null != HandleRemoved)
                HandleRemoved(handleType, handleValue, currentHandleCount);
        }

        /// <summary>
        /// Returns a <see cref="System.Collections.Hashtable"/> copy of all the currently registered handle types
        /// </summary>
        /// <exception cref="WindowsFormsNotInitializedException">Occurrs any time the HandleCollector is used before <see cref="System.Windows.Forms"/> has initialized it's internal HandleCollector</exception>
        /// <remarks>
        /// The key is the string representation of the name
        /// The value is the corresponding <see cref="RegisteredHandleType"/> type
        /// </remarks>
        public static System.Collections.Hashtable RegisteredHandleTypes
        {
            get
            {
                if (0 == handleNameType.Count)
                    ReflectHandleTypeArray();

                if (0 == handleNameType.Count)
                    throw new WindowsFormsNotInitializedException();

                return (System.Collections.Hashtable)handleNameType.Clone();
            }
        }


        /// <summary>
        /// Returns the total number of unique handle types that have been registered
        /// </summary>
        /// <exception cref="WindowsFormsNotInitializedException">Occurrs any time the HandleCollector is used before <see cref="System.Windows.Forms"/> has initialized it's internal HandleCollector</exception>
        public static int RegisteredHandleTypesCount
        {
            get
            {
                //reflect to retreive the registered type count
                int count = (int)handleTypeCount.GetValue(handleCollector);
                if (0 == count)
                    throw new WindowsFormsNotInitializedException();

                if (handleNameType.Count != count)
                    ReflectHandleTypeArray();

                return count;
            }
        }

        /// <summary>
        /// Finds if a handle has been registered, and if so, returns its <see cref="RegisteredHandleType"/>
        /// </summary>
        /// <param name="handleType">The string representation of the handle type</param>
        /// <returns>The int type corresponding to the handle name, otherwise null if it hasn't been registered,
        /// or if the internal <see cref="System.Windows.Forms"/> HandleCollector has not yet been initialized,
        /// and registered types
        /// </returns>
        public static RegisteredHandleType GetRegisteredHandleType(string handleType)
        {
            if (handleNameType.ContainsKey(handleType))
                return (RegisteredHandleType)handleNameType[handleType];
            else
            {
                ReflectHandleTypeArray();
                if (handleNameType.ContainsKey(handleType))
                    return (RegisteredHandleType)handleNameType[handleType];
            }
            
            return null;
        }

        /// <summary>
        /// Returns the total number of handles that have been registered regardless of their type
        /// </summary>
        public static int RegisteredHandleCount
        {
            get
            {
                int registered = 0;
                
                foreach (System.Collections.DictionaryEntry de in handles)
                    registered += (int)de.Value;

                return registered;
            }
        }

        /// <summary>
        /// Finds the number of handles of a given type that have been registered
        /// </summary>
        /// <param name="handleType">The string representation of a handle type</param>
        /// <returns>The number of handles registered, or 0 if there are none</returns>
        public static int GetTypeSpecificRegisteredHandleCount(string handleType)
        {
            if (handles.Contains(handleType))
                return (int)handles[handleType];
            else
                return 0;
        }

        /// <summary>
        /// Finds the number of handles of a given type that have been registered
        /// </summary>
        /// <param name="registeredHandleType"></param>
        /// <returns>The number of handles registered, or 0 if there are none</returns>
        public static int GetTypeSpecificRegisteredHandleCount(RegisteredHandleType registeredHandleType)
        {
            return GetTypeSpecificRegisteredHandleCount(registeredHandleType.Name);
        }

        /// <summary>
        /// Only set this to true if using a strongly named assembly.  Setting this to false may work around
        /// problems with loading the strongly named assembly, but might thrown unexpected exceptions.
        /// </summary>
        /// <remarks>
        /// The current strong named <see cref="System.Windows.Forms"/> assemblies supported are:
        /// 
        /// System.Windows.Forms, Version=1.0.3300.0, Culture=neutral, PublicKeyToken=b77a5c561934e089
        /// 
        /// 
        /// *** USE AT YOUR OWN RISK ***
        /// 
        /// Setting this property to false, forces an attempt to connect to a difference instance of System.Windows.Forms
        /// than those specified above, and could cause unpredictable results and/or crashes
        /// </remarks>
        public static bool UseStrongNamedAssembly
        {
            get { return useStrongNamedAssembly; }
            set
            {
                if (value != useStrongNamedAssembly)
                {
                    InitializeReflection(true);
                    useStrongNamedAssembly = value;
                }
            }
        }

        private static void InitializeReflection(bool forceReload)
        {
            lock (typeof(HandleCollector))
            {
                //load the assembly by either a strong name
                //or, give the user an option to try a partial name - which could be risky
                if ((null == winforms) || (forceReload))
                {
                    if (useStrongNamedAssembly)
                        winforms = System.Reflection.Assembly.Load("System.Windows.Forms, Version=1.0.3300.0, Culture=neutral, PublicKeyToken=b77a5c561934e089");
                    else
                        winforms = System.Reflection.Assembly.LoadWithPartialName("System.Windows.Forms");
                }
                
                //grab the HandleCollector class
                if ((null == handleCollector) || (forceReload))
                    handleCollector = winforms.GetType("System.Windows.Forms.NativeMethods+HandleCollector");

                //grab the HandleType class nested within the HandleCollector class
                if ((null == handleType) || (forceReload))
                    handleType = winforms.GetType("System.Windows.Forms.NativeMethods+HandleCollector+HandleType");
                
                //grab the name, threshHold, deltaPercent and handleCount fields from the HandleType class
                if ((null == handleTypeName) || (forceReload))
                    handleTypeName = handleType.GetField("name");
                if ((null == handleTypeThreshHold) || (forceReload))
                    handleTypeThreshHold = handleType.GetField("threshHold", System.Reflection.BindingFlags.Instance | System.Reflection.BindingFlags.NonPublic);
                if ((null == handleTypeDeltaPercent) || (forceReload))
                    handleTypeDeltaPercent = handleType.GetField("deltaPercent", System.Reflection.BindingFlags.Instance | System.Reflection.BindingFlags.NonPublic);


                //grab the Add, RegisterType, and Remove methods
                if ((null == add) || (forceReload))
                    add = handleCollector.GetMethod("Add");
                if ((null == registerType) || (forceReload))
                    registerType = handleCollector.GetMethod("RegisterType");
                if ((null == remove) || (forceReload))
                    remove = handleCollector.GetMethod("Remove");


                //grab the HandleAdded event
                if ((null == handleAdded) || (forceReload))
                {
                    //if we already have one, we must be forcing a reload
                    //which means to disconnect that event from our existing delegate type
                    if ((null != handleAdded) && (null != handleAddedDelegate))
                        handleAdded.RemoveEventHandler(typeof(HandleCollector), handleAddedDelegate);

                    handleAdded = handleCollector.GetEvent("HandleAdded");
                }
                //create a delegate from the HandleAdded event type, specifying the appropriate class method
                if ((null == handleAddedDelegate) || (forceReload))
                    handleAddedDelegate = System.Delegate.CreateDelegate(handleAdded.EventHandlerType, typeof(HandleCollector), "OnHandleAdded");
                //connect our delegate
                if ((null != handleAdded) && (null != handleAddedDelegate))
                    handleAdded.AddEventHandler(typeof(HandleCollector), handleAddedDelegate);


                //grab the HandleRemoved event
                if ((null == handleRemoved) || (forceReload))
                {
                    //if we already have one, we must be forcing a reload
                    //which means to disconnect that event from our existing delegate type
                    if ((null != handleRemoved) && (null != handleRemovedDelegate))
                        handleRemoved.RemoveEventHandler(typeof(HandleCollector), handleRemovedDelegate);

                    handleRemoved = handleCollector.GetEvent("HandleRemoved");
                }
                //create a delegate from the HandleRemoved event type, specifying the appropriate class method
                if ((null == handleRemovedDelegate) || (forceReload))
                    handleRemovedDelegate = System.Delegate.CreateDelegate(handleRemoved.EventHandlerType, typeof(HandleCollector), "OnHandleRemoved");

                //connect our delegate
                if ((null != handleRemoved) && (null != handleRemovedDelegate))
                    handleRemoved.AddEventHandler(typeof(HandleCollector), handleRemovedDelegate);

                //grab the handleTypes ArrayList field, and the handleTypeCount int field
                if ((null == handleTypes) || (forceReload))
                    handleTypes = handleCollector.GetField("handleTypes", System.Reflection.BindingFlags.Static | System.Reflection.BindingFlags.NonPublic );
                if ((null == handleTypeCount) || (forceReload))
                    handleTypeCount = handleCollector.GetField("handleTypeCount", System.Reflection.BindingFlags.Static | System.Reflection.BindingFlags.NonPublic );

                //stores our hash of 
                if ((null == handles) || (forceReload))
                    handles = new System.Collections.Hashtable();

                if ((null == handleNameType) || (forceReload))
                    handleNameType = new System.Collections.Hashtable();


                forceReload = false;
            }
        }

        private static void ReflectHandleTypeArray()
        {
            System.Array handleTypeArray = (System.Array)handleTypes.GetValue(handleCollector);
            if (null == handleTypeArray)
                return;

            string name = string.Empty;
            for (int i = 0; i < handleTypeArray.Length; ++i)
            {
                name = (string)handleTypeName.GetValue(handleTypeArray.GetValue(i));
                if ((null != name) && (string.Empty != name) && (!handleNameType.ContainsKey(name)))
                {
                    //do this outside of the lock for performance reasons
                    int threshhold = (int)handleTypeThreshHold.GetValue(handleTypeArray.GetValue(i));
                    int expense = 100 - (int)handleTypeDeltaPercent.GetValue(handleTypeArray.GetValue(i));
                    
                    lock (handleNameType.SyncRoot)
                        handleNameType.Add(name, new RegisteredHandleType(name, i + 1, expense, threshhold));
                }
            }
        }

        /// <summary>
        /// Describes a handle type that has been registered with the handle collector.
        /// Helps to ensure type-safe use of add and remove with only registered types.
        /// </summary>
        public class RegisteredHandleType
        {
            private readonly string name;
            private readonly int type;
            private readonly int expense;
            private readonly int initialThreshold;

            /// <summary>
            /// This constructor is only designed to be used by the HandleCollector
            /// </summary>
            /// <param name="name">String name of the handle type</param>
            /// <param name="type">The underlying int type </param>
            /// <param name="expense">An expense in the range of 0 - 100.  0 is the least expensive handle, and 100 is the most expensive</param>
            /// <param name="initialThreshold">Number of handles that must be added to the handle collector before garbage collection is initiated</param>
            /// <remarks>Even though this is marked internal, and could potentially be used within the library - RegisterType must be used to register and generate
            /// a valid RegisteredHandleType.
            /// </remarks>
            internal RegisteredHandleType(string name, int type, int expense, int initialThreshold)
            {
                this.name = name;
                this.type = type;
                this.expense = expense;
                this.initialThreshold = initialThreshold;
            }

            /// <summary>
            /// String name of the handle type
            /// </summary>
            public string Name 
            {
                get    { return name; }
            }
        
            /// <summary>
            /// An expense in the range of 0 - 100.  0 is the least expensive handle, and 100 is the most expensive
            /// </summary>
            public int Expense
            {
                get { return expense; }
            }

            /// <summary>
            /// Number of handles of this type that must be added to the handle collector before garbage collection is initiated
            /// </summary>
            public int InitialThreshold
            {
                get { return initialThreshold; }
            }

            /// <summary>
            /// The underlying int type used internally to identify this handle type
            /// </summary>
            public int Type
            {
                get { return type; }
            }
        }
    }
}
