// CoolMenu.cs: Contributed by Chris Burrows [chris_j_burrows@pobox.com]
// Owner-draw menu with icons (in the style of VS.NET or Windows Movie Maker).
#region Copyright 2002 The Genghis Group
///*  Copyright 2002 The Genghis Group
//    This software is provided 'as-is', without any express or implied warranty. 
//    In no event will the authors be held liable for any damages arising from the use of this software.
//    Permission is granted to anyone to use this software for any purpose, including commercial applications, 
//    subject to the following restrictions:
//    
//    The origin of this software must not be misrepresented; you must not claim that you wrote 
//    the original software. If you use this software in a product, an acknowledgment in the product 
//    documentation is required, as shown here:
//
//        Portions copyright 2002 The Genghis Group (http://www.genghisgroup.com/).
//  
//    No substantial portion of the source code of this library may be redistributed without the 
//    express written permission of the copyright holders, where "substantial" is defined as enough 
//    code to be recognizably from this library. 
//*/
#endregion Copyright
#region History
/*
 * 4/15/03 - Initial Version
 */
#endregion


using System;
using System.ComponentModel;
using System.Drawing.Design;
using System.Windows.Forms;
using System.Drawing;
using System.Collections;
namespace Genghis.Windows.Forms.Design
{

    
    /// <summary>
    /// The ImageIndexEditor is a UITypeEditor that provides design time support
    /// for displaying a drop down list of images in the property window.  It is
    /// most commonly used with an ImageIndex property so that a user can selected
    /// the image they want, rather than being forced to remember the location of
    /// the image in an image list.  
    /// 
    /// It provides exactly the same functionaly that is available on the 
    /// ImageIndex property of a button control. 
    /// 
    /// Thanks must goto Brian Ingenito(USPC.PCT.Hopewell) who pointed
    /// me in the direction of Microsoft's ImageIndexEditor which has been made private.
    /// </summary>
    /// <example>
    /// To use this image index editor you must have a property that has type of
    /// imagelist.  It doesn't have to be a property called imagelist, it just 
    /// has to return a type of imagelist.
    /// 
    /// Also, on the property that you want to display images as an image index you
    /// must add the following attributes
    /// 
    ///   [TypeConverter(typeof(ImageIndexConverter)),
    ///   Editor(typeof(Genghis.Windows.Forms.Design.ImageIndexEditor),typeof(System.Drawing.Desgin.UITypeEditor))]
    ///   
    ///</example>
    public class ImageIndexEditor:UITypeEditor
    {
        
        UITypeEditor theImageEditor;
        ImageList theImageList;
        Object currentControl;

        #region Private Properties
        /// <summary>
        /// The ImageEditor getter/setter
        /// </summary>
        /// <remarks>
        /// We need an image editor so that we can
        /// utilise it to display the images correctly
        /// </remarks>
        private UITypeEditor ImageEditor
        {
            get
            {
                return this.theImageEditor;
            }
            set
            {
                this.theImageEditor = value;
            }

        }


        /// <summary>
        /// The Image list that is assoicated with 
        /// the control that the image is associated with
        /// </summary>
        /// <remarks>
        /// The image list is where we grab the images from.
        /// </remarks>
        private ImageList ImageList
        {
            get
            {
                return this.theImageList;
            }
            set
            {
               this.theImageList = value;
            }
        }

        /// <summary>
        /// The Current control that we are working on.
        /// </summary>
        private Object Control
        {
            get
            {
                return this.currentControl;
            }
            set
            {
                if (this.currentControl != value)
                {
                    this.currentControl = value;
                }
            }
        }
        #endregion
        /// <summary>
        /// The Constructor
        /// </summary>
        public ImageIndexEditor() : base() 
        {
            //Here 
            this.ImageEditor = (UITypeEditor) TypeDescriptor.GetEditor(typeof(Image), typeof(UITypeEditor));
        }

        /// <summary>
        /// Does the component's context support painting
        /// </summary>
        /// <param name="context">The context that a component is in.  Eg the container
        /// that it is in.</param>
        /// <returns>True if the context supports painting otherwise false</returns>
        public override bool GetPaintValueSupported(ITypeDescriptorContext context) 
        {
            if (this.ImageEditor != null)
                return this.ImageEditor.GetPaintValueSupported(context);
            return false;
        }

    
        /// <summary>
        /// Overriden PaintValue method.  This is used to paint the image
        /// </summary>
        /// <param name="e">The paint event arguments</param>
        public override void PaintValue(PaintValueEventArgs e) 
        {
            Image theSelectedImage;
            
            //IF we don't have a paint value and the value that we have
            //been passed isn't null then we should look for a image in
            //the imagelist
            if (this.ImageEditor != null && e.Value != null) 
            {
                //Grab the selected passed in
                theSelectedImage = this.GetImage(e.Context, Convert.ToInt32(e.Value));
                if (theSelectedImage != null)
                {
                    PaintValueEventArgs newArguments = new PaintValueEventArgs(e.Context, theSelectedImage, e.Graphics, e.Bounds);
                    this.ImageEditor.PaintValue(newArguments);
                }
            }
        }

    
        /// <summary>
        /// Private Method used to get the image from the image list
        /// </summary>
        /// <param name="context">The context from which we are to extract the image</param>
        /// <param name="index">The index of the image of where it is to be found in
        /// the image list                  
        /// </param>
        /// <returns></returns>
        private Image GetImage(ITypeDescriptorContext context, int index) 
        {
            Image theImage = null;
            if (index >= 0) 
            {
                this.Control = context.Instance;;
                this.ImageList = this.GetImageListFromContext(this.Control);
                if (this.ImageList != null && index < this.ImageList.Images.Count)
                    theImage = this.ImageList.Images[index];
            }
        
            return theImage;
        }
    
       
        /// <summary>
        /// Function to get the image list from the given control/component.
        /// This function searchs in the current image and also it's parents
        /// if it can find any
        /// </summary>
        /// <param name="context">The control/component that we are working on</param>
        /// <returns>The image List that is assoicated with the current control,
        /// otherwise null</returns>
        private ImageList GetImageListFromContext(object control)
        {
            PropertyDescriptorCollection theControlProperties;
            PropertyDescriptor theImageListProperty;
            
            
            //Assume that the image property is null
            theImageListProperty   = null;
            //Loop through while we have a control, and we still haven't found 
            //a property of type ImageList
            while (control != null && theImageListProperty == null) 
            {
                //Grab the properties from the control that we are working on
                theControlProperties = TypeDescriptor.GetProperties(control);
                foreach (PropertyDescriptor aProperty in theControlProperties)
                {
                    //Check to see if this property has a type of image list
                    if ((typeof(ImageList).IsAssignableFrom(aProperty.PropertyType)))
                    {
                        theImageListProperty = aProperty;
                        break;
                    }
                }
                //If we haven't found a property of type imagelist then work up the tree to the
                //parent control
                if (theImageListProperty == null)
                {
                    control = theControlProperties["Parent"];
                }
            }
            //If we have found the property then assign the current imagelist property to
            //the one found
            if (theImageListProperty != null)
            {
                return (ImageList) theImageListProperty.GetValue(control);
            }
            else
            {
                return null;
            }
        }
    }
}



