// ListViewSorter.cs: Contributed by Horst Veith [hveith@intercom.at]
// A listview control that supports sorting (including those fun little triangles)
#region Copyright  2002-2003 The Genghis Group
/*
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from the
 * use of this software.
 * 
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, subject to the following restrictions:
 * 
 * 1. The origin of this software must not be misrepresented; you must not claim
 * that you wrote the original software. If you use this software in a product,
 * an acknowledgment in the product documentation is IsRequired, as shown here:
 * 
 * Portions Copyright  2002-2003 The Genghis Group (http://www.genghisgroup.com/).
 * 
 * 2. No substantial portion of the source code of this library may be redistributed
 * without the express written permission of the copyright holders, where
 * "substantial" is defined as enough code to be recognizably from this library. 
*/
#endregion
#region History
/*
 * 06/10/02 - Initial Version
 */
#endregion

using System;
using System.Collections;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.Drawing;
using System.ComponentModel;

namespace Genghis.Windows.Forms
{
    public enum ListViewCompareType
    {
        None = 0,        
        String,    
        Integer,
        Float,
        Date,
    }    
    
    /// <summary>
    /// ListViewSorter sorts a listview in reportview style.
    /// </summary>
    public class ListViewSorter : IComparer
    {
        #region Methods

        /// <summary>
        /// Sets the compare types for each column in the ListView
        /// </summary>
        /// <param name="compareTypes"></param>
        public void SetColumns(ListViewCompareType[] compareTypes)
        {
            columns = new SortColumnInfo[compareTypes.Length];
            for (int i = 0; i < compareTypes.Length; i++)
                columns[i] = new SortColumnInfo(i, compareTypes[i]);
        }

        /// <summary>
        /// Sorts the ListView
        /// </summary>
        /// <param name="listView">ListView to sort</param>
        /// <param name="column">columnindex to sort</param>
        public void SortColumn(ListView listView, int column)
        {
            SortColumn(listView, column, false);
        }

        /// <summary>
        /// Sorts the ListView
        /// </summary>
        /// <param name="listView">ListView to sort</param>
        /// <param name="column">columnindex to sort</param>
        /// <param name="addColumn">adds column to tail of already sorted columns</param>
        public void SortColumn(ListView listView, int column, bool addColumn)
        {
            // no SetColumns done before - automatic initialize each column with
            // ListViewCompareType.String
            if (columns == null)
            {
                columns = new SortColumnInfo[listView.Columns.Count];
                for (int i = 0; i < listView.Columns.Count; i++)
                    columns[i] = new SortColumnInfo(i, ListViewCompareType.String);
                
            }

            // check if column is sortable
            if (((SortColumnInfo)columns[column]).type == ListViewCompareType.None)
                return;

            // set the proper sort index
            int maxIndex = 0;
            for (int i = 0; i < columns.Length; i++)
            {
                if (!addColumn)
                {
                    if (i != column)
                        columns[i].sortIndex = SortColumnInfo.maxSortIndex;
                    else
                    {
                        columns[i].ascending = (columns[i].sortIndex != SortColumnInfo.maxSortIndex ? !columns[i].ascending : true);
                        columns[i].sortIndex = 1;
                    }
                }
                else
                {
                    if (i == column)
                        columns[i].ascending = (columns[i].sortIndex != SortColumnInfo.maxSortIndex ? !columns[i].ascending : true);
                    if (columns[i].sortIndex != SortColumnInfo.maxSortIndex && maxIndex < columns[i].sortIndex)
                        maxIndex = columns[i].sortIndex;
                }
            }

            if (columns[column].sortIndex == SortColumnInfo.maxSortIndex)
                columns[column].sortIndex = maxIndex + 1;

            // create a temporary array with all columns sorted by their sort index
            sortedColumns = new SortedList();
            foreach(SortColumnInfo si in columns)
            {
                if (si.sortIndex != SortColumnInfo.maxSortIndex)
                    sortedColumns.Add(si.sortIndex, si);
            }

            // and now perform the sorting operation
            if (listView.ListViewItemSorter != this)
                listView.ListViewItemSorter = this;
            else
                listView.Sort();

            // and set the icons
            SetIcons(listView);
        }

        public void OnListViewColumnClick(object sender, ColumnClickEventArgs e)
        {
            bool bAdd = (Control.ModifierKeys & Keys.Control) == Keys.Control;
            this.SortColumn((ListView)sender, e.Column, bAdd);
        }

        #endregion

        #region Implementation

        SortColumnInfo[]        columns = null;
        SortedList                sortedColumns = null;

        private class SortColumnInfo
        {
            public const int            maxSortIndex = 9999;

            public SortColumnInfo(int column, ListViewCompareType type)
            {
                this.column = column;
                this.type = type;
            }

            public int                    column = 0;
            public int                    sortIndex = maxSortIndex;
            public bool                    ascending = true;
            public ListViewCompareType    type = ListViewCompareType.None;
        };
        
        #endregion

        #region Graphix

        Bitmap upBM, downBM;    // the 2 bitmaps used for drawing the traingles

        [StructLayout(LayoutKind.Sequential)]
        private class HDITEM 
        {
            public int     mask; 
            public int     cxy = 0; 
            [MarshalAs(UnmanagedType.LPTStr)] 
            public string  pszText; 
            public IntPtr  hbm; 
            public int     cchTextMax; 
            public int     fmt; 
            public int     lParam = 0; 
            public int     iImage = 0;
            public int     iOrder = 0;
        };

        [DllImport("user32")]
        static extern IntPtr SendMessage(IntPtr Handle, Int32 msg, IntPtr wParam, IntPtr lParam);
        [DllImport("user32", EntryPoint="SendMessage")]
        static extern IntPtr SendMessage2(IntPtr Handle, Int32 msg, IntPtr wParam, HDITEM lParam);

        Bitmap GetBitmap(bool Ascending)
        {
            Bitmap bm = new Bitmap(8, 8);
            Graphics gfx = Graphics.FromImage(bm);

            Pen lightPen = SystemPens.ControlLightLight;
            Pen shadowPen = SystemPens.ControlDark;

            gfx.FillRectangle(SystemBrushes.ControlLight, 0, 0, 8, 8);

            if (Ascending)        
            {
                // Draw triangle pointing upwards
                gfx.DrawLine(lightPen, 0, 7, 7, 7);
                gfx.DrawLine(lightPen, 7, 7, 4, 0);
                gfx.DrawLine(shadowPen, 3, 0, 0, 7);
            }
            else
            {
                gfx.DrawLine(lightPen, 4, 7, 7, 0);
                gfx.DrawLine(shadowPen, 3, 7, 0, 0);
                gfx.DrawLine(shadowPen, 0, 0, 7, 0);
            }
            
            gfx.Dispose();

            return bm;
        }

        /// SetIcons uses LVM_GETHEADER, HDM_GEITEM and HDM_SETITEM to manipulate the header control
        /// of the underlying listview control.
        private void SetIcons(ListView listView)
        {
            Int32 LVM_GETHEADER = 0x1000 + 31;
            int HDM_GETITEM = 0x1200 + 11;
            int HDM_SETITEM = 0x1200 + 12;
            int HDI_FORMAT =  0x0004;
            int HDI_BITMAP =  0x0010;
            int HDI_TEXT = 0x0002;
            
            IntPtr hHeader = SendMessage(listView.Handle, LVM_GETHEADER, IntPtr.Zero, IntPtr.Zero);
            
            if (!hHeader.Equals(IntPtr.Zero))
            {
                if (upBM == null)
                {
                    upBM = GetBitmap(true);
                    downBM = GetBitmap(false); 
                }

                HDITEM item = new HDITEM();
                item.mask = HDI_FORMAT | HDI_BITMAP | HDI_TEXT;
                item.cchTextMax = 255;
                    
                for (int i = 0; i < listView.Columns.Count; i++)
                {
                    IntPtr result = SendMessage2(hHeader, HDM_GETITEM, new IntPtr(i), item);
                    if (result.ToInt32() > 0)
                    {
                        item.pszText = listView.Columns[i].Text;
                        item.mask = HDI_FORMAT | HDI_TEXT | HDI_BITMAP;
                        item.hbm = IntPtr.Zero;
                        
                        SortColumnInfo info = columns[i];
                        if (info.sortIndex != SortColumnInfo.maxSortIndex)
                        {
                            item.fmt |= 0x7000;
                            if (info.ascending)
                                item.hbm = upBM.GetHbitmap();
                            else
                                item.hbm = downBM.GetHbitmap();
                        }
                        else
                        {
                            item.fmt |= 0x4000;
                        }
                        SendMessage2(hHeader, HDM_SETITEM, new IntPtr(i), item);
                    }
                }
            }
        }

        #endregion

        #region ICompare interface

        int IComparer.Compare(object item1, object item2)
        {
            ListViewItem lvItem1 = item1 as ListViewItem;
            ListViewItem lvItem2 = item2 as ListViewItem;

            if (lvItem1 == null || lvItem2 == null)
                return 0; // We don't know type - so both objects are same for us

            string text1, text2;
            int n = 0;

            foreach(SortColumnInfo si in sortedColumns.Values)
            {
                text1 = si.ascending ? lvItem1.SubItems[si.column].Text : lvItem2.SubItems[si.column].Text;
                text2 = si.ascending ? lvItem2.SubItems[si.column].Text : lvItem1.SubItems[si.column].Text;
                                    
                n = CompareItems(text1, text2, si.type);
                if (n != 0)
                    return n;
            }

            return 0;
        }

        private int CompareItems(string item1, string item2, ListViewCompareType type)
        {
            if (item1.Length == 0 && item2.Length == 0) return 0;
            if (item1.Length == 0) return -1;
            if (item2.Length == 0) return 1;

            switch (type)
            {
                case ListViewCompareType.Integer:
                    return Convert.ToInt32(item1) - Convert.ToInt32(item2);
                case ListViewCompareType.Float:
                    return (Convert.ToDouble(item1) - Convert.ToDouble(item2)) > 0 ? 1 : -1;
                case ListViewCompareType.Date:
                    return DateTime.Compare(Convert.ToDateTime(item1), Convert.ToDateTime(item2));
                case ListViewCompareType.String:
                    return item1.CompareTo(item2);
            
                default:
                    return 0;
            }
        }

        #endregion
    }
}
