// MappedDrives.cs: Contributed by Ethan J. Brown (ethan_j_brown@hotmail.com)
// Used to return a collection of the drives currently mapped on the system
// and a notification system for receiving events when they change
// Can use either registry spleunking, or WMI as a basis for implementation
#region Copyright  2002-2003 The Genghis Group
/*
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from the
 * use of this software.
 * 
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, subject to the following restrictions:
 * 
 * 1. The origin of this software must not be misrepresented; you must not claim
 * that you wrote the original software. If you use this software in a product,
 * an acknowledgment in the product documentation is IsRequired, as shown here:
 * 
 * Portions Copyright  2002-2003 The Genghis Group (http://www.genghisgroup.com/).
 * 
 * 2. No substantial portion of the source code of this library may be redistributed
 * without the express written permission of the copyright holders, where
 * "substantial" is defined as enough code to be recognizably from this library. 
*/
#endregion
#region Features
/*
 * - Provides a collection of the system mapped drives - The key to the hashtable is the local drive
 *    letter and the value is the remote UNC path
 * - Provides an event thats fired when the mapped drives change, and have been reloaded
 * - A MappedDrive collection can not be instantiated - everything is done via static methods
 * - Attempts to use WMI if available, and drops down to registry use if WMI is unavailable
 * - MappedDrive collection class can not be instantiated by a non-internal class, only passed out by a class factory style method
*/
#endregion
#region History
/*
 * [7/19/2002]    First Release
 * [8/15/2002]    Fixed a minute bug in the criteria that determined that a Win32_NetworkConnection WMI class has
 * an equivalent local style name - kudos to Chris Sells for happening upon this one, immediately notifying me,
 * and doing a little research to clue me in as to the source of the problem
 * 
 * 
*/
#endregion
#region TODO
/*
 * - There appears to be a bug in the event code surrounding RegNotifyChangeKeyValue, though
 *    I'm not certain of what it is
 * - Provide Connect, and Disconnect methods that allow connecting/disconnecting mapped drives
 * - Potentially drop registry code, and only use WMI?
 */
#endregion

using System;
using System.Collections;
using System.ComponentModel;
using System.Management;
using System.Globalization;
using System.ComponentModel.Design.Serialization;
using System.Reflection;
using System.Runtime.InteropServices;
using Microsoft.Win32;

namespace Genghis.Win32
{    
    /// <summary>
    /// The delegate corresponding to the <see cref="MappedDrives.OnMappedDrivesChanged"/> event
    /// </summary>
    public delegate void MappedDrivesChangedEventHandler(object sender, System.EventArgs e);

    /// <summary>
    /// Used to load a list of mapped drives from the registry
    /// Provides an event handler to notify clients that the mapped drives have been changed
    /// </summary>
    public class MappedDrives
    {
        private const long HKEY_CURRENT_USER = 0x80000001L;
        private const long REG_NOTIFY_CHANGE_LAST_SET = 0x00000004L;
        private const long REG_NOTIFY_CHANGE_NAME = 0x00000001L;
        private const string MAPPED_DRIVE_KEY_WIN32_NT = "Network";
        private const string MAPPED_DRIVE_KEY_WIN32_WINDOWS = "Network\\Persistent";

        [DllImport("advapi32.dll", EntryPoint="RegNotifyChangeKeyValue")]
        private static extern long RegNotifyChangeKeyValue(IntPtr key, bool watchSubTree, int notifyFilter, IntPtr regEvent, bool async);

        [DllImport("advapi32.dll", EntryPoint="RegOpenKey")]
        private static extern System.IntPtr RegOpenKey(IntPtr key, string subKey, out IntPtr resultSubKey);

        [DllImport("advapi32.dll", EntryPoint="RegCloseKey")]
        private static extern System.IntPtr RegCloseKey(IntPtr key);

        private static MappedDriveMap mappedDriveMap = new MappedDriveMap();
        private static bool loadMaps = true;
        private static string currentKeyName = "";

        //our thread entry points
        private static System.Threading.ThreadStart monitorThreadStart = new System.Threading.ThreadStart(MappedDrives.MonitorRegistryKeyThread);
        private static System.Threading.Thread monitorThread = null;

        //thread status event
        private static System.Threading.AutoResetEvent threadInitializedEvent = new System.Threading.AutoResetEvent(false);
    
        //WMI state
        private static bool    usingWMI = false;
        private static bool WMIsupported = false;
        private static bool WMIchecked = false;
        private static Win32NetworkConnectionList win32NetworkConnections = null;
        private static System.Management.ManagementEventWatcher Win32_NetworkConnectionEventWatcher = null;

        /// <summary>
        /// Occurrs any time the mapped drives have been changed
        /// </summary>
        public static event MappedDrivesChangedEventHandler MappedDrivesChanged;

        private MappedDrives()
        {}

        //Finalizer
        ~MappedDrives()
        {
            //stop the WMI event from watching - if appropriate
            if (null != Win32_NetworkConnectionEventWatcher)
                Win32_NetworkConnectionEventWatcher.Stop();
        }

        /// <summary>
        ///  Overrides IDisposable.Dispose - terminates our WMI event watcher
        /// </summary>
        public void Dispose()
        {
            //stop the WMI event from watching
            if (null != Win32_NetworkConnectionEventWatcher)
                Win32_NetworkConnectionEventWatcher.Stop();        
        }


        /// <summary>
        /// Retreives a <see cref="MappedDriveMap"/> copy representing all the currently mapped drives
        /// </summary>
        /// <remarks>
        /// The key to the <see cref="MappedDriveMap"/> is the local drive letter expressed like X:\,
        /// and the value is the remote UNC path expressed like \\Server\share\file.ext
        /// </remarks>

        public static MappedDriveMap Maps
        {
            get
            {
                if (!loadMaps)
                    return mappedDriveMap.Clone();

                MappedDrives.FindOSDependentMappedDrives();
                InitializeMonitoringThread();
                return mappedDriveMap.Clone();
            }
        }

        private static void InitializeMonitoringThread()
        {
            //don't setup the terminateEvent if it's not necessary
            //nor start the registry monitoring thread
            if (!usingWMI)
            {

                lock(monitorThreadStart)
                {
                    //start up the thread that will monitor registry changes if necessary
                    if (null == monitorThread)
                    {
                        monitorThread = new System.Threading.Thread(monitorThreadStart);
                        monitorThread.IsBackground = true;
                        monitorThread.Start();

                        //wait until the monitoring thread signals us
                        MappedDrives.threadInitializedEvent.WaitOne();
                    }
                }
            }
            return;
        }

        //TODO: add these methods
        //public static bool Connect(string UNCPath, char DriveLetter)
        //{}
        //public static bool Disconnect(char DriveLetter)
        //{}

        private static void FindOSDependentMappedDrives()
        {
            if (!WMIchecked)
            {
                WMIsupported = IsWMISupported();
                WMIchecked = true;
            }

            //TODO: think about using the WMI classes exclusively instead
            //use WMI if we can
            if (WMIsupported)
                FindWMIMappedDrives();
            //load the mapped drives based on the current OS
            else if (System.PlatformID.Win32NT == Environment.OSVersion.Platform)
                FindMappedDrives(MappedDrives.MAPPED_DRIVE_KEY_WIN32_NT);
            else if (System.PlatformID.Win32Windows == Environment.OSVersion.Platform)
                FindMappedDrives(MappedDrives.MAPPED_DRIVE_KEY_WIN32_WINDOWS);
        }

        //we can use this to test the WMI waters
        private static bool IsWMISupported()
        {

            try
            {
                //this throws an exception if the class can't be found
                foreach (System.Management.ManagementObject mo in (new System.Management.ManagementObjectSearcher("SELECT * FROM Win32_NetworkConnection")).Get())
                {mo.GetHashCode();}            
            }
            catch (System.Exception)
            {
                return false;
            }


            return true;
        }


        private static void FindMappedDrives(string RegistryKey)
        {
            lock(mappedDriveMap.SyncRoot)
            {
                mappedDriveMap.Clear();
            }

            MappedDrives.currentKeyName = RegistryKey;
            RegistryKey mapsKey = Registry.CurrentUser.OpenSubKey(MappedDrives.currentKeyName);

            if (null == mapsKey)
            {
                //TODO: throw an error here? - we couldn't use WMI, and we couldn't open registry
                //but maybe there just aren't any shares
                return;
            }

            string[]    mapLocalDrives = mapsKey.GetSubKeyNames();
            string        mapUNCPath = "";

            //enumerate the mapped drive letter keys
            foreach (string mapLocalDrive in mapLocalDrives)
            {
                if (string.Empty != mapLocalDrive)
                {
                    //retreive the path value associated with this key
                    mapUNCPath = mapsKey.OpenSubKey(mapLocalDrive).GetValue("RemotePath").ToString();
                    lock(mappedDriveMap.SyncRoot)
                    {
                        mappedDriveMap.Add(mapLocalDrive + ":\\", mapUNCPath);
                    }
                }
            }

            loadMaps = false;
            return;
        }


        private static void FindWMIMappedDrives()
        {
            lock(mappedDriveMap.SyncRoot)
            {
                mappedDriveMap.Clear();
            }

            lock (typeof(MappedDrives))
            {
                if (null == win32NetworkConnections)
                {
                    win32NetworkConnections = new Win32NetworkConnectionList();

                    //attach up our event that we're exposing publicly - and start it
                    Win32_NetworkConnectionEventWatcher = new System.Management.ManagementEventWatcher();
                    Win32_NetworkConnectionEventWatcher.Query = new System.Management.EventQuery("SELECT * FROM __InstanceOperationEvent WITHIN .1 WHERE TargetInstance ISA \"Win32_NetworkConnection\"");
                    Win32_NetworkConnectionEventWatcher.EventArrived += new System.Management.EventArrivedEventHandler(MappedDrives.OnEventArrived);
                    Win32_NetworkConnectionEventWatcher.Start();
                }
                else if (0 != win32NetworkConnections.Count)
                {
                    win32NetworkConnections.Clear();
                }
            }

            System.Management.ManagementClass mapClass = new System.Management.ManagementClass("\\\\.\\ROOT\\CIMV2:Win32_NetworkConnection");
            foreach (System.Management.ManagementObject mo in mapClass.GetInstances())
            {
                NetworkConnection nc = new NetworkConnection(mo);
                
                //make sure it's a disk resource, and has a valid, non-blank LocalName
                //thanks to Chris Sells for having connected "Disk" type network connections with no local name
                if ((nc.ResourceType.ToUpper() == "Disk".ToUpper()) && (string.Empty != nc.LocalName) && (null != nc.LocalName))
                {
                    lock (win32NetworkConnections.SyncRoot)
                    {
                        win32NetworkConnections.Add(nc);
                    }
                    lock (mappedDriveMap.SyncRoot)
                    {
                        mappedDriveMap.Add(nc.LocalName + "\\", nc.RemotePath);
                    }
                }
            }
    
            usingWMI = true;
            loadMaps = false;
        
        }

        /// <summary>
        /// Occurs any time we receive a notification about mapped drives changing from the WMI event handler
        /// </summary>
        /// <param name="sender">>A <see cref="object"/> representing the originator of the event</param>
        /// <param name="e">A <see cref="System.EventArgs"/> representing additional data relevant to the event</param>
        protected static void OnEventArrived(object sender, System.Management.EventArrivedEventArgs e)
        {
            MappedDrives.FindOSDependentMappedDrives();

            //this prevents event handlers from blowing up our event generators
            try
            {
                //invoke delegates as necessary
                if (MappedDrives.MappedDrivesChanged != null)
                    MappedDrives.MappedDrivesChanged(sender, System.EventArgs.Empty);
            }
            catch (System.Exception)
            {}
        }

        /// <summary>
        /// Occurrs any time the system mapped drives have been changed
        /// </summary>
        /// <param name="sender">A <see cref="object"/> representing the originator of the event</param>
        /// <param name="e">A <see cref="System.EventArgs"/> representing additional data relevant to the event</param>
        protected static void OnMappedDrivesChanged(object sender, System.EventArgs e)
        {
            MappedDrives.FindOSDependentMappedDrives();

            //this prevents event handlers from blowing up our event generators
            try
            {
                if (MappedDrives.MappedDrivesChanged != null) 
                    // Invokes the delegates. 
                    MappedDrives.MappedDrivesChanged(sender, e);
            }
            catch (System.Exception)
            {}
        }

        private static void MonitorRegistryKeyThread()
        {
            System.IntPtr myKey = System.IntPtr.Zero;
            System.Threading.AutoResetEvent registryEvent = new System.Threading.AutoResetEvent(false);

            try
            {
                unchecked 
                {
                    RegOpenKey(new System.IntPtr((int)HKEY_CURRENT_USER), MappedDrives.currentKeyName, out myKey);
                }

                //let our main thread resume
                MappedDrives.threadInitializedEvent.Set();

                while (true)
                {
                    //HACK: it appears there's a bug here, if this isn't called each and every time
                    RegNotifyChangeKeyValue(myKey, true, (int)REG_NOTIFY_CHANGE_LAST_SET | (int)REG_NOTIFY_CHANGE_NAME, registryEvent.Handle, true);
                    //registryEvent.Reset();

                    registryEvent.WaitOne();

                    //BUG: this usually fires *twice* each time a share is added
                    //but only once when one is removed
                    OnMappedDrivesChanged(new object(), System.EventArgs.Empty);
                }
            }
            catch (System.Threading.ThreadAbortException)
            {}
            //unhandled exceptions
            catch (System.Exception)
            {}
            finally 
            {
                if (System.IntPtr.Zero != myKey)
                    RegCloseKey(myKey);
            }
        }        
    }

    
    /// <summary>
    /// A hashtable collection of mapped drive letters as keys, and their corresponding UNC paths as values
    /// Originally generated by Chris Sells CollectionGen, then modified for internal scope
    /// </summary>
    public class MappedDriveMap : IDictionary, ICollection, IEnumerable, ICloneable
    {
        protected Hashtable innerHash;
        
        #region "Constructors"
        internal  MappedDriveMap()
        {
            innerHash = new Hashtable();
        }
        internal MappedDriveMap(MappedDriveMap original)
        {
            innerHash = new Hashtable (original.innerHash);
        }
        internal MappedDriveMap(IDictionary dictionary)
        {
            innerHash = new Hashtable (dictionary);
        }

        internal MappedDriveMap(int capacity)
        {
            innerHash = new Hashtable(capacity);
        }

        internal MappedDriveMap(IDictionary dictionary, float loadFactor)
        {
            innerHash = new Hashtable(dictionary, loadFactor);
        }

        internal MappedDriveMap(IHashCodeProvider codeProvider, IComparer comparer)
        {
            innerHash = new Hashtable (codeProvider, comparer);
        }

        internal MappedDriveMap(int capacity, int loadFactor)
        {
            innerHash = new Hashtable(capacity, loadFactor);
        }

        internal MappedDriveMap(IDictionary dictionary, IHashCodeProvider codeProvider, IComparer comparer)
        {
            innerHash = new Hashtable (dictionary, codeProvider, comparer);
        }
        
        internal MappedDriveMap(int capacity, IHashCodeProvider codeProvider, IComparer comparer)
        {
            innerHash = new Hashtable (capacity, codeProvider, comparer);
        }

        internal MappedDriveMap(IDictionary dictionary, float loadFactor, IHashCodeProvider codeProvider, IComparer comparer)
        {
            innerHash = new Hashtable (dictionary, loadFactor, codeProvider, comparer);
        }

        internal MappedDriveMap(int capacity, float loadFactor, IHashCodeProvider codeProvider, IComparer comparer)
        {
            innerHash = new Hashtable (capacity, loadFactor, codeProvider, comparer);
        }

        
#endregion

        #region Implementation of IDictionary
        public System.Collections.IDictionaryEnumerator GetEnumerator()
        {
            return new MappedDriveMapEnumerator(this);
        }
        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }

        public void Remove(string key)
        {
            innerHash.Remove (key);
        }
        void IDictionary.Remove(object key)
        {
            Remove ((string)key);
        }

        public bool Contains(string key)
        {
            return innerHash.Contains(key);
        }
        bool IDictionary.Contains(object key)
        {
            return Contains((string)key);
        }

        public void Clear()
        {
            innerHash.Clear();        
        }

        public void Add(string key, string value)
        {
            innerHash.Add (key, value);
        }
        void IDictionary.Add(object key, object value)
        {
            Add ((string)key, (string)value);
        }

        public bool IsReadOnly
        {
            get
            {
                return innerHash.IsReadOnly;
            }
        }

        public string this[string key]
        {
            get
            {
                return (string) innerHash[key];
            }
            set
            {
                innerHash[key] = value;
            }
        }
        object IDictionary.this[object key]
        {
            get
            {
                return this[(string)key];
            }
            set
            {
                this[(string)key] = (string)value;
            }
        }
        
        public System.Collections.ICollection Values
        {
            get
            {
                return innerHash.Values;
            }
        }

        public System.Collections.ICollection Keys
        {
            get
            {
                return innerHash.Keys;
            }
        }

        public bool IsFixedSize
        {
            get
            {
                return innerHash.IsFixedSize;
            }
        }
        #endregion

        #region Implementation of ICollection
        public void CopyTo(System.Array array, int index)
        {
            innerHash.CopyTo (array, index);
        }

        public bool IsSynchronized
        {
            get
            {
                return innerHash.IsSynchronized;
            }
        }

        public int Count
        {
            get
            {
                return innerHash.Count;
            }
        }

        public object SyncRoot
        {
            get
            {
                return innerHash.SyncRoot;
            }
        }
        #endregion

        #region Implementation of ICloneable
        public MappedDriveMap Clone()
        {
            MappedDriveMap clone = new MappedDriveMap();
            clone.innerHash = (Hashtable) innerHash.Clone();
            
            return clone;
        }
        object ICloneable.Clone()
        {
            return Clone();
        }
        #endregion
        
        #region "HashTable Methods"
        public bool ContainsKey (string key)
        {
            return innerHash.ContainsKey(key);
        }
        public bool ContainsValue (string value)
        {
            return innerHash.ContainsValue(value);
        }
        public static MappedDriveMap Synchronized(MappedDriveMap nonSync)
        {
            MappedDriveMap sync = new MappedDriveMap();
            sync.innerHash = Hashtable.Synchronized(nonSync.innerHash);

            return sync;
        }
        #endregion

        internal Hashtable InnerHash
        {
            get
            {
                return innerHash;
            }
        }
    }
    
    public class MappedDriveMapEnumerator : IDictionaryEnumerator
    {
        private IDictionaryEnumerator innerEnumerator;
            
        internal MappedDriveMapEnumerator (MappedDriveMap enumerable)
        {
            innerEnumerator = enumerable.InnerHash.GetEnumerator();
        }

        #region Implementation of IDictionaryEnumerator
        public string Key
        {
            get
            {
                return (string)innerEnumerator.Key;
            }
        }
        object IDictionaryEnumerator.Key
        {
            get
            {
                return Key;
            }
        }


        public string Value
        {
            get
            {
                return (string)innerEnumerator.Value;
            }
        }
        object IDictionaryEnumerator.Value
        {
            get
            {
                return Value;
            }
        }

        public System.Collections.DictionaryEntry Entry
        {
            get
            {
                return innerEnumerator.Entry;
            }
        }

        #endregion

        #region Implementation of IEnumerator
        public void Reset()
        {
            innerEnumerator.Reset();
        }

        public bool MoveNext()
        {
            return innerEnumerator.MoveNext();
        }

        public object Current
        {
            get
            {
                return innerEnumerator.Current;
            }
        }
        #endregion
    }


    /// <summary>
    /// A class representing the WMI Win32_NetworkConnection class
    /// Originally generated by the VS.Net WMI Add-In, then modified for internal scope
    /// </summary>
    // Functions ShouldSerialize<PropertyName> are functions used by VS property browser to check if a particular property has to be serialized. These functions are added for all ValueType properties ( properties of type Int32, BOOL etc.. which cannot be set to null). These functions uses Is<PropertyName>Null function. These functions are also used in the TypeConverter implementation for the properties to check for NULL value of property so that a empty value can be shown in Property browser in case of Drag and Drop in Visual studio.
    // Functions Is<PropertyName>Null() are functions . These functions are to be used to check if a property is NULL.
    // Functions Reset<PropertyName> are added for Nullable Read/Write properties. These functions are used in VS designere in property browser to set a property to NULL.
    //     Every property added to the class for WMI property has attributes set to define its behaviour in Visual Studio designer and as well as to define a TypeConverter to be used.
    // DateTime Conversions are added for the class to convert DMTF date to System.DateTime and Vise-versa. Conversion from DMTF to System.DateTime conversion ignores the microseconds as System.DateTime doesn't have the microseconds part in it.
    // An Early Bound class generated for the WMI class.Win32_NetworkConnection
    internal class NetworkConnection : System.ComponentModel.Component 
    {
        
        // Private property to hold the name of WMI class which created this class.
        private static string CreatedClassName = "Win32_NetworkConnection";
        
        // Property pointing to a embeded object to get System properties of the WMI object.
        private ManagementSystemProperties PrivateSystemProperties;
        
        // Underlying LateBound WMI object.
        private System.Management.ManagementObject PrivateLateBoundObject;
        
        // Member variable to store the autocommit behaviour for the class.
        private bool AutoCommitProp = true;
        
        // Private variable to hold the embedded property representing the instance.
        private System.Management.ManagementBaseObject embeddedObj;
        
        // The current WMI object used
        private System.Management.ManagementBaseObject curObj;
        
        // Flag to indicate if t    he instance is a embedded object.
        private bool isEmbedded = false;
        
        // Below are different flavours of constructors to initialize the instance with a WMI object.
        public NetworkConnection() : 
            this(((System.Management.ManagementScope)(null)), ((System.Management.ManagementPath)(null)), ((System.Management.ObjectGetOptions)(null))) 
        {
        }
        
        public NetworkConnection(string keyName) : 
            this(((System.Management.ManagementScope)(null)), ((System.Management.ManagementPath)(new System.Management.ManagementPath(NetworkConnection.ConstructPath(keyName)))), ((System.Management.ObjectGetOptions)(null))) 
        {
        }
        
        public NetworkConnection(System.Management.ManagementScope mgmtScope, string keyName) : 
            this(((System.Management.ManagementScope)(mgmtScope)), ((System.Management.ManagementPath)(new System.Management.ManagementPath(NetworkConnection.ConstructPath(keyName)))), ((System.Management.ObjectGetOptions)(null))) 
        {
        }
        
        public NetworkConnection(System.Management.ManagementPath path, System.Management.ObjectGetOptions getOptions) : 
            this(((System.Management.ManagementScope)(null)), ((System.Management.ManagementPath)(path)), ((System.Management.ObjectGetOptions)(getOptions))) 
        {
        }
        
        public NetworkConnection(System.Management.ManagementScope mgmtScope, System.Management.ManagementPath path) : 
            this(((System.Management.ManagementScope)(mgmtScope)), ((System.Management.ManagementPath)(path)), ((System.Management.ObjectGetOptions)(null))) 
        {
        }
        
        public NetworkConnection(System.Management.ManagementPath path) : 
            this(((System.Management.ManagementScope)(null)), ((System.Management.ManagementPath)(path)), ((System.Management.ObjectGetOptions)(null))) 
        {
        }
        
        public NetworkConnection(System.Management.ManagementScope mgmtScope, System.Management.ManagementPath path, System.Management.ObjectGetOptions getOptions) 
        {
            if ((path != null)) 
            {
                if ((CheckIfProperClass(mgmtScope, path, getOptions) != true)) 
                {
                    throw new System.ArgumentException("Class name does not match.");
                }
            }
            PrivateLateBoundObject = new System.Management.ManagementObject(mgmtScope, path, getOptions);
            PrivateSystemProperties = new ManagementSystemProperties(PrivateLateBoundObject);
            curObj = PrivateLateBoundObject;
        }
        
        public NetworkConnection(System.Management.ManagementObject theObject) 
        {
            if ((CheckIfProperClass(theObject.Scope, theObject.Path, theObject.Options) == true)) 
            {
                PrivateLateBoundObject = theObject;
                PrivateSystemProperties = new ManagementSystemProperties(PrivateLateBoundObject);
                curObj = PrivateLateBoundObject;
            }
            else 
            {
                throw new System.ArgumentException("Class name does not match.");
            }
        }
        
        public NetworkConnection(System.Management.ManagementBaseObject theObject) 
        {
            if ((CheckIfProperClass(theObject) == true)) 
            {
                embeddedObj = theObject;
                PrivateSystemProperties = new ManagementSystemProperties(theObject);
                curObj = embeddedObj;
                isEmbedded = true;
            }
            else 
            {
                throw new System.ArgumentException("Class name does not match.");
            }
        }
        
        // Property returns the namespace of the WMI class.
        [Browsable(true)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public string OriginatingNamespace 
        {
            get 
            {
                return "ROOT\\CIMV2";
            }
        }
        
        [Browsable(true)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public string ManagementClassName 
        {
            get 
            {
                string strRet = CreatedClassName;
                if ((curObj != null)) 
                {
                    if ((curObj.ClassPath != null)) 
                    {
                        strRet = ((string)(curObj["__CLASS"]));
                        if (((strRet == null) 
                            || (strRet == System.String.Empty))) 
                        {
                            strRet = CreatedClassName;
                        }
                    }
                }
                return strRet;
            }
        }
        
        // Property pointing to a embeded object to get System properties of the WMI object.
        [Browsable(true)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public ManagementSystemProperties SystemProperties 
        {
            get 
            {
                return PrivateSystemProperties;
            }
        }
        
        // Property returning the underlying LateBound object.
        [Browsable(false)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public System.Management.ManagementBaseObject LateBoundObject 
        {
            get 
            {
                return curObj;
            }
        }
        
        // ManagementScope of the object.
        [Browsable(true)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public System.Management.ManagementScope Scope 
        {
            get 
            {
                if ((isEmbedded == false)) 
                {
                    return PrivateLateBoundObject.Scope;
                }
                else 
                {
                    return null;
                }
            }
            set 
            {
                if ((isEmbedded == false)) 
                {
                    PrivateLateBoundObject.Scope = value;
                }
            }
        }
        
        // Property to show the autocommit behaviour for the WMI object. If this is true then WMI object is saved to WMI then for change in every property (ie. Put is called after modification of a property).
        [Browsable(false)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public bool AutoCommit 
        {
            get 
            {
                return AutoCommitProp;
            }
            set 
            {
                AutoCommitProp = value;
            }
        }
        
        // The ManagementPath of the underlying WMI object.
        [Browsable(true)]
        public System.Management.ManagementPath Path 
        {
            get 
            {
                if ((isEmbedded == false)) 
                {
                    return PrivateLateBoundObject.Path;
                }
                else 
                {
                    return null;
                }
            }
            set 
            {
                if ((isEmbedded == false)) 
                {
                    if ((CheckIfProperClass(null, value, null) != true)) 
                    {
                        throw new System.ArgumentException("Class name does not match.");
                    }
                    PrivateLateBoundObject.Path = value;
                }
            }
        }
        
        [Browsable(false)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public bool IsAccessMaskNull 
        {
            get 
            {
                if ((curObj["AccessMask"] == null)) 
                {
                    return true;
                }
                else 
                {
                    return false;
                }
            }
        }
        
        [Browsable(true)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        [Description("The AccessMask property is a bit array representing the access rights as set remo" +
             "tely on the network share held by the user or group on whose behalf the instance" +
             " is returned. This property is only supported under Windows NT and Windows 2000." +
             " ")]
        [TypeConverter(typeof(WMIValueTypeConverter))]
        public AccessMaskValues AccessMask 
        {
            get 
            {
                if ((curObj["AccessMask"] == null)) 
                {
                    return ((AccessMaskValues)(System.Convert.ToInt32(0)));
                }
                return ((AccessMaskValues)(Convert.ToInt32(curObj["AccessMask"])));
            }
        }
        
        [Browsable(true)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        [Description("The Caption property is a short textual description (one-line string) of the obje" +
             "ct.")]
        public string Caption 
        {
            get 
            {
                return ((string)(curObj["Caption"]));
            }
        }
        
        [Browsable(true)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        [Description("The Comment property contains a comment supplied by the network provider.")]
        public string Comment 
        {
            get 
            {
                return ((string)(curObj["Comment"]));
            }
        }
        
        [Browsable(true)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        [Description("The ConnectionState property indicates the current state of the network connectio" +
             "n.")]
        public string ConnectionState 
        {
            get 
            {
                return ((string)(curObj["ConnectionState"]));
            }
        }
        
        [Browsable(true)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        [Description("The ConnectionType property indicates the persistence type of the connection used" +
             " for connecting to the network.\nExample: Permanent.")]
        public string ConnectionType 
        {
            get 
            {
                return ((string)(curObj["ConnectionType"]));
            }
        }
        
        [Browsable(true)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        [Description("The Description property provides a textual description of the object. ")]
        public string Description 
        {
            get 
            {
                return ((string)(curObj["Description"]));
            }
        }
        
        [Browsable(true)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        [Description("The DisplayType property indicates how the network object should be displayed in " +
             "a network browsing user interface.\nExample: Generic.")]
        public string DisplayType 
        {
            get 
            {
                return ((string)(curObj["DisplayType"]));
            }
        }
        
        [Browsable(false)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public bool IsInstallDateNull 
        {
            get 
            {
                if ((curObj["InstallDate"] == null)) 
                {
                    return true;
                }
                else 
                {
                    return false;
                }
            }
        }
        
        [Browsable(true)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        [Description("The InstallDate property is datetime value indicating when the object was install" +
             "ed. A lack of a value does not indicate that the object is not installed.")]
        [TypeConverter(typeof(WMIValueTypeConverter))]
        public System.DateTime InstallDate 
        {
            get 
            {
                if ((curObj["InstallDate"] != null)) 
                {
                    return ToDateTime(((string)(curObj["InstallDate"])));
                }
                else 
                {
                    return ToDateTime(null);
                }
            }
        }
        
        [Browsable(true)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        [Description("The LocalName property indicates the local name of the connected network device.\n" +
             "Example: c:\\public.")]
        public string LocalName 
        {
            get 
            {
                return ((string)(curObj["LocalName"]));
            }
        }
        
        [Browsable(true)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        [Description("The Name property indicates the name of the current network connection. It is the" +
             " combination of the value in the RemoteName property and the value in the LocalN" +
             "ame property.\nExample: \\\\NTRELEASE (c:\\public).")]
        public string Name 
        {
            get 
            {
                return ((string)(curObj["Name"]));
            }
        }
        
        [Browsable(false)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public bool IsPersistentNull 
        {
            get 
            {
                if ((curObj["Persistent"] == null)) 
                {
                    return true;
                }
                else 
                {
                    return false;
                }
            }
        }
        
        [Browsable(true)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        [Description("The Persistent property determines whether this connection will be reconnected au" +
             "tomatically by the operating system on the next logon.\nValues: TRUE or FALSE. A " +
             "value of TRUE indicates the network connection will be automatically connected.")]
        [TypeConverter(typeof(WMIValueTypeConverter))]
        public bool Persistent 
        {
            get 
            {
                if ((curObj["Persistent"] == null)) 
                {
                    return System.Convert.ToBoolean(0);
                }
                return ((bool)(curObj["Persistent"]));
            }
        }
        
        [Browsable(true)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        [Description("The ProviderName property contains the name of the provider that owns the resourc" +
             "e. This property can be NULL if the provider name is unknown.")]
        public string ProviderName 
        {
            get 
            {
                return ((string)(curObj["ProviderName"]));
            }
        }
        
        [Browsable(true)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        [Description(@"The RemoteName property contains the remote network resource name for a network resource. For a current or persistent connection, RemoteName contains the network name associated with the name of the value in the LocalName property. The name in RemoteName must follow the network provider's naming conventions. 
Example: \\NTRELEASE.")]
        public string RemoteName 
        {
            get 
            {
                return ((string)(curObj["RemoteName"]));
            }
        }
        
        [Browsable(true)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        [Description("The RemotePath property contains the full path to the network resource.\nExample: " +
             "\\\\infosrv1\\public")]
        public string RemotePath 
        {
            get 
            {
                return ((string)(curObj["RemotePath"]));
            }
        }
        
        [Browsable(true)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        [Description("The ResourceType property identifies the type of resource to enumerate or connect" +
             " to.")]
        public string ResourceType 
        {
            get 
            {
                return ((string)(curObj["ResourceType"]));
            }
        }
        
        [Browsable(true)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        [Description(@"The Status property is a string indicating the current status of the object. Various operational and non-operational statuses can be defined. Operational statuses are ""OK"", ""Degraded"" and ""Pred Fail"". ""Pred Fail"" indicates that an element may be functioning properly but predicting a failure in the near future. An example is a SMART-enabled hard drive. Non-operational statuses can also be specified. These are ""Error"", ""Starting"", ""Stopping"" and ""Service"". The latter, ""Service"", could apply during mirror-resilvering of a disk, reload of a user permissions list, or other administrative work. Not all such work is on-line, yet the managed element is neither ""OK"" nor in one of the other states.")]
        public string Status 
        {
            get 
            {
                return ((string)(curObj["Status"]));
            }
        }
        
        [Browsable(true)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        [Description("The UserName property contains the user name or the default user name used to est" +
             "ablish a network connection. \nExample: SYSTEM.")]
        public string UserName 
        {
            get 
            {
                return ((string)(curObj["UserName"]));
            }
        }
        
        private bool CheckIfProperClass(System.Management.ManagementScope mgmtScope, System.Management.ManagementPath path, System.Management.ObjectGetOptions OptionsParam) 
        {
            if (((path != null) 
                && (System.String.Compare(path.ClassName, ManagementClassName, true) == 0))) 
            {
                return true;
            }
            else 
            {
                return CheckIfProperClass(new System.Management.ManagementObject(mgmtScope, path, OptionsParam));
            }
        }
        
        private bool CheckIfProperClass(System.Management.ManagementBaseObject theObj) 
        {
            if (((theObj != null) 
                && (System.String.Compare(((string)(theObj["__CLASS"])), ManagementClassName, true) == 0))) 
            {
                return true;
            }
            else 
            {
                System.Array parentClasses = ((System.Array)(theObj["__DERIVATION"]));
                if ((parentClasses != null)) 
                {
                    Int32 count = 0;
                    for (count = 0; (count < parentClasses.Length); count = (count + 1)) 
                    {
                        if ((System.String.Compare(((string)(parentClasses.GetValue(count))), ManagementClassName, true) == 0)) 
                        {
                            return true;
                        }
                    }
                }
            }
            return false;
        }
        
        private bool ShouldSerializeAccessMask() 
        {
            if ((IsAccessMaskNull == false)) 
            {
                return true;
            }
            return false;
        }
        
        static System.DateTime ToDateTime(string dmtfDate) 
        {
            int year = System.DateTime.Now.Year;
            int month = 1;
            int day = 1;
            int hour = 0;
            int minute = 0;
            int second = 0;
            int millisec = 0;
            string dmtf = dmtfDate;
            string tempString = System.String.Empty;
            if (((System.String.Empty == dmtf) 
                || (dmtf == null))) 
            {
                return System.DateTime.MinValue;
            }
            if ((dmtf.Length != 25)) 
            {
                return System.DateTime.MinValue;
            }
            tempString = dmtf.Substring(0, 4);
            if (("****" != tempString)) 
            {
                year = System.Int32.Parse(tempString);
            }
            tempString = dmtf.Substring(4, 2);
            if (("**" != tempString)) 
            {
                month = System.Int32.Parse(tempString);
            }
            tempString = dmtf.Substring(6, 2);
            if (("**" != tempString)) 
            {
                day = System.Int32.Parse(tempString);
            }
            tempString = dmtf.Substring(8, 2);
            if (("**" != tempString)) 
            {
                hour = System.Int32.Parse(tempString);
            }
            tempString = dmtf.Substring(10, 2);
            if (("**" != tempString)) 
            {
                minute = System.Int32.Parse(tempString);
            }
            tempString = dmtf.Substring(12, 2);
            if (("**" != tempString)) 
            {
                second = System.Int32.Parse(tempString);
            }
            tempString = dmtf.Substring(15, 3);
            if (("***" != tempString)) 
            {
                millisec = System.Int32.Parse(tempString);
            }
            System.DateTime dateRet = new System.DateTime(year, month, day, hour, minute, second, millisec);
            return dateRet;
        }
        
        static string ToDMTFTime(System.DateTime dateParam) 
        {
            string tempString = dateParam.Year.ToString();
            tempString = (tempString + dateParam.Month.ToString().PadLeft(2, '0'));
            tempString = (tempString + dateParam.Day.ToString().PadLeft(2, '0'));
            tempString = (tempString + dateParam.Hour.ToString().PadLeft(2, '0'));
            tempString = (tempString + dateParam.Minute.ToString().PadLeft(2, '0'));
            tempString = (tempString + dateParam.Second.ToString().PadLeft(2, '0'));
            tempString = (tempString + ".");
            tempString = (tempString + dateParam.Millisecond.ToString().PadLeft(3, '0'));
            tempString = (tempString + "000");
            System.TimeZone curZone = System.TimeZone.CurrentTimeZone;
            System.TimeSpan tickOffset = curZone.GetUtcOffset(dateParam);
            if ((tickOffset.Ticks >= 0)) 
            {
                tempString = (tempString + "+");
                tempString = (tempString + ((tickOffset.Ticks / System.TimeSpan.TicksPerMinute)).ToString());
            }
            else 
            {
                tempString = (tempString + "-");
                tempString = (tempString + ((tickOffset.Ticks / System.TimeSpan.TicksPerMinute)).ToString().Substring(1, 3));
            }
            return tempString;
        }
        
        private bool ShouldSerializeInstallDate() 
        {
            if ((IsInstallDateNull == false)) 
            {
                return true;
            }
            return false;
        }
        
        private bool ShouldSerializePersistent() 
        {
            if ((IsPersistentNull == false)) 
            {
                return true;
            }
            return false;
        }
        
        [Browsable(true)]
        public void CommitObject() 
        {
            if ((isEmbedded == false)) 
            {
                PrivateLateBoundObject.Put();
            }
        }
        
        private static string ConstructPath(string keyName) 
        {
            string strPath = "ROOT\\CIMV2:Win32_NetworkConnection";
            strPath = (strPath 
                + (".Name=" 
                + ("\"" 
                + (keyName + "\""))));
            return strPath;
        }
        
        // Different flavours of GetInstances() help in enumerating instances of the WMI class.
        public static NetworkConnectionCollection GetInstances() 
        {
            return GetInstances(((System.Management.ManagementScope)(null)), ((System.Management.EnumerationOptions)(null)));
        }

        private void InitializeComponent()
        {}

        
        public static NetworkConnectionCollection GetInstances(string condition) 
        {
            return GetInstances(null, condition, null);
        }
        
        public static NetworkConnectionCollection GetInstances(System.String [] selectedProperties) 
        {
            return GetInstances(null, null, selectedProperties);
        }
        
        public static NetworkConnectionCollection GetInstances(string condition, System.String [] selectedProperties) 
        {
            return GetInstances(null, condition, selectedProperties);
        }
        
        public static NetworkConnectionCollection GetInstances(System.Management.ManagementScope mgmtScope, System.Management.EnumerationOptions enumOptions) 
        {
            if ((mgmtScope == null)) 
            {
                mgmtScope = new System.Management.ManagementScope();
                mgmtScope.Path.NamespacePath = "root\\CIMV2";
            }
            System.Management.ManagementPath pathObj = new System.Management.ManagementPath();
            pathObj.ClassName = "Win32_NetworkConnection";
            pathObj.NamespacePath = "root\\CIMV2";
            System.Management.ManagementClass clsObject = new System.Management.ManagementClass(mgmtScope, pathObj, null);
            if ((enumOptions == null)) 
            {
                enumOptions = new System.Management.EnumerationOptions();
                enumOptions.EnsureLocatable = true;
            }
            return new NetworkConnectionCollection(clsObject.GetInstances(enumOptions));
        }
        
        public static NetworkConnectionCollection GetInstances(System.Management.ManagementScope mgmtScope, string condition) 
        {
            return GetInstances(mgmtScope, condition, null);
        }
        
        public static NetworkConnectionCollection GetInstances(System.Management.ManagementScope mgmtScope, System.String [] selectedProperties) 
        {
            return GetInstances(mgmtScope, null, selectedProperties);
        }
        
        public static NetworkConnectionCollection GetInstances(System.Management.ManagementScope mgmtScope, string condition, System.String [] selectedProperties) 
        {
            if ((mgmtScope == null)) 
            {
                mgmtScope = new System.Management.ManagementScope();
                mgmtScope.Path.NamespacePath = "root\\CIMV2";
            }
            System.Management.ManagementObjectSearcher ObjectSearcher = new System.Management.ManagementObjectSearcher(mgmtScope, new SelectQuery("Win32_NetworkConnection", condition, selectedProperties));
            System.Management.EnumerationOptions enumOptions = new System.Management.EnumerationOptions();
            enumOptions.EnsureLocatable = true;
            ObjectSearcher.Options = enumOptions;
            return new NetworkConnectionCollection(ObjectSearcher.Get());
        }
        
        public enum AccessMaskValues : int 
        {
            
            FILE_READ_DATA_file_or_FILE_LIST_DIRECTORY_directory_ = 1,
            
            FILE_WRITE_DATA_file_or_FILE_ADD_FILE_directory_ = 2,
            
            FILE_APPEND_DATA_file_or_FILE_ADD_SUBDIRECTORY_directory_ = 4,
            
            FILE_READ_EA = 8,
            
            FILE_WRITE_EA = 16,
            
            FILE_EXECUTE_file_or_FILE_TRAVERSE_directory_ = 32,
            
            FILE_DELETE_CHILD_directory_ = 64,
            
            FILE_READ_ATTRIBUTES = 128,
            
            FILE_WRITE_ATTRIBUTES = 256,
            
            DELETE = 512,
            
            READ_CONTROL = 1024,
            
            WRITE_DAC = 2048,
            
            WRITE_OWNER = 4096,
            
            SYNCHRONIZE = 8192,
        }
        
        // Enumerator implementation for enumerating instances of the class.
        public class NetworkConnectionCollection : object, ICollection 
        {
            
            private ManagementObjectCollection ObjectCollection;
            
            public NetworkConnectionCollection(ManagementObjectCollection objCollection) 
            {
                ObjectCollection = objCollection;
            }
            
            public int Count 
            {
                get 
                {
                    return ObjectCollection.Count;
                }
            }
            
            public bool IsSynchronized 
            {
                get 
                {
                    return ObjectCollection.IsSynchronized;
                }
            }
            
            public object SyncRoot 
            {
                get 
                {
                    return this;
                }
            }
            
            public void CopyTo(System.Array array, int index) 
            {
                ObjectCollection.CopyTo(array, index);
                int nCtr;
                for (nCtr = 0; (nCtr < array.Length); nCtr = (nCtr + 1)) 
                {
                    array.SetValue(new NetworkConnection(((System.Management.ManagementObject)(array.GetValue(nCtr)))), nCtr);
                }
            }
            
            public System.Collections.IEnumerator GetEnumerator() 
            {
                return new NetworkConnectionEnumerator(ObjectCollection.GetEnumerator());
            }
            
            public class NetworkConnectionEnumerator : object, System.Collections.IEnumerator 
            {
                
                private ManagementObjectCollection.ManagementObjectEnumerator ObjectEnumerator;
                
                public NetworkConnectionEnumerator(ManagementObjectCollection.ManagementObjectEnumerator objEnum) 
                {
                    ObjectEnumerator = objEnum;
                }
                
                public object Current 
                {
                    get 
                    {
                        return new NetworkConnection(((System.Management.ManagementObject)(ObjectEnumerator.Current)));
                    }
                }
                
                public bool MoveNext() 
                {
                    return ObjectEnumerator.MoveNext();
                }
                
                public void Reset() 
                {
                    ObjectEnumerator.Reset();
                }
            }
        }
        
        // TypeConverter to handle null values for ValueType properties
        public class WMIValueTypeConverter : TypeConverter 
        {
            
            private TypeConverter baseConverter;
            
            public WMIValueTypeConverter(System.Type baseType) 
            {
                baseConverter = TypeDescriptor.GetConverter(baseType);
            }
            
            public override bool CanConvertFrom(System.ComponentModel.ITypeDescriptorContext context, System.Type srcType) 
            {
                return baseConverter.CanConvertFrom(context, srcType);
            }
            
            public override bool CanConvertTo(System.ComponentModel.ITypeDescriptorContext context, System.Type destinationType) 
            {
                return baseConverter.CanConvertTo(context, destinationType);
            }
            
            public override object ConvertFrom(System.ComponentModel.ITypeDescriptorContext context, System.Globalization.CultureInfo culture, object value) 
            {
                return baseConverter.ConvertFrom(context, culture, value);
            }
            
            public override object CreateInstance(System.ComponentModel.ITypeDescriptorContext context, System.Collections.IDictionary dictionary) 
            {
                return baseConverter.CreateInstance(context, dictionary);
            }
            
            public override bool GetCreateInstanceSupported(System.ComponentModel.ITypeDescriptorContext context) 
            {
                return baseConverter.GetCreateInstanceSupported(context);
            }
            
            public override PropertyDescriptorCollection GetProperties(System.ComponentModel.ITypeDescriptorContext context, object value, System.Attribute[] attributeVar) 
            {
                return baseConverter.GetProperties(context, value, attributeVar);
            }
            
            public override bool GetPropertiesSupported(System.ComponentModel.ITypeDescriptorContext context) 
            {
                return baseConverter.GetPropertiesSupported(context);
            }
            
            public override System.ComponentModel.TypeConverter.StandardValuesCollection GetStandardValues(System.ComponentModel.ITypeDescriptorContext context) 
            {
                return baseConverter.GetStandardValues(context);
            }
            
            public override bool GetStandardValuesExclusive(System.ComponentModel.ITypeDescriptorContext context) 
            {
                return baseConverter.GetStandardValuesExclusive(context);
            }
            
            public override bool GetStandardValuesSupported(System.ComponentModel.ITypeDescriptorContext context) 
            {
                return baseConverter.GetStandardValuesSupported(context);
            }
            
            public override object ConvertTo(System.ComponentModel.ITypeDescriptorContext context, System.Globalization.CultureInfo culture, object value, System.Type destinationType) 
            {
                if ((context != null)) 
                {
                    if ((context.PropertyDescriptor.ShouldSerializeValue(context.Instance) == false)) 
                    {
                        return "";
                    }
                }
                return baseConverter.ConvertTo(context, culture, value, destinationType);
            }
        }
        
        // Embedded class to represent WMI system Properties.
        [TypeConverter(typeof(System.ComponentModel.ExpandableObjectConverter))]
            public class ManagementSystemProperties 
        {
            
            private System.Management.ManagementBaseObject PrivateLateBoundObject;
            
            public ManagementSystemProperties(System.Management.ManagementBaseObject ManagedObject) 
            {
                PrivateLateBoundObject = ManagedObject;
            }
            
            [Browsable(true)]
            public int GENUS 
            {
                get 
                {
                    return ((int)(PrivateLateBoundObject["__GENUS"]));
                }
            }
            
            [Browsable(true)]
            public string CLASS 
            {
                get 
                {
                    return ((string)(PrivateLateBoundObject["__CLASS"]));
                }
            }
            
            [Browsable(true)]
            public string SUPERCLASS 
            {
                get 
                {
                    return ((string)(PrivateLateBoundObject["__SUPERCLASS"]));
                }
            }
            
            [Browsable(true)]
            public string DYNASTY 
            {
                get 
                {
                    return ((string)(PrivateLateBoundObject["__DYNASTY"]));
                }
            }
            
            [Browsable(true)]
            public string RELPATH 
            {
                get 
                {
                    return ((string)(PrivateLateBoundObject["__RELPATH"]));
                }
            }
            
            [Browsable(true)]
            public int PROPERTY_COUNT 
            {
                get 
                {
                    return ((int)(PrivateLateBoundObject["__PROPERTY_COUNT"]));
                }
            }
            
            [Browsable(true)]
            public string[] DERIVATION 
            {
                get 
                {
                    return ((string[])(PrivateLateBoundObject["__DERIVATION"]));
                }
            }
            
            [Browsable(true)]
            public string SERVER 
            {
                get 
                {
                    return ((string)(PrivateLateBoundObject["__SERVER"]));
                }
            }
            
            [Browsable(true)]
            public string NAMESPACE 
            {
                get 
                {
                    return ((string)(PrivateLateBoundObject["__NAMESPACE"]));
                }
            }
            
            [Browsable(true)]
            public string PATH 
            {
                get 
                {
                    return ((string)(PrivateLateBoundObject["__PATH"]));
                }
            }
        }

        protected override void Dispose(bool disposing)
        {
            if( disposing )
            {
            }
            base.Dispose( disposing );        
        }
    }


    /// <summary>
    ///        A strongly-typed collection of WMI <see cref="NetworkConnection"/> generated by the VS.Net WMI Add-In
    ///        Originally generated by Chris Sells CollectionGen, then modified for internal scope
    /// </summary>
    [Serializable]
    internal class Win32NetworkConnectionList : ICollection, IList, IEnumerable, ICloneable
    {
        #region Interfaces
        /// <summary>
        ///        Supports type-safe iteration over a <see cref="Win32NetworkConnectionList"/>.
        /// </summary>
        public interface IWin32NetworkConnectionListEnumerator
        {
            /// <summary>
            ///        Gets the current element in the collection.
            /// </summary>
            NetworkConnection Current {get;}

            /// <summary>
            ///        Advances the enumerator to the next element in the collection.
            /// </summary>
            /// <exception cref="InvalidOperationException">
            ///        The collection was modified after the enumerator was created.
            /// </exception>
            /// <returns>
            ///        <c>true</c> if the enumerator was successfully advanced to the next element; 
            ///        <c>false</c> if the enumerator has passed the end of the collection.
            /// </returns>
            bool MoveNext();

            /// <summary>
            ///        Sets the enumerator to its initial position, before the first element in the collection.
            /// </summary>
            void Reset();
        }
        #endregion

        private const int DEFAULT_CAPACITY = 16;

        #region Implementation (data)
        private NetworkConnection[] m_array;
        private int m_count = 0;
        private int m_version = 0;
        #endregion
    
        #region Static Wrappers
        /// <summary>
        ///        Creates a synchronized (thread-safe) wrapper for a 
        ///     <c>Win32NetworkConnectionList</c> instance.
        /// </summary>
        /// <returns>
        ///     An <c>Win32NetworkConnectionList</c> wrapper that is synchronized (thread-safe).
        /// </returns>
        public static Win32NetworkConnectionList Synchronized(Win32NetworkConnectionList list)
        {
            if(list==null)
                throw new ArgumentNullException("list");
            return new SyncWin32NetworkConnectionList(list);
        }
        
        /// <summary>
        ///        Creates a read-only wrapper for a 
        ///     <c>Win32NetworkConnectionList</c> instance.
        /// </summary>
        /// <returns>
        ///     An <c>Win32NetworkConnectionList</c> wrapper that is read-only.
        /// </returns>
        public static Win32NetworkConnectionList ReadOnly(Win32NetworkConnectionList list)
        {
            if(list==null)
                throw new ArgumentNullException("list");
            return new ReadOnlyWin32NetworkConnectionList(list);
        }
        #endregion

        #region Construction
        /// <summary>
        ///        Initializes a new instance of the <c>Win32NetworkConnectionList</c> class
        ///        that is empty and has the default initial capacity.
        /// </summary>
        public Win32NetworkConnectionList()
        {
            m_array = new NetworkConnection[DEFAULT_CAPACITY];
        }
        
        /// <summary>
        ///        Initializes a new instance of the <c>Win32NetworkConnectionList</c> class
        ///        that has the specified initial capacity.
        /// </summary>
        /// <param name="capacity">
        ///        The number of elements that the new <c>Win32NetworkConnectionList</c> is initially capable of storing.
        ///    </param>
        public Win32NetworkConnectionList(int capacity)
        {
            m_array = new NetworkConnection[capacity];
        }

        /// <summary>
        ///        Initializes a new instance of the <c>Win32NetworkConnectionList</c> class
        ///        that contains elements copied from the specified <c>Win32NetworkConnectionList</c>.
        /// </summary>
        /// <param name="c">The <c>Win32NetworkConnectionList</c> whose elements are copied to the new collection.</param>
        public Win32NetworkConnectionList(Win32NetworkConnectionList c)
        {
            m_array = new NetworkConnection[c.Count];
            AddRange(c);
        }

        /// <summary>
        ///        Initializes a new instance of the <c>Win32NetworkConnectionList</c> class
        ///        that contains elements copied from the specified <see cref="NetworkConnection"/> array.
        /// </summary>
        /// <param name="a">The <see cref="NetworkConnection"/> array whose elements are copied to the new list.</param>
        public Win32NetworkConnectionList(NetworkConnection[] a)
        {
            m_array = new NetworkConnection[a.Length];
            AddRange(a);
        }
        #endregion
        
        #region Operations (type-safe ICollection)
        /// <summary>
        ///        Gets the number of elements actually contained in the <c>Win32NetworkConnectionList</c>.
        /// </summary>
        public virtual int Count
        {
            get { return m_count; }
        }

        /// <summary>
        ///        Copies the entire <c>Win32NetworkConnectionList</c> to a one-dimensional
        ///        string array.
        /// </summary>
        /// <param name="array">The one-dimensional <see cref="NetworkConnection"/> array to copy to.</param>
        public virtual void CopyTo(NetworkConnection[] array)
        {
            this.CopyTo(array, 0);
        }

        /// <summary>
        ///        Copies the entire <c>Win32NetworkConnectionList</c> to a one-dimensional
        ///        <see cref="NetworkConnection"/> array, starting at the specified index of the target array.
        /// </summary>
        /// <param name="array">The one-dimensional <see cref="NetworkConnection"/> array to copy to.</param>
        /// <param name="start">The zero-based index in <paramref name="array"/> at which copying begins.</param>
        public virtual void CopyTo(NetworkConnection[] array, int start)
        {
            if (m_count > array.GetUpperBound(0) + 1 - start)
                throw new System.ArgumentException("Destination array was not long enough.");
            
            Array.Copy(m_array, start, array, 0, m_count); 
        }

        /// <summary>
        ///        Gets a value indicating whether access to the collection is synchronized (thread-safe).
        /// </summary>
        /// <returns>true if access to the ICollection is synchronized (thread-safe); otherwise, false.</returns>
        public virtual bool IsSynchronized
        {
            get { return m_array.IsSynchronized; }
        }

        /// <summary>
        ///        Gets an object that can be used to synchronize access to the collection.
        /// </summary>
        public virtual object SyncRoot
        {
            get { return m_array.SyncRoot; }
        }
        #endregion
        
        #region Operations (type-safe IList)
        /// <summary>
        ///        Gets or sets the <see cref="NetworkConnection"/> at the specified index.
        /// </summary>
        /// <param name="index">The zero-based index of the element to get or set.</param>
        /// <exception cref="ArgumentOutOfRangeException">
        ///        <para><paramref name="index"/> is less than zero</para>
        ///        <para>-or-</para>
        ///        <para><paramref name="index"/> is equal to or greater than <see cref="Win32NetworkConnectionList.Count"/>.</para>
        /// </exception>
        public virtual NetworkConnection this[int index]
        {
            get
            {
                ValidateIndex(index); // throws
                return m_array[index]; 
            }
            set
            {
                ValidateIndex(index); // throws
                ++m_version; 
                m_array[index] = value; 
            }
        }

        /// <summary>
        ///        Adds a <see cref="NetworkConnection"/> to the end of the <c>Win32NetworkConnectionList</c>.
        /// </summary>
        /// <param name="item">The <see cref="NetworkConnection"/> to be added to the end of the <c>Win32NetworkConnectionList</c>.</param>
        /// <returns>The index at which the value has been added.</returns>
        public virtual int Add(NetworkConnection item)
        {
            if (m_count == m_array.Length)
                EnsureCapacity(m_count + 1);

            m_array[m_count] = item;
            m_version++;

            return m_count++;
        }
        
        /// <summary>
        ///        Removes all elements from the <c>Win32NetworkConnectionList</c>.
        /// </summary>
        public virtual void Clear()
        {
            ++m_version;
            m_array = new NetworkConnection[DEFAULT_CAPACITY];
            m_count = 0;
        }
        
        /// <summary>
        ///        Creates a shallow copy of the <see cref="Win32NetworkConnectionList"/>.
        /// </summary>
        public virtual object Clone()
        {
            Win32NetworkConnectionList newColl = new Win32NetworkConnectionList(m_count);
            Array.Copy(m_array, 0, newColl.m_array, 0, m_count);
            newColl.m_count = m_count;
            newColl.m_version = m_version;

            return newColl;
        }

        /// <summary>
        ///        Determines whether a given <see cref="NetworkConnection"/> is in the <c>Win32NetworkConnectionList</c>.
        /// </summary>
        /// <param name="item">The <see cref="NetworkConnection"/> to check for.</param>
        /// <returns><c>true</c> if <paramref name="item"/> is found in the <c>Win32NetworkConnectionList</c>; otherwise, <c>false</c>.</returns>
        public virtual bool Contains(NetworkConnection item)
        {
            for (int i=0; i != m_count; ++i)
                if (m_array[i].Equals(item))
                    return true;
            return false;
        }

        /// <summary>
        ///        Returns the zero-based index of the first occurrence of a <see cref="NetworkConnection"/>
        ///        in the <c>Win32NetworkConnectionList</c>.
        /// </summary>
        /// <param name="item">The <see cref="NetworkConnection"/> to locate in the <c>Win32NetworkConnectionList</c>.</param>
        /// <returns>
        ///        The zero-based index of the first occurrence of <paramref name="item"/> 
        ///        in the entire <c>Win32NetworkConnectionList</c>, if found; otherwise, -1.
        ///    </returns>
        public virtual int IndexOf(NetworkConnection item)
        {
            for (int i=0; i != m_count; ++i)
                if (m_array[i].Equals(item))
                    return i;
            return -1;
        }

        /// <summary>
        ///        Inserts an element into the <c>Win32NetworkConnectionList</c> at the specified index.
        /// </summary>
        /// <param name="index">The zero-based index at which <paramref name="item"/> should be inserted.</param>
        /// <param name="item">The <see cref="NetworkConnection"/> to insert.</param>
        /// <exception cref="ArgumentOutOfRangeException">
        ///        <para><paramref name="index"/> is less than zero</para>
        ///        <para>-or-</para>
        ///        <para><paramref name="index"/> is equal to or greater than <see cref="Win32NetworkConnectionList.Count"/>.</para>
        /// </exception>
        public virtual void Insert(int index, NetworkConnection item)
        {
            ValidateIndex(index, true); // throws
            
            if (m_count == m_array.Length)
                EnsureCapacity(m_count + 1);

            if (index < m_count)
            {
                Array.Copy(m_array, index, m_array, index + 1, m_count - index);
            }

            m_array[index] = item;
            m_count++;
            m_version++;
        }

        /// <summary>
        ///        Removes the first occurrence of a specific <see cref="NetworkConnection"/> from the <c>Win32NetworkConnectionList</c>.
        /// </summary>
        /// <param name="item">The <see cref="NetworkConnection"/> to remove from the <c>Win32NetworkConnectionList</c>.</param>
        /// <exception cref="ArgumentException">
        ///        The specified <see cref="NetworkConnection"/> was not found in the <c>Win32NetworkConnectionList</c>.
        /// </exception>
        public virtual void Remove(NetworkConnection item)
        {           
            int i = IndexOf(item);
            if (i < 0)
                throw new System.ArgumentException("Cannot remove the specified item because it was not found in the specified Collection.");
            
            ++m_version;
            RemoveAt(i);
        }

        /// <summary>
        ///        Removes the element at the specified index of the <c>Win32NetworkConnectionList</c>.
        /// </summary>
        /// <param name="index">The zero-based index of the element to remove.</param>
        /// <exception cref="ArgumentOutOfRangeException">
        ///        <para><paramref name="index"/> is less than zero</para>
        ///        <para>-or-</para>
        ///        <para><paramref name="index"/> is equal to or greater than <see cref="Win32NetworkConnectionList.Count"/>.</para>
        /// </exception>
        public virtual void RemoveAt(int index)
        {
            ValidateIndex(index); // throws
            
            m_count--;

            if (index < m_count)
            {
                Array.Copy(m_array, index + 1, m_array, index, m_count - index);
            }
            
            // We can't set the deleted entry equal to null, because it might be a value type.
            // Instead, we'll create an empty single-element array of the right type and copy it 
            // over the entry we want to erase.
            NetworkConnection[] temp = new NetworkConnection[1];
            Array.Copy(temp, 0, m_array, m_count, 1);
            m_version++;
        }

        /// <summary>
        ///        Gets a value indicating whether the collection has a fixed size.
        /// </summary>
        /// <value>true if the collection has a fixed size; otherwise, false. The default is false</value>
        public virtual bool IsFixedSize
        {
            get { return false; }
        }

        /// <summary>
        ///        gets a value indicating whether the IList is read-only.
        /// </summary>
        /// <value>true if the collection is read-only; otherwise, false. The default is false</value>
        public virtual bool IsReadOnly
        {
            get { return false; }
        }
        #endregion

        #region Operations (type-safe IEnumerable)
        
        /// <summary>
        ///        Returns an enumerator that can iterate through the <c>Win32NetworkConnectionList</c>.
        /// </summary>
        /// <returns>An <see cref="Enumerator"/> for the entire <c>Win32NetworkConnectionList</c>.</returns>
        public virtual IWin32NetworkConnectionListEnumerator GetEnumerator()
        {
            return new Enumerator(this);
        }
        #endregion

        #region Public helpers (just to mimic some nice features of ArrayList)
        
        /// <summary>
        ///        Gets or sets the number of elements the <c>Win32NetworkConnectionList</c> can contain.
        /// </summary>
        public virtual int Capacity
        {
            get { return m_array.Length; }
            
            set
            {
                if (value < m_count)
                    value = m_count;

                if (value != m_array.Length)
                {
                    if (value > 0)
                    {
                        NetworkConnection[] temp = new NetworkConnection[value];
                        Array.Copy(m_array, temp, m_count);
                        m_array = temp;
                    }
                    else
                    {
                        m_array = new NetworkConnection[DEFAULT_CAPACITY];
                    }
                }
            }
        }

        /// <summary>
        ///        Adds the elements of another <c>Win32NetworkConnectionList</c> to the current <c>Win32NetworkConnectionList</c>.
        /// </summary>
        /// <param name="x">The <c>Win32NetworkConnectionList</c> whose elements should be added to the end of the current <c>Win32NetworkConnectionList</c>.</param>
        /// <returns>The new <see cref="Win32NetworkConnectionList.Count"/> of the <c>Win32NetworkConnectionList</c>.</returns>
        public virtual int AddRange(Win32NetworkConnectionList x)
        {
            if (m_count + x.Count >= m_array.Length)
                EnsureCapacity(m_count + x.Count);
            
            Array.Copy(x.m_array, 0, m_array, m_count, x.Count);
            m_count += x.Count;
            m_version++;

            return m_count;
        }

        /// <summary>
        ///        Adds the elements of a <see cref="NetworkConnection"/> array to the current <c>Win32NetworkConnectionList</c>.
        /// </summary>
        /// <param name="x">The <see cref="NetworkConnection"/> array whose elements should be added to the end of the <c>Win32NetworkConnectionList</c>.</param>
        /// <returns>The new <see cref="Win32NetworkConnectionList.Count"/> of the <c>Win32NetworkConnectionList</c>.</returns>
        public virtual int AddRange(NetworkConnection[] x)
        {
            if (m_count + x.Length >= m_array.Length)
                EnsureCapacity(m_count + x.Length);

            Array.Copy(x, 0, m_array, m_count, x.Length);
            m_count += x.Length;
            m_version++;

            return m_count;
        }
        
        /// <summary>
        ///        Sets the capacity to the actual number of elements.
        /// </summary>
        public virtual void TrimToSize()
        {
            this.Capacity = m_count;
        }

        #endregion

        #region Implementation (helpers)

        /// <exception cref="ArgumentOutOfRangeException">
        ///        <para><paramref name="index"/> is less than zero</para>
        ///        <para>-or-</para>
        ///        <para><paramref name="index"/> is equal to or greater than <see cref="Win32NetworkConnectionList.Count"/>.</para>
        /// </exception>
        private void ValidateIndex(int i)
        {
            ValidateIndex(i, false);
        }

        /// <exception cref="ArgumentOutOfRangeException">
        ///        <para><paramref name="index"/> is less than zero</para>
        ///        <para>-or-</para>
        ///        <para><paramref name="index"/> is equal to or greater than <see cref="Win32NetworkConnectionList.Count"/>.</para>
        /// </exception>
        private void ValidateIndex(int i, bool allowEqualEnd)
        {
            int max = (allowEqualEnd)?(m_count):(m_count-1);
            if (i < 0 || i > max)
                throw new System.ArgumentOutOfRangeException("Index was out of range.  Must be non-negative and less than the size of the collection.", (object)i, "Specified argument was out of the range of valid values.");
        }

        private void EnsureCapacity(int min)
        {
            int newCapacity = ((m_array.Length == 0) ? DEFAULT_CAPACITY : m_array.Length * 2);
            if (newCapacity < min)
                newCapacity = min;

            this.Capacity = newCapacity;
        }

        #endregion
        
        #region Implementation (ICollection)

        void ICollection.CopyTo(Array array, int start)
        {
            this.CopyTo((NetworkConnection[])array, start);
        }

        #endregion

        #region Implementation (IList)

        object IList.this[int i]
        {
            get { return (object)this[i]; }
            set { this[i] = (NetworkConnection)value; }
        }

        int IList.Add(object x)
        {
            return this.Add((NetworkConnection)x);
        }

        bool IList.Contains(object x)
        {
            return this.Contains((NetworkConnection)x);
        }

        int IList.IndexOf(object x)
        {
            return this.IndexOf((NetworkConnection)x);
        }

        void IList.Insert(int pos, object x)
        {
            this.Insert(pos, (NetworkConnection)x);
        }

        void IList.Remove(object x)
        {
            this.Remove((NetworkConnection)x);
        }

        void IList.RemoveAt(int pos)
        {
            this.RemoveAt(pos);
        }

        #endregion

        #region Implementation (IEnumerable)

        IEnumerator IEnumerable.GetEnumerator()
        {
            return (IEnumerator)(this.GetEnumerator());
        }

        #endregion

        #region Nested enumerator class
        /// <summary>
        ///        Supports simple iteration over a <see cref="Win32NetworkConnectionList"/>.
        /// </summary>
        private class Enumerator : IEnumerator, IWin32NetworkConnectionListEnumerator
        {
            #region Implementation (data)
            
            private Win32NetworkConnectionList m_collection;
            private int m_index;
            private int m_version;
            
            #endregion
        
            #region Construction
            
            /// <summary>
            ///        Initializes a new instance of the <c>Enumerator</c> class.
            /// </summary>
            /// <param name="tc"></param>
            internal Enumerator(Win32NetworkConnectionList tc)
            {
                m_collection = tc;
                m_index = -1;
                m_version = tc.m_version;
            }
            
            #endregion
    
            #region Operations (type-safe IEnumerator)
            
            /// <summary>
            ///        Gets the current element in the collection.
            /// </summary>
            public NetworkConnection Current
            {
                get { return m_collection[m_index]; }
            }

            /// <summary>
            ///        Advances the enumerator to the next element in the collection.
            /// </summary>
            /// <exception cref="InvalidOperationException">
            ///        The collection was modified after the enumerator was created.
            /// </exception>
            /// <returns>
            ///        <c>true</c> if the enumerator was successfully advanced to the next element; 
            ///        <c>false</c> if the enumerator has passed the end of the collection.
            /// </returns>
            public bool MoveNext()
            {
                if (m_version != m_collection.m_version)
                    throw new System.InvalidOperationException("Collection was modified; enumeration operation may not execute.");

                ++m_index;
                return (m_index < m_collection.Count) ? true : false;
            }

            /// <summary>
            ///        Sets the enumerator to its initial position, before the first element in the collection.
            /// </summary>
            public void Reset()
            {
                m_index = -1;
            }
            #endregion
    
            #region Implementation (IEnumerator)
            
            object IEnumerator.Current
            {
                get { return (object)(this.Current); }
            }
            
            #endregion
        }
        #endregion
        
        #region Nested Syncronized Wrapper class
        private class SyncWin32NetworkConnectionList : Win32NetworkConnectionList
        {
            #region Implementation (data)
            private Win32NetworkConnectionList m_collection;
            private object m_root;
            #endregion

            #region Construction
            internal SyncWin32NetworkConnectionList(Win32NetworkConnectionList list)
            {
                m_root = list.SyncRoot;
                m_collection = list;
            }
            #endregion
            
            #region Type-safe ICollection
            public override void CopyTo(NetworkConnection[] array)
            {
                lock(this.m_root)
                    m_collection.CopyTo(array);
            }

            public override void CopyTo(NetworkConnection[] array, int start)
            {
                lock(this.m_root)
                    m_collection.CopyTo(array,start);
            }
            public override int Count
            {
                get
                { 
                    lock(this.m_root)
                        return m_collection.Count;
                }
            }

            public override bool IsSynchronized
            {
                get { return true; }
            }

            public override object SyncRoot
            {
                get { return this.m_root; }
            }
            #endregion
            
            #region Type-safe IList
            public override NetworkConnection this[int i]
            {
                get
                {
                    lock(this.m_root)
                        return m_collection[i];
                }
                set
                {
                    lock(this.m_root)
                        m_collection[i] = value; 
                }
            }

            public override int Add(NetworkConnection x)
            {
                lock(this.m_root)
                    return m_collection.Add(x);
            }
            
            public override void Clear()
            {
                lock(this.m_root)
                    m_collection.Clear();
            }

            public override bool Contains(NetworkConnection x)
            {
                lock(this.m_root)
                    return m_collection.Contains(x);
            }

            public override int IndexOf(NetworkConnection x)
            {
                lock(this.m_root)
                    return m_collection.IndexOf(x);
            }

            public override void Insert(int pos, NetworkConnection x)
            {
                lock(this.m_root)
                    m_collection.Insert(pos,x);
            }

            public override void Remove(NetworkConnection x)
            {           
                lock(this.m_root)
                    m_collection.Remove(x);
            }

            public override void RemoveAt(int pos)
            {
                lock(this.m_root)
                    m_collection.RemoveAt(pos);
            }
            
            public override bool IsFixedSize
            {
                get {return m_collection.IsFixedSize;}
            }

            public override bool IsReadOnly
            {
                get {return m_collection.IsReadOnly;}
            }
            #endregion

            #region Type-safe IEnumerable
            public override IWin32NetworkConnectionListEnumerator GetEnumerator()
            {
                lock(m_root)
                    return m_collection.GetEnumerator();
            }
            #endregion

            #region Public Helpers
            // (just to mimic some nice features of ArrayList)
            public override int Capacity
            {
                get
                {
                    lock(this.m_root)
                        return m_collection.Capacity;
                }
                
                set
                {
                    lock(this.m_root)
                        m_collection.Capacity = value;
                }
            }

            public override int AddRange(Win32NetworkConnectionList x)
            {
                lock(this.m_root)
                    return m_collection.AddRange(x);
            }

            public override int AddRange(NetworkConnection[] x)
            {
                lock(this.m_root)
                    return m_collection.AddRange(x);
            }
            #endregion
        }
        #endregion

        #region Nested Read Only Wrapper class
        private class ReadOnlyWin32NetworkConnectionList : Win32NetworkConnectionList
        {
            #region Implementation (data)
            private Win32NetworkConnectionList m_collection;
            #endregion

            #region Construction
            internal ReadOnlyWin32NetworkConnectionList(Win32NetworkConnectionList list)
            {
                m_collection = list;
            }
            #endregion
            
            #region Type-safe ICollection
            public override void CopyTo(NetworkConnection[] array)
            {
                m_collection.CopyTo(array);
            }

            public override void CopyTo(NetworkConnection[] array, int start)
            {
                m_collection.CopyTo(array,start);
            }
            public override int Count
            {
                get {return m_collection.Count;}
            }

            public override bool IsSynchronized
            {
                get { return m_collection.IsSynchronized; }
            }

            public override object SyncRoot
            {
                get { return this.m_collection.SyncRoot; }
            }
            #endregion
            
            #region Type-safe IList
            public override NetworkConnection this[int i]
            {
                get { return m_collection[i]; }
                set { throw new NotSupportedException("This is a Read Only Collection and can not be modified"); }
            }

            public override int Add(NetworkConnection x)
            {
                throw new NotSupportedException("This is a Read Only Collection and can not be modified");
            }
            
            public override void Clear()
            {
                throw new NotSupportedException("This is a Read Only Collection and can not be modified");
            }

            public override bool Contains(NetworkConnection x)
            {
                return m_collection.Contains(x);
            }

            public override int IndexOf(NetworkConnection x)
            {
                return m_collection.IndexOf(x);
            }

            public override void Insert(int pos, NetworkConnection x)
            {
                throw new NotSupportedException("This is a Read Only Collection and can not be modified");
            }

            public override void Remove(NetworkConnection x)
            {           
                throw new NotSupportedException("This is a Read Only Collection and can not be modified");
            }

            public override void RemoveAt(int pos)
            {
                throw new NotSupportedException("This is a Read Only Collection and can not be modified");
            }
            
            public override bool IsFixedSize
            {
                get {return true;}
            }

            public override bool IsReadOnly
            {
                get {return true;}
            }
            #endregion

            #region Type-safe IEnumerable
            public override IWin32NetworkConnectionListEnumerator GetEnumerator()
            {
                return m_collection.GetEnumerator();
            }
            #endregion

            #region Public Helpers
            // (just to mimic some nice features of ArrayList)
            public override int Capacity
            {
                get { return m_collection.Capacity; }
                
                set { throw new NotSupportedException("This is a Read Only Collection and can not be modified"); }
            }

            public override int AddRange(Win32NetworkConnectionList x)
            {
                throw new NotSupportedException("This is a Read Only Collection and can not be modified");
            }

            public override int AddRange(NetworkConnection[] x)
            {
                throw new NotSupportedException("This is a Read Only Collection and can not be modified");
            }
            #endregion
        }
        #endregion
    }
}
