// ScreenSaver.cs: contributed by Shawn Van Ness [shawnv@arithex.com]
// A base class for managed screen savers (rev: 2002.08.15)

#region Copyright  2002-2003 The Genghis Group
/*
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from the
 * use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not claim
 * that you wrote the original software. If you use this software in a product,
 * an acknowledgment in the product documentation is required, as shown here:
 *
 * Portions Copyright  2002-2003 The Genghis Group (http://www.genghisgroup.com/).
 *
 * 2. No substantial portion of the source code of this library may be redistributed
 * without the express written permission of the copyright holders, where
 * "substantial" is defined as enough code to be recognizably from this library.
*/
#endregion
#region History
/*
 * 08/18/02 - Initial Version
 */
#endregion
#region Usage
// 1)Derive a class from this form, and override OnLoad and OnPaint to do your 
//   bidding.
//
//2)Early in your Main function (before instantiating this form class!) you should 
//  query the static property ScreenSaver.ConfigurationMode to determine if you're 
//  being requested to show your configuration UI -- and then do so, or at least 
//  display a msgbox indicating that you have no user-configurable properties.
//
//3)To deploy your screensaver, you must rename your file extenstion from 
//  "EXE" to "SCR", then copy it into the %SystemRoot%\system32 directory.
//
//4)To provide a friendly description string (as seen in the desk.cpl's 
//  combobox), you must link an unmanaged string-table resource into your assembly, 
//  with your description string as ID #1.  <groan>  The /win32res switch on csc.exe 
//  will do this for you, but it means abandoning VS.NET in favor of NAnt or NMake.
#endregion
#region TODO
//-- there seems to be some disagreement in the docs about what to do when run w/o 
//    any command line arguments: configure, or launch.  resolve this! (what do 
//    the ms-authored savers do?)
//
//-- solve the /p:<parentHandle> problem
//
//-- test on multimon system
#endregion

using System;
using System.Drawing;
using System.Windows.Forms;
using System.Threading;
using System.Diagnostics;

namespace Genghis.Windows.Forms
{
    /// <summary>
    /// The main screensaver window -- big, black, borderless, and topmost.
    /// The window closes itself on keystrokes and (optionally) mouse movement.
    /// </summary>
    public class ScreenSaver : Form
    {
        #region Member variables

        // Whether or not to exit on mousemove
        private bool ignoreMouseMove = false;

        // Ignore early, errant mousemove msgs
        private DateTime failSafe = DateTime.Now + new TimeSpan( 0, 0, 2); // 2sec into the future
        #endregion

        #region Initialization and finalization

        public ScreenSaver()
        {
            // Required for Windows Form Designer support
            InitializeComponent();

            // Being TopMost makes debugging difficult
#if DEBUG
            this.TopMost = false;
#endif
        }

        protected override void Dispose( bool disposing)
        { base.Dispose( disposing); }

        #region Windows Form Designer generated code
        private void InitializeComponent()
        {
            // 
            // ScreenSaver
            // 
            this.AutoScale = false;
            this.BackColor = System.Drawing.Color.Black;
            this.CausesValidation = false;
            this.ControlBox = false;
            this.ForeColor = System.Drawing.SystemColors.GrayText;
            this.FormBorderStyle = System.Windows.Forms.FormBorderStyle.None;
            this.MaximizeBox = false;
            this.MinimizeBox = false;
            this.Name = "ScreenSaver";
            this.ShowInTaskbar = false;
            this.StartPosition = System.Windows.Forms.FormStartPosition.Manual;
            this.TopMost = true;
            this.Visible = false;
            this.WindowState = System.Windows.Forms.FormWindowState.Maximized;
        }
        #endregion
        #endregion

        #region Overrides

        protected override void OnLoad( EventArgs ea)
        {
            base.OnLoad(ea);

            // Bail, if another instance is already running in our session
            bool createdNew = false;
            Mutex singletonGuard = new Mutex( false, "Local\\mutex6642D0F2-6EA8-4911-8B16-7FD886C055D4", out createdNew);

            if ( !createdNew)
            {
                this.Close();
                return;
            }

            // Are we being parented? ([/p <handle>|/p:<handle>])
            string[] args = Environment.GetCommandLineArgs();
            IntPtr parentHandle = IntPtr.Zero;
            if ( args.Length > 1 && ( args[1].Substring( 0, 2).ToLower() == "/p"))
            {
//                string argParent = null;
//                if ( args.Length == 2)
//                    argParent = args[1].Substring(2);
//                else
//                    argParent = args[2];
//
//                parentHandle = int.Parse(argParent);
//
                //todo: figure out what to do with parentHandle!
                //    maybe p/invoke User32!SetParent
                //    maybe redesign this whole mess around NativeWindow
                this.WindowState = FormWindowState.Normal;
                this.Size = new Size( 320, 240);
                this.Location = new Point( 160, 120);
            }

            // Let's take it nice and easy, from here on in
            Process.GetCurrentProcess().PriorityClass = ProcessPriorityClass.Idle;

            // Show ourselves
            this.Visible = true;

            // Hide the cursor, initially
            Cursor.Hide();
        }

        protected override void OnClosed( EventArgs ea)
        {
            // Restore the cursor (just good karma)
            Cursor.Show();
            base.OnClosed(ea);
        }

        protected override void OnClick( EventArgs ea)
        {
            base.OnClick(ea);
            this.Close();
        }

        protected override void OnKeyDown( KeyEventArgs ea)
        {
            base.OnKeyDown(ea);
            this.Close();
        }

        protected override void OnMouseMove( MouseEventArgs ea)
        {
            base.OnMouseMove(ea);
            if ( !this.ignoreMouseMove && DateTime.Now > this.failSafe)
                this.Close();
        }
        #endregion

        #region Properties

        public static bool ConfigurationMode
        {
            get
            {
                // ([/c|/c:<handle>|/c <handle>])
                string[] args = Environment.GetCommandLineArgs();
                if ( args.Length > 1 && ( args[1].Substring( 0, 2).ToLower() == "/c"))
                    return true;
                else
                    return false;
            }
        }

        protected bool IgnoreMouseMove
        {
            get
            { return this.ignoreMouseMove; }
            set
            { this.ignoreMouseMove = value; }
        }
        #endregion
    }
}
