// ScrollablePictureBox.cs: Contributed by Husein Choroomi [hchoroomi@usermail.com]
// A Scrollable PictureBox Control
#region Copyright  2002-2003 The Genghis Group
/*
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from the
 * use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not claim
 * that you wrote the original software. If you use this software in a product,
 * an acknowledgment in the product documentation is required, as shown here:
 *
 * Portions Copyright  2002-2003 The Genghis Group (http://www.genghisgroup.com/).
 *
 * 2. No substantial portion of the source code of this library may be redistributed
 * without the express written permission of the copyright holders, where
 * "substantial" is defined as enough code to be recognizably from this library.
*/
#endregion
#region History
/*
 * 04/15/03 - Initial Version
 */
#endregion

using System;
using System.Collections;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Windows.Forms;


namespace Genghis.Windows.Forms
{
    /// <summary>
    /// A Scrollable PictureBox Class
    /// </summary>
    [
    DesignTimeVisible(true),
    ToolboxItem(true),
    ]
    public class ScrollablePictureBox : System.Windows.Forms.PictureBox
    {
        #region Private Fields
        
        private System.ComponentModel.Container components = null;
        private System.Windows.Forms.HScrollBar horScrollBar;
        private System.Windows.Forms.VScrollBar verScrollBar;
        private Image pic= null;
        private int x = 0;
        private int y = 0;
        
        #endregion

        #region  Initialization and finalization

        #region Constructor/Destructor

        public ScrollablePictureBox()
        {
            InitializeComponent();
            Adjust();
        }

        protected override void Dispose( bool disposing )
        {
            if( disposing )
            {
                if(components != null)
                {
                    components.Dispose();
                }
            }
            base.Dispose( disposing );
        }
        
        #endregion

        #region Windows Form Designer generated code

        private void InitializeComponent()
        {
            this.verScrollBar = new System.Windows.Forms.VScrollBar();
            this.horScrollBar = new System.Windows.Forms.HScrollBar();
            this.SuspendLayout();
            // 
            // verScrollBar
            // 
            this.verScrollBar.Location = new System.Drawing.Point(136, 8);
            this.verScrollBar.Name = "verScrollBar";
            this.verScrollBar.TabIndex = 0;
            this.verScrollBar.Scroll += new System.Windows.Forms.ScrollEventHandler(this.VerScrollBarScroll);
            // 
            // horScrollBar
            // 
            this.horScrollBar.Location = new System.Drawing.Point(0, 128);
            this.horScrollBar.Name = "horScrollBar";
            this.horScrollBar.TabIndex = 1;
            this.horScrollBar.Scroll += new System.Windows.Forms.ScrollEventHandler(this.HorScrollBarScroll);
            // 
            // ScrollablePictureBox
            // 
            this.Controls.AddRange(new System.Windows.Forms.Control[] {
                                                                          this.verScrollBar,
                                                                          this.horScrollBar});
            this.Name = "ScrollablePictureBox";
            this.SizeChanged += new System.EventHandler(this.SizeChanging);
            this.Paint += new System.Windows.Forms.PaintEventHandler(this.ScrollablePictureBoxPaint);
            this.ResumeLayout(false);

        }
        
        #endregion

        #endregion

        #region private Methods

        /// <summary>
        /// Adjust
        /// </summary>
        private void Adjust()
        {
            horScrollBar.SetBounds(0, ClientRectangle.Height - horScrollBar.Height, ClientRectangle.Width - verScrollBar.Width, horScrollBar.Height);
            verScrollBar.SetBounds(ClientRectangle.Right - verScrollBar.Width, 0, verScrollBar.Width, ClientRectangle.Height - horScrollBar.Height);  

            if (pic != null)
            {
                horScrollBar.Maximum = pic.Width + verScrollBar.Width*2  - ClientRectangle.Width;
                verScrollBar.Maximum = pic.Height + horScrollBar.Height*2 - ClientRectangle.Height;
            }

            Invalidate();
        }
        /// <summary>
        /// SizeChanging
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void SizeChanging(object sender, System.EventArgs e)
        {
            Adjust();
        }

        /// <summary>
        /// ScrollablePictureBoxPaint
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ScrollablePictureBoxPaint(object sender, System.Windows.Forms.PaintEventArgs e)
        {
            Graphics g = e.Graphics;
            if (pic != null)
            {
                g.DrawImageUnscaled(pic, -HorizontalValue, -VerticalValue, pic.Width, pic.Height);
                g.FillRectangle(System.Drawing.SystemBrushes.Control, ClientRectangle.Width - verScrollBar.Width, ClientRectangle.Height - horScrollBar.Height, verScrollBar.Width, horScrollBar.Height);
            }
            TestSize();
        }
    
        /// <summary>
        /// HorScrollBarScroll
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void HorScrollBarScroll(object sender, System.Windows.Forms.ScrollEventArgs e)
        {
            HorizontalValue = e.NewValue ;
            Invalidate();
        }
        
        /// <summary>
        /// VerScrollBarScroll
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void VerScrollBarScroll(object sender, System.Windows.Forms.ScrollEventArgs e)
        {
            VerticalValue = e.NewValue ;
            Invalidate();
        }
        
        /// <summary>
        /// TestSize
        /// </summary>
        private void TestSize()
        {
            if (pic != null)
            {
                if (pic.Width <= horScrollBar.Width)
                {
                    horScrollBar.Enabled = false;
                }

                if (pic.Height <= verScrollBar.Height)
                {
                    verScrollBar.Enabled = false;
                }
            }
            else
            {
                horScrollBar.Enabled = false;
                verScrollBar.Enabled = false;
            }
        }

        #endregion

        #region Properties

        /// <summary>
        /// Image
        /// </summary>
        [Description("The image displayed in the ScrollablePictureBox."), Category("Appearance")]
        public Image SrcImage
        {
            get 
            {
                return pic;
            }
            set
            {
                pic= value;
                Adjust();
            }
        }
                
        /// <summary>
        /// VerticalValue
        /// </summary>
        [Description("The vertical position of the thumb."), Category("Appearance")]
        public int VerticalValue
        {
            get
            {
                return y;
            }
            set
            {
                y = value;
            }
        }
        
        /// <summary>
        /// HorizontalValue
        /// </summary>
        [Description("The horizontal position of the thumb."), Category("Appearance")]
        public int HorizontalValue
        {
            get
            {
                return x;
            }
            set
            {
                x = value;
            }
        }

        #endregion
    }
}
