// StatusBarExtender.cs: Contributed by Chris Sells [csells@sellsbrothers.com]
// Inspired by a question from Shawn Wildermuth
// A class to give each control on a form a StatusText property
// and to set the Text of a status bar based on that property
// when the control gets focus
#region Copyright  2002-2003 The Genghis Group
/*
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from the
 * use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not claim
 * that you wrote the original software. If you use this software in a product,
 * an acknowledgment in the product documentation is required, as shown here:
 *
 * Portions Copyright  2002-2003 The Genghis Group (http://www.genghisgroup.com/).
 *
 * 2. No substantial portion of the source code of this library may be redistributed
 * without the express written permission of the copyright holders, where
 * "substantial" is defined as enough code to be recognizably from this library.
*/
#endregion
#region History
/*
 * 1.0.0.0 Initial Version  
*/
#endregion
#region Usage
// 1. Drop an instance of the StatusBarExtender onto the form.
// 2. Set the StatusBar property of StatusBarExtender component.
// 3. Set the StatusText property of each control on the form.
// 4. At run-time, as each control gets focus, the status bar
//    will be updated to show the StatusText property of that
//    control.
#endregion

using System;
using System.Windows.Forms;
using System.ComponentModel;
using System.Collections;
using System.Diagnostics;

namespace Genghis.Windows.Forms {

  /// <summary>
  /// A class to give each control on a form a StatusText property
  /// and to set the Text of a status bar based on that property
  /// when the control gets focus
  /// </summary>
  [ProvideProperty("StatusText", typeof(Control))]
  public class StatusBarExtender : Component, IExtenderProvider {
    /// Required designer variable
    private System.ComponentModel.Container components = null;

    public StatusBarExtender(System.ComponentModel.IContainer container) {
      /// Required for Windows.Forms Class Composition Designer support
      container.Add(this);
      InitializeComponent();
    }

    public StatusBarExtender() {
      /// Required for Windows.Forms Class Composition Designer support
      InitializeComponent();
    }

        #region Component Designer generated code
    /// <summary>
    /// Required method for Designer support - do not modify
    /// the contents of this method with the code editor.
    /// </summary>
    private void InitializeComponent() {
      components = new System.ComponentModel.Container();
    }
        #endregion

    /// <summary>
    /// The statusbar that gets set when a new control gets focus
    /// </summary>
    private StatusBar statusbar = null;
    public StatusBar StatusBar {
      get {
        return this.statusbar;
      }
      set {
        this.statusbar = value;
      }
    }

    // Mapping of components to status text
    Hashtable statusTexts = new Hashtable();

    public string GetStatusText(Control extendee) {
      // Return component's status text
      return (string)this.statusTexts[extendee];
    }

    public void SetStatusText(Control extendee, object value) {
      // Add the status text
      this.statusTexts[extendee] = (value != null ? value.ToString() : null);

      // Watch for the focus event
      if( !this.DesignMode ) {
        extendee.GotFocus += new EventHandler(extendee_GotFocus);
      }
    }

    void extendee_GotFocus(object sender, EventArgs e) {
      if( this.statusbar == null ) return;

      // Set status bar text
      string text = (string)statusTexts[sender];
      this.statusbar.Text = (text == null ? this.defaultStatusText : text);
    }

    #region Implementation of IExtenderProvider
    public bool CanExtend(object extendee) {
      return true;
    }
    #endregion

    /// <summary>
    /// Default status text for controls that don't have one
    /// </summary>
    string defaultStatusText = "Ready";
    public string DefaultStatusText {
      get {
        return this.defaultStatusText;
      }
      set {
        this.defaultStatusText = value;
      }
    }

  }
}






