// TraceEx is an extension to the Trace class, supporting individual
// TraceSwitches for each assembly and a customizable formatstring to specify the output format.
// Contributed by Horst Veith [hveith@intercom.at].
#region Copyright  2002-2003 The Genghis Group
/*
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from the
 * use of this software.
 * 
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, subject to the following restrictions:
 * 
 * 1. The origin of this software must not be misrepresented; you must not claim
 * that you wrote the original software. If you use this software in a product,
 * an acknowledgment in the product documentation is required, as shown here:
 * 
 * Portions Copyright  2002-2003 The Genghis Group (http://www.genghisgroup.com/).
 * 
 * 2. No substantial portion of the source code of this library may be redistributed
 * without the express written permission of the copyright holders, where
 * "substantial" is defined as enough code to be recognizably from this library. 
*/
#endregion

using System;
using System.Text;
using System.Reflection;
using System.Collections;
using System.Diagnostics;
using System.Configuration;

namespace Genghis.Diagnostics
{
    /// <summary>
    /// TraceEx is an extension to the Trace class, supporting individuell
    /// TraceSwitches for each assembly and a customizable formatstring to specify the output format.
    /// 
    /// <para>
    /// All settings are configurable in the app.config of the application.
    /// </para>
    /// 
    /// <para>
    ///    TraceFormat in appSettings:
    ///    
    ///    <list type="table">
    /// <listheader><term>Format</term><description>Descriptions</description></listheader>
    /// <item><term>{T}</term><description>timestamp</description></item>
    /// <item><term>{A}</term><description>name of calling assembly</description></item>
    /// <item><term>{M}</term><description>name of calling method</description></item>
    /// <item><term>{m}</term><description>trace message</description></item>
    /// <item><term>\n</term><description>CRLF</description></item>
    /// <item><term>\t</term><description>Tag</description></item>
    /// </list>
    /// 
    /// </para>        
    /// 
    /// <para>
    ///    Example: {T}:{A}::{M}\n{m}
    ///    </para>
    ///     
    /// </summary>
    public class TraceEx
    {
        #region WriteMethods

        /// <summary>
        /// Writes a category name and message to the trace listeners in the Listeners collection.
        /// </summary>
        /// <param name="message">A message to write</param>
        /// <param name="category">A category to write</param>
        static public void Write(string message, string category)
        {
            TraceEx.DoWriteMessage(Assembly.GetCallingAssembly(), TraceLevel.Off, message, category);
        }

        /// <summary>
        /// Writes a message to the trace listeners in the Listeners collection.
        /// </summary>
        /// <param name="message">A message to write</param>
        static public void Write(string message)
        {
            TraceEx.DoWriteMessage(Assembly.GetCallingAssembly(), TraceLevel.Off, message, string.Empty);
        }

        /// <summary>
        /// Writes a category name and message to the trace listeners in the Listeners collection
        /// if the TraceLevel for the calling module is set to TraceLevel.Error or higher.
        /// </summary>
        /// <param name="message">A message to write</param>
        /// <param name="category">A category to write</param>
        static public void WriteError(string message, string category)
        {
            TraceEx.DoWriteMessage(Assembly.GetCallingAssembly(), TraceLevel.Error, message, category);
        }

        /// <summary>
        /// Writes a message to the trace listeners in the Listeners collection
        /// if the TraceLevel for the calling module is set to TraceLevel.Error or higher.
        /// </summary>
        /// <param name="message">A message to write</param>
        static public void WriteError(string message)
        {
            TraceEx.DoWriteMessage(Assembly.GetCallingAssembly(), TraceLevel.Error, message, string.Empty);
        }

        /// <summary>
        /// Writes a category name and message to the trace listeners in the Listeners collection
        /// if the TraceLevel for the calling module is set to TraceLevel.Warning or higher.
        /// </summary>
        /// <param name="message">A message to write</param>
        /// <param name="category">A category to write</param>
        static public void WriteWarning(string message, string category)
        {
            TraceEx.DoWriteMessage(Assembly.GetCallingAssembly(), TraceLevel.Warning, message, category);
        }

        /// <summary>
        /// Writes a message to the trace listeners in the Listeners collection
        /// if the TraceLevel for the calling module is set to TraceLevel.Warning or higher.
        /// </summary>
        /// <param name="message">A message to write</param>
        static public void WriteWarning(string message)
        {
            TraceEx.DoWriteMessage(Assembly.GetCallingAssembly(), TraceLevel.Warning, message, string.Empty);
        }

        /// <summary>
        /// Writes a category name and message to the trace listeners in the Listeners collection
        /// if the TraceLevel for the calling module is set to TraceLevel.Info or higher.
        /// </summary>
        /// <param name="message">A message to write</param>
        /// <param name="category">A category to write</param>
        static public void WriteInfo(string message, string category)
        {
            TraceEx.DoWriteMessage(Assembly.GetCallingAssembly(), TraceLevel.Info, message, category);
        }

        /// <summary>
        /// Writes a message to the trace listeners in the Listeners collection
        /// if the TraceLevel for the calling module is set to TraceLevel.Info or higher.
        /// </summary>
        /// <param name="message">A message to write</param>
        static public void WriteInfo(string message)
        {
            TraceEx.DoWriteMessage(Assembly.GetCallingAssembly(), TraceLevel.Info, message, string.Empty);
        }

        /// <summary>
        /// Writes a category name and message to the trace listeners in the Listeners collection
        /// if the TraceLevel for the calling module is set to TraceLevel.Verbose.
        /// </summary>
        /// <param name="message">A message to write</param>
        /// <param name="category">A category to write</param>
        static public void WriteVerbose(string message, string category)
        {
            TraceEx.DoWriteMessage(Assembly.GetCallingAssembly(), TraceLevel.Verbose, message, category);
        }

        /// <summary>
        /// Writes a message to the trace listeners in the Listeners collection 
        /// if the TraceLevel for the calling module is set to TraceLevel.Verbose.
        /// </summary>
        /// <param name="message">A message to write</param>
        static public void WriteVerbose(string message)
        {
            TraceEx.DoWriteMessage(Assembly.GetCallingAssembly(), TraceLevel.Verbose, message, string.Empty);
        }

        /// <summary>
        /// Checks for a condition and writes a message to the trace listeners in the Listeners collection
        /// if the condition is false.
        /// <note type="note">
        /// The message is only written if the TraceLevel for the calling module is set to TraceLevel.Error
        /// or higher.
        /// </note>
        /// </summary>
        /// <param name="condition">&lt;b&gt;true&lt;/b&gt; to prevent a message being written; otherwise, &lt;b&gt;false&lt;/b&gt;.</param>
        /// <param name="message">A message to write</param>
        static public void Assert(bool condition, string message)
        {
            if (!condition)
                DoWriteMessage(Assembly.GetCallingAssembly(), TraceLevel.Error, "Assert: " + message, string.Empty);
        }

        /// <summary>
        /// Writes a textual representation of the gives exception to the trace listeners in the Listeners collection.
        /// <note type="note">
        /// The message is only written if the TraceLevel for the calling module is set to TraceLevel.Error
        /// or higher.
        /// </note>
        /// </summary>
        /// <param name="e">Exception being thrown</param>
        static public void WriteException(Exception e)
        {
            StringBuilder str = new StringBuilder("Exception: " + e.Message + " in " + e.Source + "\r\n");
            str.Append(e.StackTrace);
            DoWriteMessage(Assembly.GetCallingAssembly(), TraceLevel.Error, str.ToString(), string.Empty);
        }

        #endregion

        #region Implementation

        private static void DoWriteMessage(Assembly assembly, TraceLevel level, string message, string category)
        {
            TraceSettings settings = GetSettings(assembly);
            if (settings != null && settings.traceSwitch.Level >= level)
            {
                StringBuilder str = new StringBuilder(TraceEx.fmtString);
                

                if (((flags & TraceFlags.ShowTimestamp) == TraceFlags.ShowTimestamp))
                    str.Replace("{T}", DateTime.Now.ToString("s") + ":");
                if (((flags & TraceFlags.ShowAssembly) == TraceFlags.ShowAssembly))
                    str.Replace("{A}", settings.traceSwitch.DisplayName);
                if (((flags & TraceFlags.ShowMethod) == TraceFlags.ShowMethod))
                {
                    StackTrace stackTrace = new StackTrace(2);
                    MethodBase methodInfo = stackTrace.GetFrame(0).GetMethod();
                    str.Replace("{M}", methodInfo.ReflectedType.ToString() + "." + methodInfo.Name);
                }

                str.Replace("{m}", message);
                
                if (category == string.Empty)
                    Trace.WriteLine(str.ToString());
                else
                    Trace.WriteLine(str.ToString(), category);
                Trace.Flush();
            }
        }

        private static TraceSettings GetSettings(Assembly assembly)
        {
            TraceSettings settings = null;

            if (assembly != null)
            {
                string name = assembly.GetName().Name;
                settings = (TraceSettings)moduleSettings[name];
                if (settings == null)
                {
                    settings = new TraceSettings(name);
                    moduleSettings.Add(name, settings);
                }
            }
            return settings;
        }

        private static string Init()
        {
            string str = ConfigurationSettings.AppSettings["TraceFormat"];
            if (str == null)
                str = "{T}[{A}::{M}]\\n\\t{m}";
 
            str = str.Replace("\\t", "\t");
            str = str.Replace("\\n", "\r\n");
            
            if (-1 != str.IndexOf("{T}")) flags |= TraceFlags.ShowTimestamp;
            if (-1 != str.IndexOf("{A}")) flags |= TraceFlags.ShowAssembly;
            if (-1 != str.IndexOf("{M}")) flags |= TraceFlags.ShowMethod;

            return str;
        }

        enum TraceFlags
        {
            Default          = 0x00,
            ShowTimestamp = 0x01,
            ShowAssembly  = 0x02,
            ShowMethod      = 0x04,
        }

        class TraceSettings
        {
            public TraceSettings(string switchName)
            {
                this.traceSwitch = new TraceSwitch(switchName, "");
            }
            public TraceSwitch    traceSwitch;
        }

        static Hashtable    moduleSettings = new Hashtable();
        static TraceFlags    flags = TraceFlags.Default;
        static string        fmtString = Init();

        #endregion
    }
}
