  // WizardSheet.cs: Contributed by Shawn Wildermuth [swildermuth@adoguy.com]
  // Classes for creation of Wizard-like User Interfaces
#region Copyright  2002-2003 The Genghis Group
  /*
   * This software is provided 'as-is', without any express or implied warranty.
   * In no event will the authors be held liable for any damages arising from the
   * use of this software.
   * 
   * Permission is granted to anyone to use this software for any purpose,
   * including commercial applications, subject to the following restrictions:
   * 
   * 1. The origin of this software must not be misrepresented; you must not claim
   * that you wrote the original software. If you use this software in a product,
   * an acknowledgment in the product documentation is required, as shown here:
   * 
   * Portions Copyright  2002-2003 The Genghis Group (http://www.genghisgroup.com/).
   * 
   * 2. No substantial portion of the source code of this library may be redistributed
   * without the express written permission of the copyright holders, where
   * "substantial" is defined as enough code to be recognizably from this library. 
  */
#endregion

#region Features
/*
 * -Creation of Wizards-type User Interfaces with individual pages and groups
 * -Allows pages to be any object that derives from System.Windows.Forms.Control
 * -Handles First/Last/Finish buttons.
 * -Resizes dialog based on the pages.
 * -Base WizardPage class to allow Windows Forms Inheritance for Wizard Pages.
 * -Allows forming groups of wizard pages and switching of them on the fly.
 * -Autonumbers pages to provide feedback to the user of their location in the wizard.
 * -Supports CanFinish to determine whether a wizard has all the required elements yet.
 */
#endregion
#region Limitations
/*
 * 
*/
#endregion
#region History
/*
 * 6/22/2002:
 * -Initial Development
 * 8/31/2002
 * -Brian Hormann changed buttons to Protected to support changing the buttons
 * -Changed name of helpButton to hlpButton to make it CLS Compliant
 * -Brian Hormann Added Support for Design-Time editing of Inherited classes by 
 *  only running code in Form.Load when not in Design mode
 * -Keith Brown added support for CanFinish to the wizard and wizard pages.
*/
#endregion

using System;
using System.Collections;
using System.Collections.Specialized;
using System.ComponentModel;
using IDesignerHost = System.ComponentModel.Design.IDesignerHost;
using System.Windows.Forms;
using System.Drawing;
using System.Text;
using Genghis;

namespace Genghis.Windows.Forms 
{
  /// <summary>
  /// Frame to hold WizardPages for a Wizard-like User Interface.
  /// </summary>
  /// <example>Typical usage case:
  /// <code>
  /// WizardSheet sht = new WizardSheet();
  /// sht.Add(new WizardPage());
  /// sht.ShowDialog();
  /// </code></example>
  public class WizardSheet : System.Windows.Forms.Form
  {

    #region Constructor
    /// <summary>
    /// Default Constructor
    /// </summary>
    /// <example>To create a new WizardSheet object without specifying the dialog title:
    /// <code>
    /// Wizard sheet = new WizardSheet();
    /// </code>
    /// </example>
    public WizardSheet()
    {
      // Designer Support
      InitializeComponent();
    }

    /// <summary>
    /// Constructs a new Wizard Sheet specifying the name
    /// of the wizard (for the Titlebar).
    /// </summary>
    /// <param name="titleBarText">Text for the Titlebar</param>
    /// <example>To create a new WizardSheet object while specifying the dialog title:
    /// <code>
    /// Wizard sheet = new WizardSheet("My Wizard");
    /// </code>
    /// </example>
    public WizardSheet(string titleBarText)
    {
      // Designer Support
      InitializeComponent();

      // Set the titlebar text
      _TitleBarText = titleBarText;
    }
    #endregion

    #region public Interface
    /// <summary>
    /// Adds a new page
    /// </summary>
    /// <param name="page">The WizardPage object to add to the collection of pages.</param>
    public virtual void AddPage(WizardPage page)
    {
      AddPage(page as Control);
    }

    /// <summary>
    /// Adds a new Page, assigning it to a single group.
    /// </summary>
    /// <param name="page">The WizardPage object to add to the collection of pages.</param>
    /// <param name="group">The name of the group to add this page to.</param>
    public virtual void AddPage(WizardPage page, string group)
    {
      AddPage(page as Control, group);
    }

    /// <summary>
    /// Adds a new Page, assigning it to multiple groups.
    /// </summary>
    /// <param name="page">The WizardPage object to add to the collection of pages.</param>
    /// <param name="groups">An array of strings specifying which groups to add the pages to.</param>
    public virtual void AddPage(WizardPage page, string[] groups)
    {
      AddPage(page as Control, groups);
    }

    /// <summary>
    /// Add a page to the default group.
    /// </summary>
    /// <param name="page">The Control object to add to the collection of pages.</param>
    public virtual void AddPage(Control page)
    {
      // Add to default Group
      AddPage(page, new string[] { DEFAULTGROUP });
    }

    /// <summary>
    /// Adds a page to the Wizard, specifying the specific group to add it to.
    /// </summary>
    /// <param name="page">The Control object to add to the collection of pages.</param>
    /// <param name="group">The name of the group to add this page to.</param>
    public virtual void AddPage(Control page, string group)
    {
      AddPage(page, new string[] { group });
    }

    /// <summary>
    /// Adds a page to the Wizard, specifying the groups in which this new page will belong.
    /// </summary>
    /// <param name="page">The Control object to add to the collection of pages.</param>
    /// <param name="groups">An array of strings specifying which groups to add the pages to.</param>
    public virtual void AddPage(Control page, string[] groups)
    {
      // Check to make sure there is at least one group
      if (groups == null || groups.Length == 0)
      {
        throw new ArgumentNullException("Cannot send in null groups or empty group collection.  Use AddPage(page) syntax to use default group");
      }

      // If this is the first page to be added, figure out the
      // Initial Group Name
      if (_Pages.Count == 0)
      {
        _CurrentGroupName = groups[0];
      }

      // Add it to the page collection
      _Pages.Add(new PageObject(page, groups));
  
      if (_isLoaded)
      {
        // Calculate the Current Group
        CalculatePageGroup();
        
        // Resize
        ResizeDialog();
      }
    }

    /// <summary>
    /// Gets the page in the current group by ordinal number.
    /// </summary>
    /// <param name="item">The zero-based index to a page in the current group.</param>
    public Control GetPage(int item)
    {
      if (_CurrentGroup.Count >= item) throw new ArgumentException("Item number is greater than page count", "item");
      return _CurrentGroup[item] as Control;
    }

    /// <summary>
    /// Determines whether a wizard can successfully Finish itself yet.
    /// </summary>
    /// <returns>True if wizard can be completed.</returns>
    public bool CanFinish()
    {
      return CanFinish(false);
    }

    /// <summary>
    /// Determines whether a wizard can successfully Finish itself yet.
    /// </summary>
    /// <param name="showUI">Boolean that determines if the first WizardPage that needs to be satisfied.</param>
    /// <returns>True if wizard can be completed.</returns>
    public bool CanFinish(bool showUI) 
    {
      // ask all pages in current group if they can proceed
      foreach (object o in _CurrentGroup) 
      {
        WizardPage validatablePage = o as WizardPage;
        if (null != o && !validatablePage.CanProceed(false)) 
        {
          if (showUI) 
          {
            // find the first page that's not ready
            // show it, then let the page put up its UI
            _CurrentPage = _CurrentGroup.IndexOf(validatablePage);
            ShowPage();
            validatablePage.CanProceed(true);
          }
                    
          return false;
        }
      }
      return true;
    }

    #endregion

    #region public Members
    /// <summary>
    /// The current group of wizard pages shown in the wizard.
    /// </summary>
    /// <value>A string that matches groups that you specified while adding
    /// wizard pages.</value>
    public virtual string CurrentGroup
    {
      get
      {
        return _CurrentGroupName;
      }
      set
      {
        _CurrentGroupName = value;
        this.CalculatePageGroup();
        if (_CurrentGroup.Count == 0)
        {
          throw new ArgumentOutOfRangeException("CurrentGroup", value, "Group does not exist");
        }

        if (_isLoaded)
        {
          // Cache the current Page
          Control currentPage = pagePanel.Controls[0] as Control;

          // Set the current page if out of scope
          if (_CurrentPage > _CurrentGroup.Count)
          {
            _CurrentPage = _CurrentGroup.Count - 1;
            ShowPage();
          }
          else if (_CurrentGroup[_CurrentPage] != currentPage)
          {
            // Set Minimum Size
            this.Size = this.MinimumSize;

            // Resize
            ResizeDialog();
          }

          // Re-show title (to fix page counts)
          ShowTitleBar();

          // Recalc the prev/next buttons to allow for a group of 1
          SetActiveButtons();
        }

      }
    }

    /// <summary>
    /// The text for the titlebar of the wizard.
    /// </summary>
    public new string Text
    {
      get
      {
        return _TitleBarText;
      }
      set
      {
        _TitleBarText = value;
      }
    }

    /// <summary>
    /// The number of pages in the current group.
    /// </summary>
    public int PageCount
    {
      get
      {
        CalculatePageGroup();
        return _CurrentGroup.Count;
      }
    }

    /// <summary>
    /// Shows or hides the help button.
    /// </summary>
    public bool ShowHelpButton
    {
      get
      {
        return hlpButton.Visible;
      }
      set
      {
        hlpButton.Visible = value;
      }
    }

    /// <summary>
    /// Enables or disables the Finish button.  Usually called from
    /// the pages of the Wizard Page to ensure that some pages are viewed.
    /// </summary>
    public bool EnableFinishButton
    {
      get
      {
        return finishButton.Enabled;
      }
      set
      {
        finishButton.Enabled = value;
      }
    }
    #endregion

    #region protected Implementation
    /// <summary>
    /// Shows the current page in the wizard.
    /// </summary>
    protected virtual void ShowPage()
    {
      // Check for empty collection
      System.Diagnostics.Debug.Assert(_Pages.Count > 0, "Page Collection Empty");

      // Show the Page
      if (CurrentPage != null)
      {
        if (!pagePanel.Controls.Contains(CurrentPage))
        {
          pagePanel.Controls.Clear();
          pagePanel.Controls.Add(CurrentPage);
          CurrentPage.Top = 0;
          CurrentPage.Left = 0;
          CurrentPage.Dock = DockStyle.Fill;
        }
      }

      // Show title
      ShowTitleBar();

      // Set the Active Buttons
      SetActiveButtons();

      // Activate the Page
      CurrentPage.Focus();

      // Repaint
      this.Refresh();
      
    }

    /// <summary>
    /// Shows the title of the page and the page numbers on the TitleBar
    /// </summary>
    protected virtual void ShowTitleBar()
    {
      // Show Title
      base.Text = string.Format("{0} ({1} of {2})", _TitleBarText, _CurrentPage + 1, _CurrentGroup.Count);
    }

    /// <summary>
    /// Enables/Disables next/prev buttons based on the pages
    /// </summary>
    protected virtual void SetActiveButtons()
    {
      // Enable/Disable next/prev buttons
      prevButton.Enabled = (_CurrentPage != 0);
      nextButton.Enabled = (_CurrentPage + 1 != _CurrentGroup.Count);

    }

    /// <summary>
    /// resizes the dialog based on the pages in the current group.
    /// </summary>
    protected virtual void ResizeDialog()
    {
      // Go through each page in the group and resize the containing dialog
      foreach (Control page in _CurrentGroup)
      {
        Size size = ClientSize;
        if (size.Width < page.Width)
        {
          size.Width = page.Width;
          pagePanel.Width = page.Width;
        }
        if (size.Height < page.Height + BottomPanel.Height)
        {
          size.Height = page.Height + BottomPanel.Height;
        }
        ClientSize = size;
      }
    }

    /// <summary>
    /// Virtual method called with the help button is pressed.  The default
    /// behavior is to do nothing.
    /// </summary>
    protected virtual void OnHelp()
    {
    }
    #endregion

    #region private Methods
    /// <summary>
    /// Returns the current page in the current group.
    /// </summary>
    /// <returns>The current page as a Control object.</returns>
    private Control CurrentPage
    {
      get
      {
        return _CurrentGroup[_CurrentPage] as Control;
      }
    }

    /// <summary>
    /// Calculates the pages for the current page group.
    /// </summary>
    private void CalculatePageGroup()
    {
      // Current Page 
      int currentPointer = 0;
      
      // Clear the current Group
      _CurrentGroup.Clear();

      // Construct the Current Group Array
      foreach (PageObject obj in _Pages)
      {
        if (obj.Groups.Contains(_CurrentGroupName))
        {
          _CurrentGroup.Add(obj.Page);
          ++currentPointer;
        }
      }

      // If we can't find any pages
      if (currentPointer == 0) 
      {
        // Throw an exception
        throw new Exception(string.Format("The {0} group contains no pages, please set the CurrentGroup of the WizardSheet", _CurrentGroupName));
      }
    }
    #endregion

    #region Private Fields
    // Default Group
    private const string DEFAULTGROUP = "__DefaultGroup";

    // List of Page Structures
    private ArrayList _Pages = new ArrayList();

    // Current Group of Pages
    private ArrayList _CurrentGroup = new ArrayList();

    // Page Cursor
    private int _CurrentPage = 0;

    // Current Group
    private string _CurrentGroupName = DEFAULTGROUP;

    // Titlebar Text
    private string _TitleBarText = "WizardSheet";
        

    // Boolean to indicate whether WizardSheet_Load has happened yet
    private bool _isLoaded = false;
    #endregion

    #region Event Handlers
    private void WizardSheet_Load(object sender, System.EventArgs e)
    {
      // Only setup the page in Design Mode
      if(this.DesignMode == false)
      {
        // Calculate the Group
        CalculatePageGroup();

        // Resize the Dialog
        ResizeDialog();

        // Throw the isloaded flag
        _isLoaded = true;
      
        // Show the current Page
        ShowPage();
      }
    }

    private void nextButton_Click(object sender, System.EventArgs e)
    {
      _CurrentPage++;
      ShowPage();
    }

    private void prevButton_Click(object sender, System.EventArgs e)
    {
      _CurrentPage--;
      ShowPage();
    }

    private void finishButton_Click(object sender, System.EventArgs e)
    {
      DialogResult = DialogResult.OK;
      Close();
    }
    private void InitializeComponent()
    {
      this.BottomPanel = new System.Windows.Forms.Panel();
      this.cancelBtn = new System.Windows.Forms.Button();
      this.hlpButton = new System.Windows.Forms.Button();
      this.prevButton = new System.Windows.Forms.Button();
      this.nextButton = new System.Windows.Forms.Button();
      this.finishButton = new System.Windows.Forms.Button();
      this.pagePanel = new System.Windows.Forms.Panel();
      this.BottomPanel.SuspendLayout();
      this.SuspendLayout();
      // 
      // BottomPanel
      // 
      this.BottomPanel.Anchor = ((System.Windows.Forms.AnchorStyles.Bottom | System.Windows.Forms.AnchorStyles.Left) 
        | System.Windows.Forms.AnchorStyles.Right);
      this.BottomPanel.Controls.AddRange(new System.Windows.Forms.Control[] {
                                                                              this.cancelBtn,
                                                                              this.hlpButton,
                                                                              this.prevButton,
                                                                              this.nextButton,
                                                                              this.finishButton});
      this.BottomPanel.Location = new System.Drawing.Point(0, 272);
      this.BottomPanel.Name = "BottomPanel";
      this.BottomPanel.Size = new System.Drawing.Size(400, 40);
      this.BottomPanel.TabIndex = 0;
      // 
      // cancelBtn
      // 
      this.cancelBtn.Anchor = (System.Windows.Forms.AnchorStyles.Bottom | System.Windows.Forms.AnchorStyles.Right);
      this.cancelBtn.DialogResult = System.Windows.Forms.DialogResult.Cancel;
      this.cancelBtn.Location = new System.Drawing.Point(256, 8);
      this.cancelBtn.Name = "cancelBtn";
      this.cancelBtn.Size = new System.Drawing.Size(64, 23);
      this.cancelBtn.TabIndex = 3;
      this.cancelBtn.Text = "&Cancel";
      this.cancelBtn.Click += new System.EventHandler(this.cancelBtn_Click);
      // 
      // hlpButton
      // 
      this.hlpButton.Anchor = (System.Windows.Forms.AnchorStyles.Bottom | System.Windows.Forms.AnchorStyles.Left);
      this.hlpButton.Location = new System.Drawing.Point(8, 8);
      this.hlpButton.Name = "hlpButton";
      this.hlpButton.Size = new System.Drawing.Size(64, 23);
      this.hlpButton.TabIndex = 0;
      this.hlpButton.Text = "&Help";
      this.hlpButton.Visible = false;
      this.hlpButton.Click += new System.EventHandler(this.hlpButton_Click);
      // 
      // prevButton
      // 
      this.prevButton.Anchor = (System.Windows.Forms.AnchorStyles.Bottom | System.Windows.Forms.AnchorStyles.Right);
      this.prevButton.Location = new System.Drawing.Point(112, 8);
      this.prevButton.Name = "prevButton";
      this.prevButton.Size = new System.Drawing.Size(64, 23);
      this.prevButton.TabIndex = 1;
      this.prevButton.Text = "< &Last";
      this.prevButton.Click += new System.EventHandler(this.prevButton_Click);
      // 
      // nextButton
      // 
      this.nextButton.Anchor = (System.Windows.Forms.AnchorStyles.Bottom | System.Windows.Forms.AnchorStyles.Right);
      this.nextButton.Location = new System.Drawing.Point(184, 8);
      this.nextButton.Name = "nextButton";
      this.nextButton.Size = new System.Drawing.Size(64, 23);
      this.nextButton.TabIndex = 2;
      this.nextButton.Text = "&Next >";
      this.nextButton.Click += new System.EventHandler(this.nextButton_Click);
      // 
      // finishButton
      // 
      this.finishButton.Anchor = (System.Windows.Forms.AnchorStyles.Bottom | System.Windows.Forms.AnchorStyles.Right);
      this.finishButton.DialogResult = System.Windows.Forms.DialogResult.OK;
      this.finishButton.Location = new System.Drawing.Point(328, 8);
      this.finishButton.Name = "finishButton";
      this.finishButton.Size = new System.Drawing.Size(64, 23);
      this.finishButton.TabIndex = 4;
      this.finishButton.Text = "&Finish";
      this.finishButton.Click += new System.EventHandler(this.finishButton_Click);
      // 
      // pagePanel
      // 
      this.pagePanel.Anchor = (((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Bottom) 
        | System.Windows.Forms.AnchorStyles.Left) 
        | System.Windows.Forms.AnchorStyles.Right);
      this.pagePanel.Name = "pagePanel";
      this.pagePanel.Size = new System.Drawing.Size(400, 272);
      this.pagePanel.TabIndex = 1;
      // 
      // WizardSheet
      // 
      this.AcceptButton = this.finishButton;
      this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
      this.CancelButton = this.cancelBtn;
      this.ClientSize = new System.Drawing.Size(402, 312);
      this.Controls.AddRange(new System.Windows.Forms.Control[] {
                                                                  this.pagePanel,
                                                                  this.BottomPanel});
      this.FormBorderStyle = System.Windows.Forms.FormBorderStyle.FixedSingle;
      this.MaximizeBox = false;
      this.MinimizeBox = false;
      this.MinimumSize = new System.Drawing.Size(408, 296);
      this.Name = "WizardSheet";
      this.ShowInTaskbar = false;
      this.StartPosition = System.Windows.Forms.FormStartPosition.CenterParent;
      this.Load += new System.EventHandler(this.WizardSheet_Load);
      this.HelpRequested += new System.Windows.Forms.HelpEventHandler(this.WizardSheet_HelpRequested);
      this.BottomPanel.ResumeLayout(false);
      this.ResumeLayout(false);

    }

    private void cancelBtn_Click(object sender, System.EventArgs e)
    {
      this.DialogResult = DialogResult.Cancel;
      this.Close();
    }

    private void WizardSheet_HelpRequested(object sender, System.Windows.Forms.HelpEventArgs hlpevent)
    {
      hlpButton_Click(sender, hlpevent as EventArgs);
    }

    private void hlpButton_Click(object sender, System.EventArgs e)
    {
      OnHelp();      
    }

    #endregion

    #region Page Controls
    protected System.Windows.Forms.Panel BottomPanel;
    protected System.Windows.Forms.Button finishButton;
    protected System.Windows.Forms.Button nextButton;
    protected System.Windows.Forms.Button prevButton;
    protected System.Windows.Forms.Button hlpButton;
    protected System.Windows.Forms.Button cancelBtn;
    protected System.Windows.Forms.Panel pagePanel;
    #endregion

  }

  /// <summary>
  /// A simple structure to hold the page and groups that the page belongs to.
  /// </summary>
  internal class PageObject
  {
    public PageObject(Control page, string[] groups)
    {
      Page = page;
      Groups = new StringCollection();
      Groups.AddRange(groups);
    }

    public Control Page;
    public StringCollection Groups;
  }

  /// <summary>
  /// A User Control that provides a base class to derive from to create 
  /// standard looking wizard pages.  It supports a top pane that contains a 
  /// title and description of a page.
  /// </summary>
  public class WizardPage : System.Windows.Forms.UserControl
  {
    /// <summary> 
    /// Required designer variable.
    /// </summary>
    private System.ComponentModel.Container components = null;

    /// <summary>
    /// The default constructor.
    /// </summary>
    public WizardPage()
    {
      // This call is required by the Windows.Forms Form Designer.
      InitializeComponent();

      // Set the title and description
      titleLabel.Text = string.Format("{0} Wizard", this.GetType().Name);
      descriptionLabel.Text = string.Format("The {0} Wizard.", this.GetType().Name);
    }

    /// <summary>
    /// Creates a new wizard page, specifying the title and description.
    /// </summary>
    /// <param name="title">The Title to be shown in the top pane of the page.</param>
    /// <param name="description">The description to be shown in the top pane of the page.</param>
    public WizardPage(string title, string description)
    {
      // This call is required by the Windows.Forms Form Designer.
      InitializeComponent();

      // Set the title and description
      titleLabel.Text = title;
      descriptionLabel.Text = description;
    }

    /// <summary> 
    /// Clean up any resources being used.
    /// </summary>
    protected override void Dispose( bool disposing )
    {
      if( disposing )
      {
        if(components != null)
        {
          components.Dispose();
        }
      }
      base.Dispose( disposing );
    }

        #region Component Designer generated code
    /// <summary> 
    /// Required method for Designer support - do not modify 
    /// the contents of this method with the code editor.
    /// </summary>
    private void InitializeComponent()
    {
      this.panel1 = new System.Windows.Forms.Panel();
      this.descriptionLabel = new System.Windows.Forms.Label();
      this.titleLabel = new System.Windows.Forms.Label();
      this.panel1.SuspendLayout();
      this.SuspendLayout();
      // 
      // panel1
      // 
      this.panel1.BackColor = System.Drawing.SystemColors.Info;
      this.panel1.Controls.AddRange(new System.Windows.Forms.Control[] {
                                                                         this.descriptionLabel,
                                                                         this.titleLabel});
      this.panel1.Dock = System.Windows.Forms.DockStyle.Top;
      this.panel1.Name = "panel1";
      this.panel1.Size = new System.Drawing.Size(400, 56);
      this.panel1.TabIndex = 0;
      // 
      // descriptionLabel
      // 
      this.descriptionLabel.Anchor = ((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Left) 
        | System.Windows.Forms.AnchorStyles.Right);
      this.descriptionLabel.Location = new System.Drawing.Point(16, 24);
      this.descriptionLabel.Name = "descriptionLabel";
      this.descriptionLabel.Size = new System.Drawing.Size(376, 32);
      this.descriptionLabel.TabIndex = 1;
      this.descriptionLabel.Text = "<Description>";
      // 
      // titleLabel
      // 
      this.titleLabel.Anchor = ((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Left) 
        | System.Windows.Forms.AnchorStyles.Right);
      this.titleLabel.Font = new System.Drawing.Font("Microsoft Sans Serif", 10F, System.Drawing.FontStyle.Bold, System.Drawing.GraphicsUnit.Point, ((System.Byte)(0)));
      this.titleLabel.Location = new System.Drawing.Point(8, 8);
      this.titleLabel.Name = "titleLabel";
      this.titleLabel.Size = new System.Drawing.Size(384, 16);
      this.titleLabel.TabIndex = 0;
      this.titleLabel.Text = "<WizardPageTitle>";
      // 
      // WizardPage
      // 
      this.Controls.AddRange(new System.Windows.Forms.Control[] {
                                                                  this.panel1});
      this.Name = "WizardPage";
      this.Size = new System.Drawing.Size(400, 272);
      this.panel1.ResumeLayout(false);
      this.ResumeLayout(false);

    }
        #endregion

    private System.Windows.Forms.Panel panel1;
    private System.Windows.Forms.Label descriptionLabel;
    private System.Windows.Forms.Label titleLabel;

    /// <summary>
    /// The current WizardSheet that the WizardPage is shown within.
    /// </summary>
    /// <value>A reference to the WizardSheet</value>
    public WizardSheet WizardSheet
    {
      get
      {
        return this.ParentForm as WizardSheet;
      }
    }

    /// <summary>
    /// Determines if the Wizard can pass this page.  This is called from WizardSheet.CanFinish() to see if all pages are
    /// satisfied.
    /// </summary>
    /// <param name="showUI">Boolean that indicates whether the WizardPage should show itself if the wizard cannot proceed.</param>
    /// <returns>Boolean that indicates whether a page can proceed or not.</returns>
    public virtual bool CanProceed(bool showUI) 
    {
      return true;
    }

    /// <summary>
    /// Sets the title of the page shown within the header of the 
    /// page.
    /// </summary>
    public string PageTitle
    {
      set
      {
        titleLabel.Text = value;
      }
    }

    /// <summary>
    /// Sets the description of the page shown within the header of the page.
    /// </summary>
    public string PageDescription
    {
      set
      {
        descriptionLabel.Text = value;
      }
    }
  }

}
